@@ -5,6 +5,7 @@ from abc import abstractmethod
 import contextlib
 from functools import partial
 import inspect
+from io import BytesIO
 from itertools import chain
 import logging
 import os
@@ -32,10 +33,6 @@ from typing import (
 import desert
 from marshmallow import fields
 import numpy as np
-from opacus import GradSampleModule, PrivacyEngine
-from opacus.accountants import RDPAccountant
-from opacus.validators import ModuleValidator, register_module_fixer
-from opacus.validators.errors import UnsupportedModuleError
 import pandas as pd
 import pytorch_lightning as pl
 from pytorch_lightning.callbacks import Callback, ModelCheckpoint
@@ -44,15 +41,14 @@ from pytorch_lightning.callbacks.progress import TQDMProgressBar
 from pytorch_lightning.loggers import LightningLoggerBase, TensorBoardLogger
 from sklearn.base import BaseEstimator
 import torch
-from torch import Tensor
-from torch import nn as nn
+from torch import Tensor, nn as nn
 
 # Convention is to import functional as F
 # noinspection PyPep8Naming
 from torch.nn import functional as F
 import torch.optim as optimizers
-import torch.optim.lr_scheduler as schedulers
 from torch.optim.lr_scheduler import _LRScheduler
+import torch.optim.lr_scheduler as schedulers
 from torch.utils.data import DataLoader
 import torch_optimizer as torch_optimizers
 
@@ -75,7 +71,7 @@ from bitfount.backends.pytorch.types import (
 )
 from bitfount.backends.pytorch.utils import LoggerType, _autodetect_gpu
 from bitfount.backends.pytorch.weight_clipper import _PytorchParamConstraint
-from bitfount.config import BITFOUNT_LOGS_DIR, BITFOUNT_OUTPUT_DIR
+from bitfount.config import BITFOUNT_LOGS_DIR, BITFOUNT_OUTPUT_DIR, DP_AVAILABLE
 from bitfount.data.databunch import BitfountDataBunch
 from bitfount.data.dataloaders import BitfountDataLoader
 from bitfount.data.datasets import _BitfountDataset
@@ -103,6 +99,16 @@ from bitfount.utils.logging_utils import filter_stderr
 
 logger = logging.getLogger(__name__)
 
+
+try:
+    from opacus import GradSampleModule, PrivacyEngine
+    from opacus.accountants import RDPAccountant
+    from opacus.validators import ModuleValidator, register_module_fixer
+    from opacus.validators.errors import UnsupportedModuleError
+except ImportError:
+    logger.debug("Opacus not installed, Differential Privacy will not be available.")
+
+
 _OptimizerType = Union[torch_optimizers.Optimizer, optimizers.Optimizer]
 
 _OPTIMIZERS: Dict[str, Type[_OptimizerType]] = {
@@ -162,34 +168,36 @@ def _calculate_embedding_sizes(
     return embedding_sizes
 
 
-# Module fixer implementation that is more friendly with our expected layer sizes, etc.
-# We register this as the preferred module replacer for BatchNorm layers in Opacus.
-@register_module_fixer(  # type: ignore[misc] # Reason: untyped decorator but not ours
-    [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
-)
-def _fix_batch_norm(
-    module: Union[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
-) -> nn.GroupNorm:
-    # A default value of 32 is chosen for the number of groups based on the paper
-    # *Group Normalization* https://arxiv.org/abs/1803.08494.
-    # If this is not appropriate for the num_features in the layer we then try to
-    # find an alternative.
-    num_groups = min(32, module.num_features)
-    while num_groups > 1:
-        # If num_features is divisible by num_groups we can use it
-        if module.num_features % num_groups == 0:
-            break
-        # Otherwise, try the next number down
-        num_groups -= 1
-    else:
-        # If we reach the end of the while-loop we couldn't find a compatible
-        # number of groups.
-        raise UnsupportedModuleError(
-            "Unable to find compatible number of groups for GroupNorm replacement "
-            "of BatchNorm. Tried [2,32]."
-        )
+if DP_AVAILABLE:
+    # Module fixer implementation that is more friendly with our expected layer sizes,
+    # etc. We register this as the preferred module replacer for BatchNorm layers in
+    # Opacus.
+    @register_module_fixer(  # type: ignore[misc] # Reason: untyped decorator (not ours)
+        [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
+    )
+    def _fix_batch_norm(
+        module: Union[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
+    ) -> nn.GroupNorm:
+        # A default value of 32 is chosen for the number of groups based on the paper
+        # *Group Normalization* https://arxiv.org/abs/1803.08494.
+        # If this is not appropriate for the num_features in the layer we then try to
+        # find an alternative.
+        num_groups = min(32, module.num_features)
+        while num_groups > 1:
+            # If num_features is divisible by num_groups we can use it
+            if module.num_features % num_groups == 0:
+                break
+            # Otherwise, try the next number down
+            num_groups -= 1
+        else:
+            # If we reach the end of the while-loop we couldn't find a compatible
+            # number of groups.
+            raise UnsupportedModuleError(
+                "Unable to find compatible number of groups for GroupNorm replacement "
+                "of BatchNorm. Tried [2,32]."
+            )
 
-    return nn.GroupNorm(num_groups, module.num_features, affine=module.affine)
+        return nn.GroupNorm(num_groups, module.num_features, affine=module.affine)
 
 
 @delegates()
@@ -576,6 +584,7 @@ weight-averaging).
         ):
             # TODO: [BIT-1412] allow the user to provide these arguments in the
             # constructor
+            gpu_kwargs = _autodetect_gpu()
             trainer = pl.Trainer(
                 enable_checkpointing=True if self._dp_config else False,
                 max_epochs=self.epochs or -1,
@@ -585,9 +594,9 @@ weight-averaging).
                 auto_lr_find=True,
                 logger=self._pl_logger,
                 callbacks=callbacks,
-                gpus=_autodetect_gpu(),  # will be 1 (first GPU) or 0 (CPU)
                 check_val_every_n_epoch=1,
                 default_root_dir=str(BITFOUNT_OUTPUT_DIR),
+                **gpu_kwargs,
             )
             return trainer
 
@@ -622,7 +631,7 @@ weight-averaging).
         assert self._model is not None  # nosec[assert_used]
         torch.save(self._model.state_dict(), filename)
 
-    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
+    def deserialize(self, content: Union[str, os.PathLike, bytes]) -> None:
         """Deserialize model.
 
         :::danger
@@ -633,14 +642,15 @@ weight-averaging).
         :::
 
         Args:
-            filename: Path to file containing serialized model.
+            content: Path to file containing serialized model.
         """
         if not self._initialised:
             logger.info("Model not yet initialised. Auto-initialising model.")
             self.initialise_model()
         # Model has been initialised, assuring mypy of this
         assert self._model is not None  # nosec[assert_used]
-        self._model.load_state_dict(torch.load(filename))
+        load_contents = BytesIO(content) if isinstance(content, bytes) else content
+        self._model.load_state_dict(torch.load(load_contents))
 
     @staticmethod
     def _get_layer_sizes(
@@ -1263,7 +1273,7 @@ weight-averaging).
             if isinstance(test_dl, BitfountDataLoader):
                 logger.info(
                     f"Using test portion of dataset for inference - this has "
-                    f"{len(test_dl)} record(s)."
+                    f"{len(test_dl.dataset)} record(s)."
                 )
             else:
                 raise ValueError("No test data to infer in the provided datasource.")
@@ -1508,7 +1518,9 @@ class BaseTabNetModel(
             return
         torch.save(self._model.network.state_dict(), filename)  # type: ignore[union-attr] # reason: model is initialised # noqa: B950
 
-    def deserialize(self, filename: Union[str, os.PathLike], **kwargs: Any) -> None:
+    def deserialize(
+        self, content: Union[str, os.PathLike, bytes], **kwargs: Any
+    ) -> None:
         """Deserialize model.
 
         :::danger
@@ -1519,7 +1531,7 @@ class BaseTabNetModel(
         :::
 
         Args:
-            filename: Path to file containing serialized model.
+            content: Bytestream or path to file containing serialized model.
             **kwargs: Keyword arguments provided to `torch.load` under the hood.
         """
         if not self._initialised:
@@ -1529,8 +1541,8 @@ class BaseTabNetModel(
                 "Model can't be initialised without data."
             )
             return
-
-        self._model.network.load_state_dict(torch.load(filename, **kwargs))  # type: ignore[union-attr] # reason: model is initialised # noqa: B950
+        load_contents = BytesIO(content) if isinstance(content, bytes) else content
+        self._model.network.load_state_dict(torch.load(load_contents, **kwargs))  # type: ignore[union-attr] # reason: model is initialised # noqa: B950
 
     def _fit_local(
         self,
