@@ -191,16 +191,16 @@ class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             # self.model_structure.dropout_probs is set in post_init,
             # so we can just cast below
             model = _PyTorchConvNeuralNet(
-                layer_sizes,
-                cast(List[float], self.model_structure.dropout_probs),
-                self.model_structure.mish_activation_function,
-                head_sizes,
-                ff_layer_sizes,
-                self.model_structure.ff_dropout_probs,
-                self.model_structure.kernel_size,
-                self.model_structure.padding,
-                self.model_structure.stride,
-                self.model_structure.pooling_function,
+                layer_sizes=layer_sizes,
+                dropout_probs=cast(List[float], self.model_structure.dropout_probs),
+                mish=self.model_structure.mish_activation_function,
+                head_sizes=head_sizes,
+                ff_layer_sizes=ff_layer_sizes,
+                ff_dropout_probs=self.model_structure.ff_dropout_probs,
+                kernel_size=self.model_structure.kernel_size,
+                padding=self.model_structure.padding,
+                stride=self.model_structure.stride,
+                pooling_function=self.model_structure.pooling_function,
             )
         return model
 
@@ -656,7 +656,6 @@ class TabNetClassifier(PyTorchClassifierMixIn, BaseTabNetModel):
         ]
 
         if self.embedding_sizes is None:
-
             embedding_sizes = _calculate_embedding_sizes(
                 table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
             )
