@@ -25,7 +25,11 @@ import pandas as pd
 from skimage import color, io
 from sqlalchemy import text
 
-from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.base_source import (
+    BaseSource,
+    FileSystemIterableSource,
+    IterableSource,
+)
 from bitfount.data.datasources.database_source import DatabaseSource
 from bitfount.data.datasplitters import DatasetSplitter, PercentageSplitter
 from bitfount.data.exceptions import DataNotLoadedError
@@ -35,7 +39,6 @@ from bitfount.transformations.processor import TransformationProcessor
 from bitfount.utils import _array_version
 
 if TYPE_CHECKING:
-
     from bitfount.data.schema import TableSchema
     from bitfount.data.types import (
         _DataEntry,
@@ -110,7 +113,7 @@ class _BaseBitfountDataset(ABC):
         that we don't end up removing the extra columns in our dataframe that are
         used during training (e.g. loss_weights_col, etc.) but aren't part of the
         schema. Applying the schema adds extra columns to the dataframe if they
-        are missing. Therefore we need to subset the data columns here to ensure
+        are missing. Therefore, we need to subset the data columns here to ensure
         we are only using the columns specified for this task
         """
         diff = list(sorted(set(self.selected_cols) - set(data.columns)))
@@ -268,7 +271,7 @@ class _BaseBitfountDataset(ABC):
 
     def _set_image_values(self, data: pd.DataFrame) -> None:
         """Sets `self.image`."""
-        for (i, col) in enumerate(self.image_columns):
+        for i, col in enumerate(self.image_columns):
             x_img = np.expand_dims(data.loc[:, col].values, axis=1)
             # If there are multiple images, we start concatentating `self.image`
             # with each next image
@@ -368,7 +371,7 @@ class _BaseBitfountDataset(ABC):
 
         The data splitter is resolved in the following order:
             1. BaseSource data_splitter if specified
-            2. Provided data_splitter if specified
+            2. Provided data_splitter if specified (from datastructure)
             3. PercentageSplitter (default)
 
         Returns:
@@ -389,106 +392,6 @@ class _BaseBitfountDataset(ABC):
 
         return data_splitter
 
-    def _split_data(self, data_splitter: Optional[DatasetSplitter] = None) -> None:
-        """Split the data into training, validation and test datasets.
-
-        This method is idempotent so it can be called multiple times without
-        re-splitting the data.
-
-        Args:
-            data_splitter: An optional data splitter object.
-        """
-        if not self.datasource._data_is_split:
-            data_splitter = self._resolve_data_splitter(data_splitter)
-            assert self.datasource.data is not None  # nosec[assert_used]
-
-            (
-                self.datasource._train_idxs,
-                self.datasource._validation_idxs,
-                self.datasource._test_idxs,
-            ) = data_splitter.create_dataset_splits(self.datasource.data)
-
-            self.datasource._data_is_split = True
-        else:
-            logger.debug("Data is already split, keeping the current split.")
-
-    def get_dataset_split_length(
-        self, split: DataSplit, data_splitter: Optional[DatasetSplitter] = None
-    ) -> int:
-        """Returns the length of the specified dataset split.
-
-        Args:
-            split: The split to get the length of.
-            data_splitter: The splitter object used to split the data if the BaseSource
-                does not have one.
-
-        Returns:
-            The length of the specified dataset split.
-
-        Raises:
-            DataNotLoadedError: If unable to get the length of the dataset split.
-        """
-        data_splitter = self._resolve_data_splitter(data_splitter)
-
-        # If PercentageSplitter is used regardless of the data loader.
-        if isinstance(data_splitter, PercentageSplitter):
-            len_loader = len(self.datasource)
-            if split == DataSplit.TRAIN:
-                return int(len_loader * data_splitter.train_percentage / 100)
-            elif split == DataSplit.VALIDATION:
-                return int(len_loader * data_splitter.validation_percentage / 100)
-            elif split == DataSplit.TEST:
-                return int(len_loader * data_splitter.test_percentage / 100)
-
-        # If the data is held in memory with any other splitter.
-        elif self.datasource._data_is_loaded:
-            return len(self.get_dataset_split(split=split, data_splitter=data_splitter))
-
-        # If the data is held in a database with any other splitter. This is because
-        # it requires the database to be queried in order to determine the length of
-        # a split e.g. `SplitterDefinedInData`
-        elif isinstance(self.datasource, DatabaseSource):
-            query = self._resolve_data_splitter(data_splitter).get_split_query(
-                self.datasource, split
-            )
-            with self.datasource.con.connect() as con:
-                # Ignoring the security warning because the sql query is trusted and
-                # will be executed regardless.
-                result = con.execute(text(f"SELECT COUNT(*) FROM ({query}) q"))  # nosec
-                return cast(int, result.scalar_one())
-
-        # `load_data` should be called to avoid this error being raised
-        raise DataNotLoadedError("Unable to get length of dataset split")
-
-    def get_dataset_split(
-        self,
-        split: DataSplit,
-        data_splitter: Optional[DatasetSplitter] = None,
-    ) -> pd.DataFrame:
-        """Returns the relevant portion of `self.data`.
-
-        Args:
-            split: The portion of data to return.
-            data_splitter: The splitter object used to split the data.
-
-        Returns:
-            A dataframe-type object containing the data points specified by the data
-            splitter.
-
-        Raises:
-            DataNotLoadedError: If data has not been loaded.
-        """
-        if not self.datasource._data_is_loaded:
-            raise DataNotLoadedError(
-                "Please load data before accessing a split. "
-                "If the data is remote, it must be iterated."
-            )
-
-        self._split_data(data_splitter)  # idempotent
-        indices: np.ndarray = getattr(self.datasource, f"_{split.value}_idxs")
-        df: pd.DataFrame = self.datasource.data.loc[indices.tolist()]
-        return df.reset_index(drop=True)
-
 
 class _BitfountDataset(_BaseBitfountDataset):
     """A dataset for supervised tasks.
@@ -536,6 +439,58 @@ class _BitfountDataset(_BaseBitfountDataset):
     def __len__(self) -> int:
         return len(self.x_var[0])
 
+    def get_dataset_split(
+        self,
+        split: DataSplit,
+        data_splitter: Optional[DatasetSplitter] = None,
+    ) -> pd.DataFrame:
+        """Returns the relevant portion of `self.data`.
+
+        Args:
+            split: The portion of data to return.
+            data_splitter: The splitter object used to split the data.
+
+        Returns:
+            A dataframe-type object containing the data points specified by the data
+            splitter.
+
+        Raises:
+            DataNotLoadedError: If data has not been loaded.
+        """
+        if not self.datasource._data_is_loaded:
+            raise DataNotLoadedError(
+                "Please load data before accessing a split. "
+                "If the data is remote, it must be iterated."
+            )
+
+        self._split_data(data_splitter)  # idempotent
+        indices: np.ndarray = getattr(self.datasource, f"_{split.value}_idxs")
+        df: pd.DataFrame = self.datasource.data.loc[indices.tolist()]
+        return df.reset_index(drop=True)
+
+    def _split_data(self, data_splitter: Optional[DatasetSplitter] = None) -> None:
+        """Split the data into training, validation and test datasets.
+
+        This method is idempotent so it can be called multiple times without
+        re-splitting the data.
+
+        Args:
+            data_splitter: An optional data splitter object.
+        """
+        if not self.datasource._data_is_split:
+            data_splitter = self._resolve_data_splitter(data_splitter)
+            assert self.datasource.data is not None  # nosec[assert_used]
+
+            (
+                self.datasource._train_idxs,
+                self.datasource._validation_idxs,
+                self.datasource._test_idxs,
+            ) = data_splitter.create_dataset_splits(self.datasource.data)
+
+            self.datasource._data_is_split = True
+        else:
+            logger.debug("Data is already split, keeping the current split.")
+
 
 class _IterableBitfountDataset(_BaseBitfountDataset):
     """Iterable Dataset.
@@ -543,6 +498,8 @@ class _IterableBitfountDataset(_BaseBitfountDataset):
     Currently, this is only used for Database connections.
     """
 
+    datasource: IterableSource
+
     def __iter__(self) -> Iterator[_DataEntry]:
         """Iterates over the dataset."""
         for data_partition in self.yield_dataset_split(
@@ -553,6 +510,48 @@ class _IterableBitfountDataset(_BaseBitfountDataset):
             for idx in range(len(self.data)):
                 yield self._getitem(idx)
 
+    def get_dataset_split_length(
+        self, split: DataSplit, data_splitter: Optional[DatasetSplitter] = None
+    ) -> int:
+        """Returns the length of the specified dataset split.
+
+        Args:
+            split: The split to get the length of.
+            data_splitter: The splitter object used to split the data if the BaseSource
+                does not have one.
+
+        Returns:
+            The length of the specified dataset split.
+
+        Raises:
+            DataNotLoadedError: If unable to get the length of the dataset split.
+        """
+        data_splitter = self._resolve_data_splitter(data_splitter)
+
+        # If PercentageSplitter is used regardless of the data loader.
+        if isinstance(data_splitter, PercentageSplitter):
+            len_datasource = len(self.datasource)
+            if split == DataSplit.TRAIN:
+                return int(len_datasource * data_splitter.train_percentage / 100)
+            elif split == DataSplit.VALIDATION:
+                return int(len_datasource * data_splitter.validation_percentage / 100)
+            elif split == DataSplit.TEST:
+                return int(len_datasource * data_splitter.test_percentage / 100)
+
+        # If the data is held in a database with any other splitter. This is because
+        # it requires the database to be queried in order to determine the length of
+        # a split e.g. `SplitterDefinedInData`
+        elif isinstance(self.datasource, DatabaseSource):
+            query = data_splitter.get_split_query(self.datasource, split)
+            with self.datasource.con.connect() as con:
+                # Ignoring the security warning because the sql query is trusted and
+                # will be executed regardless.
+                result = con.execute(text(f"SELECT COUNT(*) FROM ({query}) q"))  # nosec
+                return cast(int, result.scalar_one())
+
+        # `load_data` should be called to avoid this error being raised
+        raise DataNotLoadedError("Unable to get length of dataset split")
+
     @cached_property
     def _len(self) -> int:
         """Returns the length of the dataset."""
@@ -580,22 +579,18 @@ class _IterableBitfountDataset(_BaseBitfountDataset):
         Raises:
             ValueError: If no query or table name has been supplied for multi-table
                 data.
-            ValueError: If attempting to iterate over in-memory data.
         """
+        kwargs: Dict[str, Any] = {}
+        data_splitter = self._resolve_data_splitter(data_splitter)
+
         if isinstance(self.datasource, DatabaseSource):
             if not self.datasource.query:
                 raise ValueError("No query or table name specified.")
 
-            query = self._resolve_data_splitter(data_splitter).get_split_query(
-                self.datasource, split
-            )
-            with self.datasource.con.connect() as con:
-                result = con.execute(text(query))
-                for partition in result.partitions(
-                    self.datasource.database_partition_size
-                ):
-                    yield pd.DataFrame(partition, columns=list(result.keys()))
-        else:
-            raise ValueError(
-                "Iterating over in-memory data is currently not supported."
-            )
+            kwargs["query"] = data_splitter.get_split_query(self.datasource, split)
+
+        elif isinstance(self.datasource, FileSystemIterableSource):
+            kwargs["file_names"] = data_splitter.get_filenames(self.datasource, split)
+
+        for data_partition in self.datasource.yield_data(**kwargs):
+            yield data_partition
