@@ -97,10 +97,13 @@ class PercentageSplitter(DatasetSplitter):
             dataset. This sorts the dataframe by the values of those features
             ensuring the validation and test sets come after the training set data
             to remove potential bias during training and evaluation. Defaults to None.
+        shuffle: A bool indicating whether we shuffle the data for the splits.
+            Defaults to True.
     """
 
     validation_percentage: int = 10
     test_percentage: int = 10
+    shuffle: bool = True
     time_series_sort_by: Optional[Union[List[str], str]] = None
 
     def __post_init__(self) -> None:
@@ -135,7 +138,7 @@ class PercentageSplitter(DatasetSplitter):
         # Get indices of dataframe
         indices = data.index.to_series().to_list()
 
-        if not self.time_series_sort_by:
+        if not self.time_series_sort_by and self.shuffle is True:
             random.shuffle(indices)
 
         # Create dataset splits
@@ -164,11 +167,10 @@ class PercentageSplitter(DatasetSplitter):
                 train_idxs, test_idxs = self._split_indices_time_series(
                     train_idxs, test_idxs, data
                 )
-
-            np.random.shuffle(train_idxs)
-            np.random.shuffle(validation_idxs)
-            np.random.shuffle(test_idxs)
-
+            if self.shuffle is True:
+                np.random.shuffle(train_idxs)
+                np.random.shuffle(validation_idxs)
+                np.random.shuffle(test_idxs)
         return train_idxs, validation_idxs, test_idxs
 
     def get_split_query(self, datasource: DatabaseSource, split: DataSplit) -> str:
@@ -304,7 +306,10 @@ class PercentageSplitter(DatasetSplitter):
         else:
             raise ValueError("Split not recognised")
 
-        return datasource.file_names[offset : offset + limit]
+        filenames = datasource.file_names[offset : offset + limit]
+        if datasource.selected_file_names:
+            return [i for i in filenames if i in datasource.selected_file_names]
+        return filenames
 
 
 @dataclass
