@@ -6,9 +6,11 @@ import ast
 from collections import Counter, OrderedDict
 import copy
 from enum import Enum
+import logging
 from typing import (
     TYPE_CHECKING,
     Any,
+    Callable,
     Dict,
     Iterable,
     List,
@@ -22,30 +24,71 @@ from typing import (
     Union,
 )
 
-import databricks.koalas as ks
-from marshmallow import Schema as MarshmallowSchema
-from marshmallow import ValidationError, fields, post_dump, post_load, pre_dump
+from marshmallow import (
+    Schema as MarshmallowSchema,
+    ValidationError,
+    fields,
+    post_dump,
+    post_load,
+    pre_dump,
+)
 from marshmallow.fields import Field
 from marshmallow_enum import EnumField
+from mypy_extensions import Arg, KwArg
 import numpy as np
 import pandas as pd
 from pandas._typing import Dtype
 from pandas.core.dtypes.common import pandas_dtype
 from typing_extensions import NotRequired
 
-from bitfount.types import _JSONDict
+from bitfount.types import _Dtypes, _JSONDict
+from bitfount.utils import delegates
 
 if TYPE_CHECKING:
+    from bitfount.data.datasources.base_source import BaseSource
     from bitfount.data.schema import TableSchema
 
+logger = logging.getLogger(__name__)
+
 # Type aliases for dunder methods
 T = TypeVar("T")
 _SingleOrMulti = Union[T, Sequence[T]]
 _ImagesData = _SingleOrMulti[np.ndarray]
+_TabularData = np.ndarray
+_SupportData = np.ndarray
 _Y_VAR = np.ndarray
-_ImageAndTabularEntry = Tuple[Tuple[np.ndarray, _ImagesData, np.ndarray], _Y_VAR]
-_TabularEntry = Tuple[Tuple[np.ndarray, np.ndarray], _Y_VAR]
-_ImageEntry = Tuple[Tuple[_ImagesData, np.ndarray], _Y_VAR]
+_TabularEntry = Tuple[Tuple[_TabularData, _SupportData], _Y_VAR]
+_ImageEntry = Tuple[Tuple[_ImagesData, _SupportData], _Y_VAR]
+_ImageAndTabularEntry = Tuple[Tuple[_TabularData, _ImagesData, _SupportData], _Y_VAR]
+_Segmentation_ImageEntry = Tuple[Tuple[_ImagesData, _SupportData], _ImagesData]
+_Segmentation_ImageAndTabEntry = Tuple[
+    Tuple[_TabularData, _ImagesData, _SupportData], _ImagesData
+]
+_DataEntry = Union[
+    _TabularEntry,
+    _ImageEntry,
+    _ImageAndTabularEntry,
+    _Segmentation_ImageEntry,
+    _Segmentation_ImageAndTabEntry,
+]
+_DataBatch = List[_DataEntry]
+
+# Type alias for the values in the `SemanticType` enum
+_SemanticTypeValue = Literal["categorical", "continuous", "image", "text"]
+S = TypeVar("S", bound="BaseSource")
+_Column = Union[np.ndarray, pd.Series]
+_GetColumnCallable = Callable[
+    [Arg(S, "self"), Arg(str, "col_name"), KwArg(Any)], _Column  # noqa: F821
+]
+_GetDtypesCallable = Callable[[Arg(S, "self"), KwArg(Any)], _Dtypes]  # noqa: F821
+
+
+class DataSplit(Enum):
+    """Enum corresponding to the available data splits."""
+
+    TRAIN = "train"
+    VALIDATION = "validation"
+    TEST = "test"
 
 
 class _CamelCaseSchema(MarshmallowSchema):
@@ -88,7 +131,7 @@ class _ExtendableLabelEncoder:
             new_encoder.classes = dict(data["classes"])
             return new_encoder
 
-    def add_values(self, values: Union[np.ndarray, pd.Series, ks.Series]) -> None:
+    def add_values(self, values: Union[np.ndarray, pd.Series]) -> None:
         """Adds all entries in the column to the set."""
         uniques: Iterable
         if isinstance(values, np.ndarray):
@@ -96,7 +139,7 @@ class _ExtendableLabelEncoder:
         else:
             uniques = sorted(set(values.astype(self.dtype).tolist()))
 
-        # Removes classes that are already present in `self.classes`
+        # Remove classes that are already present in `self.classes`
         new_uniques = sorted(u for u in uniques if u not in self.classes)
         cur_size = len(self.classes)
 
@@ -116,12 +159,51 @@ class _ExtendableLabelEncoder:
         for value, mapping in values.items():
             self.classes[value] = mapping
 
-    def transform(self, values: Union[pd.Series, ks.Series]) -> List[int]:
-        """Transforms the given column."""
+    def transform(self, values: pd.Series) -> List[int]:
+        """Transforms the given column and returns it as a list of encoded values.
+
+        Args:
+            values: The column to encode as a `pandas.Series`.
+
+        Returns:
+            A list of encoded values.
+
+        Raises:
+            ValueError: If the encoder encounters a previously unseen label.
+        """
+        previously_unseen_label: str
         try:
             return [self.classes[v] for v in values.astype(self.dtype).tolist()]
         except KeyError as err:
-            raise ValueError("Previously unseen label: %s" % str(err))
+            previously_unseen_label = str(err)
+            logger.debug(f"Previously unseen label: {previously_unseen_label}")
+
+        # If the encoder encounters a previously unseen label, we instead try to
+        # interpret it as a datetime and see if there is a match there instead.
+        # This is done because the same datetime can have multiple different formats
+        # and pandas and numpy display these differently when converting to a string
+        # so we try to convert them into the same format before encoding.
+        logger.debug(f"Interpreting {previously_unseen_label} as a datetime.")
+        try:
+            classes = {}
+            for _cls, val in self.classes.items():
+                date = str(pd.to_datetime(_cls))
+                classes[date] = val
+
+            # We first try to see if there is a match once the schema label has been
+            # converted into a datetime. If there is no match, this could be due to an
+            # empty time component in the value to be encoded so we try again after
+            # adding this.
+            return [
+                classes.get(v, classes[str(pd.to_datetime(v))])
+                for v in values.astype(self.dtype).tolist()
+            ]
+        except Exception:
+            logger.debug(
+                f"Failed to interpret {previously_unseen_label} as a datetime."
+            )
+
+        raise ValueError(f"Previously unseen label: {previously_unseen_label}")
 
     @property
     def size(self) -> int:
@@ -146,9 +228,6 @@ class SemanticType(Enum):
     TEXT = "text"
 
 
-_SemanticTypeValue = Literal["categorical", "continuous", "image", "text"]
-
-
 class _SemanticTypeRecord:
     """Simple semantic type wrapper for an individual record.
 
@@ -183,10 +262,12 @@ class _SemanticTypeRecord:
         raise NotImplementedError
 
     class _Schema(_CamelCaseSchema):
-
         feature_name = fields.Str()
         dtype = fields.Str()
-        semantic_type = EnumField(SemanticType, by_value=True)
+        semantic_type = EnumField(
+            SemanticType,
+            by_value=True,
+        )
         description = fields.Str(allow_none=True)
 
         @post_dump
@@ -216,6 +297,7 @@ class _SemanticTypeRecord:
             return data
 
 
+@delegates()
 class CategoricalRecord(_SemanticTypeRecord):
     """Stores information for a categorical feature in the schema.
 
@@ -260,7 +342,6 @@ class CategoricalRecord(_SemanticTypeRecord):
             schema.features["categorical"][record.feature_name] = record
 
     class _Schema(_SemanticTypeRecord._Schema):
-
         encoder = fields.Nested(_ExtendableLabelEncoder._Schema)
 
         @post_load
@@ -270,6 +351,7 @@ class CategoricalRecord(_SemanticTypeRecord):
             return CategoricalRecord(**data)
 
 
+@delegates()
 class TextRecord(_SemanticTypeRecord):
     """Stores information for a text feature in the schema."""
 
@@ -306,6 +388,7 @@ class TextRecord(_SemanticTypeRecord):
             return TextRecord(**data)
 
 
+@delegates()
 class ContinuousRecord(_SemanticTypeRecord):
     """Stores information for a continuous feature in the schema."""
 
@@ -358,6 +441,7 @@ class StrDictField(fields.Field):
             raise ValidationError("Field should be str or dict")
 
 
+@delegates()
 class ImageRecord(_SemanticTypeRecord):
     """Stores information for an image feature in the schema."""
 
@@ -410,7 +494,6 @@ class ImageRecord(_SemanticTypeRecord):
             schema.features["image"][record.feature_name] = record
 
     class _Schema(_SemanticTypeRecord._Schema):
-
         dimensions = fields.Dict()
         formats = fields.Dict()
         modes = fields.Dict()
