@@ -7,8 +7,10 @@ from typing import Any, ClassVar, Dict, List, Mapping, Optional, Union, cast, ov
 import numpy as np
 import pandas as pd
 
-from bitfount.federated.aggregators.aggregator import _ModellerSide as ModellerSide_
-from bitfount.federated.aggregators.aggregator import _WorkerSide as WorkerSide_
+from bitfount.federated.aggregators.aggregator import (
+    _ModellerSide as ModellerSide_,
+    _WorkerSide as WorkerSide_,
+)
 from bitfount.federated.aggregators.base import AggregationType, _BaseAggregatorFactory
 from bitfount.federated.exceptions import AggregatorError
 from bitfount.federated.secure import SecureShare, _secure_share_registry
@@ -289,7 +291,6 @@ class SecureAggregator(_BaseAggregatorFactory, _InterPodAggregatorWorkerFactory)
         secure_share: Optional[SecureShare] = None,
         **kwargs: Any,
     ):
-
         super().__init__(**kwargs)
         self._tensor_shim = _load_default_tensor_shim()
         self._secure_share = secure_share if secure_share else SecureShare()
