@@ -8,9 +8,9 @@ import numpy as np
 
 from bitfount.data.datasources.base_source import BaseSource
 from bitfount.federated.algorithms.base import (
-    _BaseAlgorithmFactory,
-    _BaseModellerAlgorithm,
-    _BaseWorkerAlgorithm,
+    BaseAlgorithmFactory,
+    BaseModellerAlgorithm,
+    BaseWorkerAlgorithm,
 )
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.privacy.differential import DPPodConfig
@@ -20,7 +20,7 @@ from bitfount.utils import delegates
 logger = _get_federated_logger(__name__)
 
 
-class _ModellerSide(_BaseModellerAlgorithm):
+class _ModellerSide(BaseModellerAlgorithm):
     """Modeller side of the ColumnAverage algorithm."""
 
     def initialise(
@@ -38,7 +38,7 @@ class _ModellerSide(_BaseModellerAlgorithm):
         return dict(results)
 
 
-class _WorkerSide(_BaseWorkerAlgorithm):
+class _WorkerSide(BaseWorkerAlgorithm):
     """Worker side of the ColumnAverage algorithm."""
 
     def __init__(self, *, field: str, table_name: str, **kwargs: Any) -> None:
@@ -65,7 +65,7 @@ class _WorkerSide(_BaseWorkerAlgorithm):
 
 
 @delegates()
-class ColumnAverage(_BaseAlgorithmFactory):
+class ColumnAverage(BaseAlgorithmFactory):
     """Simple algorithm for taking the arithmetic mean of a column in a table.
 
     Args:
@@ -80,8 +80,8 @@ class ColumnAverage(_BaseAlgorithmFactory):
     """
 
     fields_dict: ClassVar[T_FIELDS_DICT] = {
-        "field": fields.Str(),
-        "table_name": fields.Str(),
+        "field": fields.Str(required=True),
+        "table_name": fields.Str(required=True),
     }
 
     def __init__(self, *, field: str, table_name: str, **kwargs: Any):
