@@ -6,6 +6,7 @@ from dataclasses import dataclass
 from datetime import datetime, timezone
 from enum import Enum
 from typing import (
+    TYPE_CHECKING,
     Callable,
     Final,
     List,
@@ -38,7 +39,6 @@ from bitfount.federated.types import (
     _PodResponseType,
     _TaskRequestMessageGenerator,
 )
-from bitfount.hub.api import BitfountAM
 from bitfount.hub.authentication_flow import _get_auth_environment
 from bitfount.hub.authentication_handlers import (
     _AUTHORIZATION_PENDING_ERROR,
@@ -53,6 +53,8 @@ from bitfount.hub.types import (
 )
 from bitfount.utils import web_utils
 
+if TYPE_CHECKING:
+    from bitfount.hub.api import BitfountAM
 logger = _get_federated_logger(__name__)
 
 # This forces `requests` to make IPv4 connections
@@ -151,6 +153,7 @@ def _make_encrypted_task_request_message(
     aes_key: bytes,
     pod_public_key: RSAPublicKey,
     project_id: Optional[str],
+    run_on_new_data_only: bool = False,
 ) -> bytes:
     """Common function for generating messages with _EncryptedTaskRequest contents."""
     # Create task request
@@ -163,7 +166,11 @@ def _make_encrypted_task_request_message(
 
     # Create outer message
     message = _TaskRequestMessage(
-        serialized_protocol, auth_type.value, packed_tr.serialize(), project_id
+        serialized_protocol,
+        auth_type.value,
+        packed_tr.serialize(),
+        project_id,
+        run_on_new_data_only,
     )
     return message.serialize()
 
@@ -234,6 +241,7 @@ class _LocalAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
         project_id: Optional[str],
+        run_on_new_data_only: bool = False,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.LOCAL,
@@ -242,6 +250,7 @@ class _LocalAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
             aes_key,
             pod_public_key,
             project_id,
+            run_on_new_data_only,
         )
 
     @classmethod
@@ -295,6 +304,7 @@ class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskReq
             aes_key: bytes,
             pod_public_key: RSAPublicKey,
             project_id: Optional[str] = None,
+            run_on_new_data_only: bool = False,
         ) -> bytes:
             tr = _TaskRequest(serialized_protocol, pod_identifiers, aes_key)
 
@@ -307,6 +317,7 @@ class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskReq
                 IdentityVerificationMethod.KEYS.value,
                 packed_tr.serialize(),
                 project_id,
+                run_on_new_data_only,
             )
             return message.serialize()
 
@@ -387,6 +398,7 @@ class _SAMLAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
         project_id: Optional[str],
+        run_on_new_data_only: bool = False,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.SAML,
@@ -395,6 +407,7 @@ class _SAMLAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
             aes_key,
             pod_public_key,
             project_id,
+            run_on_new_data_only,
         )
 
     @classmethod
@@ -494,6 +507,7 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
         project_id: Optional[str],
+        run_on_new_data_only: bool = False,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.OIDC_ACF_PKCE,
@@ -502,6 +516,7 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
             aes_key,
             pod_public_key,
             project_id,
+            run_on_new_data_only,
         )
 
     @classmethod
@@ -634,6 +649,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
         project_id: Optional[str],
+        run_on_new_data_only: bool = False,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.OIDC_DEVICE_CODE,
@@ -642,6 +658,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
             aes_key,
             pod_public_key,
             project_id,
+            run_on_new_data_only,
         )
 
     @classmethod
