@@ -6,6 +6,8 @@ from asyncio import Task
 import logging
 from typing import TYPE_CHECKING, Any, Set
 
+from bitfount.utils.concurrency_utils import ThreadWithException
+
 if TYPE_CHECKING:
     from bitfount.federated.transport.base_transport import _BaseMailbox
 
@@ -25,23 +27,63 @@ class _FederatedLogger(logging.getLoggerClass()):  # type: ignore[misc] # Reason
 
     def federated_debug(self, msg: object, *args: Any, **kwargs: Any) -> None:
         """Calls `self.debug` under the hood and sets the `federated` attribute."""
-        self.debug(msg, *args, **kwargs, extra={"federated": True})
+        self.debug(
+            msg,
+            *args,
+            **kwargs,
+            # Set stack level to 2 so the record is made from
+            # the _caller_ of this method
+            stacklevel=2,
+            extra={"federated": True},
+        )
 
     def federated_info(self, msg: object, *args: Any, **kwargs: Any) -> None:
         """Calls `self.info` under the hood and sets the `federated` attribute."""
-        self.info(msg, *args, **kwargs, extra={"federated": True})
+        self.info(
+            msg,
+            *args,
+            **kwargs,
+            # Set stack level to 2 so the record is made from
+            # the _caller_ of this method
+            stacklevel=2,
+            extra={"federated": True},
+        )
 
     def federated_warning(self, msg: object, *args: Any, **kwargs: Any) -> None:
         """Calls `self.warning` under the hood and sets the `federated` attribute."""
-        self.warning(msg, *args, **kwargs, extra={"federated": True})
+        self.warning(
+            msg,
+            *args,
+            **kwargs,
+            # Set stack level to 2 so the record is made from
+            # the _caller_ of this method
+            stacklevel=2,
+            extra={"federated": True},
+        )
 
     def federated_error(self, msg: object, *args: Any, **kwargs: Any) -> None:
         """Calls `self.error` under the hood and sets the `federated` attribute."""
-        self.error(msg, *args, **kwargs, extra={"federated": True})
+        self.error(
+            msg,
+            *args,
+            **kwargs,
+            # Set stack level to 2 so the record is made from
+            # the _caller_ of this method
+            stacklevel=2,
+            extra={"federated": True},
+        )
 
     def federated_critical(self, msg: object, *args: Any, **kwargs: Any) -> None:
         """Calls `self.critical` under the hood and sets the `federated` attribute."""
-        self.critical(msg, *args, **kwargs, extra={"federated": True})
+        self.critical(
+            msg,
+            *args,
+            **kwargs,
+            # Set stack level to 2 so the record is made from
+            # the _caller_ of this method
+            stacklevel=2,
+            extra={"federated": True},
+        )
 
 
 class _MailboxHandler(logging.Handler):
@@ -78,16 +120,24 @@ class _MailboxHandler(logging.Handler):
             # TODO: [BIT-1619] [BIT-2260] Add support for federated exception logging.
             record.exc_info = None
 
-            logging_task = asyncio.create_task(
-                self.mailbox.log(record.__dict__), name="Federated Log Sending"
-            )
-
             if record.levelno >= logging.ERROR:
                 # Block flow until high priority messages have completed.
-                asyncio.get_running_loop().run_until_complete(logging_task)
+                # As can't run second event loop in the same thread we farm the
+                # priority logging out to a second thread.
+                message_send_thread = ThreadWithException(
+                    target=asyncio.run,
+                    args=(self.mailbox.log(record.__dict__),),
+                    daemon=True,
+                    name="message_send_thread",
+                )
+                message_send_thread.start()
+                message_send_thread.join()
             else:
                 # Ensure strong reference kept to task so doesn't get garbage
                 # collected, but also that it's still "fire-and-forget".
+                logging_task = asyncio.create_task(
+                    self.mailbox.log(record.__dict__), name="Federated Log Sending"
+                )
                 _background_logging_tasks.add(logging_task)
                 logging_task.add_done_callback(_background_logging_tasks.discard)
         except Exception:
@@ -159,9 +209,7 @@ def _get_federated_logger(name: str) -> _FederatedLogger:
     """
     if not name.startswith("bitfount.federated"):
         raise ValueError("This can only be used from the federated package.")
-
     logging_class = logging.getLoggerClass()
-
     # Use the global logging lock for thread safety
     logging._acquireLock()  # type: ignore[attr-defined] # Reason: known private attribute, see comment # noqa: B950
     try:
