@@ -19,10 +19,8 @@ from typing import (
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
 
+from bitfount.config import BITFOUNT_STORAGE_PATH
 from bitfount.federated.aggregators.secure import SecureAggregator
-from bitfount.federated.algorithms.model_algorithms.base import (
-    _BaseModelAlgorithmFactory,
-)
 from bitfount.federated.authorisation_checkers import (
     _IDENTITY_VERIFICATION_METHODS_MAP,
     IdentityVerificationMethod,
@@ -36,6 +34,12 @@ from bitfount.federated.helper import (
     _create_message_service,
     _get_idp_url,
 )
+from bitfount.federated.keys_setup import (
+    RSAKeyPair,
+    _get_key_id,
+    _get_modeller_keys,
+    _store_key_id,
+)
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.model_reference import BitfountModelReference
 from bitfount.federated.monitoring import task_monitor_context
@@ -61,10 +65,9 @@ from bitfount.hub.api import BitfountHub
 from bitfount.hub.authentication_flow import _get_auth_environment
 from bitfount.hub.exceptions import PodDoesNotExistError
 from bitfount.hub.helper import _default_bitfounthub, _get_pod_public_keys
-from bitfount.types import DistributedModelProtocol
 
 if TYPE_CHECKING:
-    from bitfount.federated.protocols.base import _BaseProtocolFactory
+    from bitfount.federated.protocols.base import BaseProtocolFactory
 
 
 logger = _get_federated_logger(__name__)
@@ -109,7 +112,7 @@ class _Modeller:
 
     def __init__(
         self,
-        protocol: _BaseProtocolFactory,
+        protocol: BaseProtocolFactory,
         message_service: Optional[_MessageService] = None,
         bitfounthub: Optional[BitfountHub] = None,
         pod_public_key_paths: Optional[Mapping[str, Path]] = None,
@@ -132,10 +135,9 @@ class _Modeller:
             check_identity_verification_method(identity_verification_method)
         )
 
-        # Load private key if needed
-        if isinstance(private_key, Path):
-            private_key = _RSAEncryption.load_private_key(private_key)
-        self._private_key = private_key
+        self._private_key = _Modeller._process_private_key(
+            private_key, self._identity_verification_method, self._hub
+        )
 
         self._idp_url: Optional[str]
         if self._identity_verification_method == IdentityVerificationMethod.SAML:
@@ -144,38 +146,144 @@ class _Modeller:
         else:
             self._idp_url = None
 
-    @property
-    def _private_key(self) -> Optional[RSAPrivateKey]:
-        """Private key attribute."""
-        return self.__private_key
+    @staticmethod
+    def _is_public_key_registered(
+        username: str, public_key: RSAPublicKey, hub: BitfountHub
+    ) -> bool:
+        key_id = None
+        try:
+            key_id = _get_key_id(_Modeller._get_modeller_key_storage_path(username))
+        except IOError:
+            logger.debug(
+                "Key present, but no key ID file found. "
+                "Will re-register key with the hub"
+            )
 
-    @_private_key.setter
-    def _private_key(self, private_key: Optional[RSAPrivateKey]) -> None:
-        """Set _private_key attribute."""
-        self.__private_key: Optional[RSAPrivateKey] = None
+        if key_id:
+            # Validate against registered public key or register if not already present
+            logger.debug(
+                f"Checking hub for registered modeller public key"
+                f" for modeller {username} with ID {key_id}"
+            )
+            key_with_metadata = hub.check_public_key_registered_and_active(key_id)
+            if key_with_metadata:
+                if _RSAEncryption.public_keys_equal(
+                    public_key, key_with_metadata["public_key"]
+                ):
+                    logger.debug(
+                        f"Public key already registered for modeller {username}"
+                    )
+                    return True
+                else:
+                    logger.debug(
+                        f"Key provided doesn't match "
+                        f"the one stored on the hub with ID {key_id}"
+                    )
+            else:
+                logger.debug("Key provided, but it's not registered on the hub")
+        return False
 
-        if self._identity_verification_method == IdentityVerificationMethod.KEYS:
-            if not private_key:
-                raise ValueError(
-                    "Key-based identity verification selected but "
-                    "no private key provided."
-                )
-            self.__private_key = private_key
-        else:
+    @staticmethod
+    def _process_private_key(
+        private_key: Optional[Union[RSAPrivateKey, Path]],
+        identity_verification_method: IdentityVerificationMethod,
+        hub: BitfountHub,
+    ) -> Optional[RSAPrivateKey]:
+        # NOTE: This method is static to avoid the risk of it being called during
+        #       _Modeller initialisation before the identity_verification_method
+        #       or hub attributes have been set.
+        # Fail out fast if not using key-based authorisation
+        if identity_verification_method != IdentityVerificationMethod.KEYS:
             if private_key:
                 logger.warning(
                     f"Private key provided but identity verification method "
-                    f'"{self._identity_verification_method.value}" was chosen. '
+                    f'"{identity_verification_method.value}" was chosen. '
                     f"Private key will be ignored."
                 )
+            return None
+
+        username = hub.username
+
+        if isinstance(private_key, Path):
+            # Load private key file if needed
+            logger.debug(
+                f"Loading private key for modeller {username} from {str(private_key)}"
+            )
+            private_key = _RSAEncryption.load_private_key(private_key)
+            public_key = private_key.public_key()
+            # If we loaded it from a file, it may already be registered
+            if _Modeller._is_public_key_registered(username, public_key, hub):
+                return private_key
+
+        elif private_key is None:
+            # Generate/load stored keys
+            logger.info(
+                f"No keys provided for modeller {username}; loading/generating keys"
+            )
+            key_pair = _Modeller._get_modeller_keys(username)
+            private_key = key_pair.private
+            public_key = key_pair.public
+
+            # If the key was loaded it might already be registered
+            if _Modeller._is_public_key_registered(username, public_key, hub):
+                return private_key
+        elif isinstance(private_key, RSAPrivateKey):
+            # Just use private key as is
+            # We can't feasibly check this is registered without iterating
+            # through them all, so we're going to register it
+            public_key = private_key.public_key()
+        else:
+            # Must be invalid type
+            raise TypeError(
+                f"Error processing private key; expected a Path, RSAPrivateKey"
+                f" or None, got {type(private_key)}"
+            )
+
+        # Register new public key
+        logger.info(f"Registering public key for modeller {username}")
+        _store_key_id(
+            _Modeller._get_modeller_key_storage_path(username),
+            str(hub.register_user_public_key(public_key)),
+        )
+
+        return private_key
+
+    @staticmethod
+    def _get_modeller_key_storage_path(username: str) -> Path:
+        """Get the appropriate storage directory for the modeller keys."""
+        # NOTE: We explicitly construct the storage path (rather than relying on
+        #       the user_storage_path from the hub/session) to guarantee that the
+        #       key storage location is explicitly linked to the username and avoids
+        #       using `_default`.
+        return Path(BITFOUNT_STORAGE_PATH / username / "modeller" / "keys")
+
+    @staticmethod
+    def _get_modeller_keys(username: str) -> RSAKeyPair:
+        """Load existing modeller keys or generate new ones.
+
+        Args:
+            username: The user to generate/load the keys for.
+
+        Returns:
+            The loaded/generated keys.
+        """
+        key_storage_path = _Modeller._get_modeller_key_storage_path(username)
+        keys = _get_modeller_keys(key_storage_path)
+        return keys
 
     async def _send_task_requests(
-        self, pod_identifiers: Iterable[str]
+        self,
+        pod_identifiers: Iterable[str],
+        project_id: Optional[str],
+        run_on_new_data_only: bool = False,
     ) -> _ModellerMailbox:
         """Sends task requests to pods.
 
         Args:
             pod_identifiers: The group of pods to run the task with.
+            project_id: Project Id the task belongs to.
+            run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
         Returns:
             The created ModellerMailbox for the task.
@@ -197,6 +305,8 @@ class _Modeller:
             pod_public_keys=loaded_pod_public_keys,
             task_request_msg_gen=task_request_msg_gen,
             message_service=self._message_service,
+            project_id=project_id,
+            run_on_new_data_only=run_on_new_data_only,
         )
 
         return modeller_mailbox
@@ -282,6 +392,8 @@ class _Modeller:
         self,
         pod_identifiers: Iterable[str],
         require_all_pods: bool,
+        project_id: Optional[str] = None,
+        run_on_new_data_only: bool = False,
     ) -> Union[Literal[False], Optional[Any]]:
         """Runs the modeller's task with a set of pods.
 
@@ -290,6 +402,10 @@ class _Modeller:
 
         Args:
             pod_identifiers: The identifiers for the pods to run the task on.
+            require_all_pods: Only run task if all pods are online.
+            project_id: Project Id the task belongs to. Defaults to None.
+            run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
         Returns:
             Whatever the protocol's return value is.
@@ -304,7 +420,7 @@ class _Modeller:
 
             logger.info("Sending task requests...")
             modeller_mailbox: _ModellerMailbox = await self._send_task_requests(
-                pod_identifiers
+                pod_identifiers, project_id, run_on_new_data_only
             )
 
             with task_monitor_context(
@@ -388,34 +504,34 @@ class _Modeller:
         return response_handler
 
     def _serialize(self, model_out: Union[str, os.PathLike]) -> None:
-        """Save model to file.
+        """Save any models to file.
 
-        Save the model to the file specified.
+        Save any algorithms which have models to the filename specified. If there is
+        more than one algorithm with a model, the model name will be be appended with a
+        number, starting from 1.
 
         Args:
             model_out: The path to save the model out to.
         """
-        if self.protocol.class_name != "ResultsOnly":
-            # If protocol is not ResultsOnly, then the algorithm has a model
-            # Assertion to reassure mypy of this fact
-            assert isinstance(  # nosec[assert_used]
-                self.protocol.algorithm, _BaseModelAlgorithmFactory
-            )
-            self.protocol.algorithm.model = cast(
-                DistributedModelProtocol, self.protocol.algorithm.model
-            )
-            logger.info(f'Saving model out to "{model_out}"')
-            self.protocol.algorithm.model.serialize(model_out)
-        else:
-            logger.warning(
-                "Model not saved. ResultsOnly protocol does not produce model."
+        model_algorithms = [i for i in self.protocol.algorithms if hasattr(i, "model")]
+
+        for i, algo in enumerate(model_algorithms, start=1):
+            model_out = (
+                f"{str(model_out)}_{i}" if len(model_algorithms) > 1 else str(model_out)
             )
+            try:
+                algo.model.serialize(model_out)  # type: ignore[attr-defined] # Reason: Captured by try clause  # noqa: B950
+                logger.info(f'Model saved successfully to "{model_out}"')
+            except Exception as e:
+                logger.exception(e)
 
     def run(
         self,
         pod_identifiers: Iterable[str],
         require_all_pods: bool = False,
         model_out: Optional[Union[Path, str]] = None,
+        project_id: Optional[str] = None,
+        run_on_new_data_only: bool = False,
     ) -> Union[Literal[False], Optional[Any]]:
         """Runs the modeller's task with a set of pods.
 
@@ -424,7 +540,11 @@ class _Modeller:
 
         Args:
             pod_identifiers: The identifiers for the pods to run the task on.
+            require_all_pods: Only run task if all pods are online.
             model_out: If specified, path to save the model out to.
+            project_id: Project_id the task belongs to.
+            run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
         Returns:
             Whatever the protocol's return value is.
@@ -441,36 +561,44 @@ class _Modeller:
         # if steps_between_parameter_updates is 1, then we do
         # the clipping at the 'SecureShare' level.
         if (
-            (
-                hasattr(self.protocol, "aggregator")
-                and isinstance(self.protocol.aggregator, SecureAggregator)  # type: ignore[attr-defined] # Reason: see below # noqa: B950
-            )
-            and (
-                hasattr(self.protocol, "steps_between_parameter_updates")
-                and self.protocol.steps_between_parameter_updates != 1  # type: ignore[attr-defined] # Reason: see below # noqa: B950
-            )
-            and hasattr(self.protocol.algorithm, "model")
+            hasattr(self.protocol, "aggregator")
+            and isinstance(self.protocol.aggregator, SecureAggregator)  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+        ) and (
+            hasattr(self.protocol, "steps_between_parameter_updates")
+            and self.protocol.steps_between_parameter_updates != 1  # type: ignore[attr-defined] # Reason: see below # noqa: B950
         ):
-            logger.warning(
-                "SecureAggregation in use. "
-                "We recommend normalization of continuous features prior to training."
-            )
+            if any(hasattr(algo, "model") for algo in self.protocol.algorithms):
+                logger.warning(
+                    "SecureAggregation in use. We recommend normalization "
+                    "of continuous features prior to training."
+                )
+
             # Show warning to user if using custom models that
             # they might have to implement clipping in their custom model.
-
-            if isinstance(self.protocol.algorithm.model, BitfountModelReference):  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+            if any(
+                hasattr(algo, "model")
+                and isinstance(algo.model, BitfountModelReference)  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+                for algo in self.protocol.algorithms
+            ):
                 logger.warning(
                     "You are using a custom model with Secure Aggregation."
                     "We recommend clipping the model parameters."
                 )
+
             # Mypy: hasattr check on the protocol aggregator, model and steps.
             # Pass the `SecureShare` parameters to the model for clipping
-            self.protocol.algorithm.model.param_clipping = {  # type: ignore[attr-defined] # Reason: see above # noqa: B950
-                "prime_q": self.protocol.aggregator._secure_share.prime_q,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
-                "precision": self.protocol.aggregator._secure_share.precision,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
-                "num_workers": len(pod_identifiers),
-            }
-        result = asyncio.run(self.run_async(pod_identifiers, require_all_pods))
+            for algo in self.protocol.algorithms:
+                if hasattr(algo, "model"):
+                    algo.model.param_clipping = {  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                        "prime_q": self.protocol.aggregator._secure_share.prime_q,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                        "precision": self.protocol.aggregator._secure_share.precision,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                        "num_workers": len(pod_identifiers),
+                    }
+        result = asyncio.run(
+            self.run_async(
+                pod_identifiers, require_all_pods, project_id, run_on_new_data_only
+            )
+        )
         if model_out and result is not False:
             self._serialize(model_out)
         return result
