@@ -2,10 +2,8 @@
 import asyncio
 from contextlib import nullcontext
 from dataclasses import asdict
-import hashlib
 import os
 from pathlib import Path
-import sqlite3
 import threading
 from typing import (
     Any,
@@ -23,13 +21,13 @@ from typing import (
 )
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
-import numpy as np
 import pandas as pd
 from requests import HTTPError, RequestException
 import yaml
 
 from bitfount.config import BITFOUNT_STORAGE_PATH
-from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.base_source import BaseSource, MultiTableSource
+from bitfount.data.datasources.database_source import DatabaseSource
 from bitfount.data.exceptions import DataSourceError
 from bitfount.data.schema import BitfountSchema
 from bitfount.federated.aggregators.secure import _is_secure_share_task_request
@@ -51,6 +49,7 @@ from bitfount.federated.helper import (
 from bitfount.federated.keys_setup import RSAKeyPair, _get_pod_keys
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.monitoring import task_monitor_context
+from bitfount.federated.pod_db_utils import _add_data_to_pod_db
 from bitfount.federated.pod_response_message import _PodResponseMessage
 from bitfount.federated.pod_vitals import _PodVitals, _PodVitalsHandler
 from bitfount.federated.privacy.differential import DPPodConfig
@@ -180,6 +179,7 @@ class Pod:
         show_datapoints_with_results_in_db: Whether the original datapoints should be
             included in the results database. Defaults to True. This argument is ignored
             if pod_db argument is set to False.
+        project_id: The project id. Defaults to None.
         update_schema: Whether the schema needs to be re-generated even if provided.
             Defaults to False.
         secrets: Secrets for authenticating with Bitfount services.
@@ -292,7 +292,14 @@ class Pod:
         self.public_metadata = self._get_public_metadata(
             pod_details_config, self.schema
         )
-        self.pod_db = pod_db
+        if isinstance(self.datasource, DatabaseSource):
+            self.pod_db = False
+            logger.warning(
+                "Pod database not supported for DatabaseSource. "
+                "Starting pod without database."
+            )
+        else:
+            self.pod_db = pod_db
         if self.pod_db:
             self._update_pod_db()
         self.show_datapoints_with_results_in_db = (
@@ -417,97 +424,38 @@ class Pod:
 
         This is a static database on the pod with the datapoint hashes so we only
         compute them once. For each datapoint row in the datasource, a hash value
-        is computed. Then the data from the datasource, together with the hash value,
-        are written to the database.
+        is computed. Then the data from (each table of) the datasource,
+        together with the hash value,are written to the database.
 
-        :::note
 
-        Only works for single-table, non-iterable datasources.
+        :::caution
+
+        Does not work for multi-table `DatabaseSource`s as we cannot load the data into
+        memory.
 
         :::
-
-        Raises:
-            ValueError: If there are clashing column names in the datasource
-                and the pod database.
         """
-        if self.datasource.multi_table or self.datasource.iterable:
-            # TODO: [BIT-2486] - Look into supporting multi-table datasources.
-            logger.warning(
-                "Pod database only supported with single-table, "
-                "non-iterable datasources."
-            )
-            self.pod_db = False
+        if self.datasource.multi_table:
+            datasource = cast(MultiTableSource, self.datasource)
+            for table in datasource.table_names:
+                # if table name is given as an arg to get_data
+                # then it will always return a df, so we can cast
+                new_data = cast(
+                    pd.DataFrame, datasource.get_data(table_name=table)
+                ).copy()
+                _add_data_to_pod_db(
+                    pod_name=self.name,
+                    data=new_data,
+                    table_name=table,
+                )
         else:
-            con = sqlite3.connect(f"{self.name}.db")
-            cur = con.cursor()
-            cur.execute(
-                """CREATE TABLE IF NOT EXISTS "datasource" ('rowID' INTEGER PRIMARY KEY)"""  # noqa: B950
-            )
-            con.commit()
-
-            # get the existing datasource data
+            # This works regardless of whether or not the datasource is iterable
             self.datasource.load_data()
-            new_data = self.datasource.data
-            if "datapoint_hash" in new_data.columns:
-                raise ValueError(
-                    "`datapoint_hash` not supported as column name in the datasource."
-                )
-            # as placeholder column for the hash
-            new_data["datapoint_hash"] = np.nan
-            # change type to object so it will match the db type
-            new_data["datapoint_hash"] = new_data["datapoint_hash"].astype("object")
-
-            # sqlite transforms bool values to int, so we need to make sure that
-            # they are the same in the df so the hashes match
-            bool_cols = [col for col in new_data.columns if new_data[col].dtype == bool]
-            # replace bools by their int value, as it will be done by
-            # sqlite in the db anyway
-            new_data[bool_cols] *= 1
-            # read the db data for the datasource
-            existing_data = pd.read_sql('SELECT * FROM "datasource"', con)
-            existing_cols_without_index = set(existing_data.columns)
-            existing_cols_without_index.remove("rowID")
-            # check if df is empty or if columns all columns are the same,
-            # if not all the hashes will have to be recomputed
-            if (
-                not existing_data.empty
-                and set(new_data.columns) == existing_cols_without_index
-            ):
-                data = pd.merge(
-                    new_data.drop(columns=["datapoint_hash"]),
-                    existing_data.drop(columns=["datapoint_hash", "rowID"]),
-                    how="outer",
-                    indicator=True,
-                ).loc[lambda x: x["_merge"] != "both"]
-                data.drop(columns=["_merge"], inplace=True)
-                data.drop_duplicates()
-                hashed_list = []
-                for _, row in data.iterrows():
-                    hashed_list.append(
-                        hashlib.sha256(str(row).encode("utf-8")).hexdigest()
-                    )
-                data["datapoint_hash"] = hashed_list
-                data.to_sql("datasource", con=con, if_exists="append", index=False)
-            else:
-
-                # replace table if columns are mismatched
-                cur.execute("DROP TABLE 'datasource'")
-                cur.execute(
-                    """CREATE TABLE "datasource" ('rowID' INTEGER PRIMARY KEY)"""
-                )
-                for col in new_data.columns:
-                    cur.execute(
-                        f"ALTER TABLE 'datasource' ADD COLUMN '{col}' {new_data[col].dtype}"  # noqa: B950
-                    )
-
-                hashed_list = []
-                for _, row in new_data.iterrows():
-                    hashed_list.append(
-                        hashlib.sha256(str(row).encode("utf-8")).hexdigest()
-                    )
-                data = new_data
-                data["datapoint_hash"] = hashed_list
-                data.to_sql("datasource", con=con, if_exists="append", index=False)
+            _add_data_to_pod_db(
+                pod_name=self.name,
+                data=self.datasource.data.copy(),
+                table_name=self.name,
+            )
 
     def _get_default_pod_details_config(self) -> PodDetailsConfig:
         """Get default pod details config."""
@@ -671,9 +619,16 @@ class Pod:
         task_request_message: _TaskRequestMessage = _TaskRequestMessage.deserialize(
             message.body
         )
-        # TODO: [BIT-2486] There should be check to see if the
-        #  task request contains a project id, and allow
-        #  for pod_db only in that case.
+        try:
+            if task_request_message.project_id is None:
+                # if there is no project id we set pod_db to False
+                # old versions of bitfount don't include the project id
+                # in the task request message, so added check for missing attribute
+                self.pod_db = False
+        except AttributeError:
+            # this check is to ensure backward compatibility for task request messages
+            self.pod_db = False
+
         auth_type: IdentityVerificationMethod = check_identity_verification_method(
             task_request_message.auth_type
         )
@@ -701,6 +656,7 @@ class Pod:
         # Similarly, if we're not using secure aggregation we just create a normal
         # mailbox as inter-pod communication won't be needed.
         worker_mailbox: _WorkerMailbox
+
         if _is_secure_share_task_request(task_request) and not unapproved_workers:
             logger.debug("Creating mailbox with inter-pod support.")
 
@@ -753,8 +709,11 @@ class Pod:
         )
 
         logger.debug("Creating worker.")
-        # TODO: [BIT-2486] Add the project id to the worker to ensure
-        #  results are written to the correct project db.
+        if not hasattr(task_request_message, "project_id"):
+            task_request_message.project_id = None
+        if not hasattr(task_request_message, "run_on_new_data_only"):
+            task_request_message.run_on_new_data_only = False
+
         worker = _Worker(
             datasource=self.datasource,
             mailbox=worker_mailbox,
@@ -766,6 +725,8 @@ class Pod:
             serialized_protocol=task_request.serialized_protocol,
             pod_db=self.pod_db,
             show_datapoints_in_results_db=self.show_datapoints_with_results_in_db,
+            run_on_new_data_only=task_request_message.run_on_new_data_only,
+            project_id=task_request_message.project_id,
         )
 
         # If interacting with an older modeller version then task_id won't be supplied
@@ -901,6 +862,7 @@ class Pod:
         immediately when the previous iteration finished.
 
         Args:
+            stop_event: threading.Event to stop the loop
             interval: run interval in seconds
             func: function to call which returns a coroutine to await
         """
