@@ -24,9 +24,11 @@ from typing import (
     Mapping,
     Optional,
     Protocol,
+    Sequence,
     Type,
     TypeVar,
     Union,
+    cast,
 )
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
@@ -55,6 +57,7 @@ from bitfount.federated.transport.modeller_transport import _ModellerMailbox
 from bitfount.federated.transport.worker_transport import _WorkerMailbox
 from bitfount.federated.types import ProtocolType, SerializedProtocol
 from bitfount.hub.helper import _default_bitfounthub
+from bitfount.schemas.utils import bf_dump
 from bitfount.types import T_FIELDS_DICT, T_NESTED_FIELDS, _BaseSerializableObjectMixIn
 
 if TYPE_CHECKING:
@@ -73,24 +76,34 @@ class _BaseProtocol(Generic[MB], ABC):
     def __init__(
         self,
         *,
-        algorithm: Union[_BaseCompatibleModeller, _BaseCompatibleWorker],
+        algorithm: Union[
+            BaseCompatibleModellerAlgorithm, BaseCompatibleWorkerAlgorithm
+        ],
         mailbox: MB,
         **kwargs: Any,
     ):
-
         self.algorithm = algorithm
         self.mailbox = mailbox
 
         super().__init__(**kwargs)
 
+    @property
+    def algorithms(
+        self,
+    ) -> List[Union[BaseCompatibleModellerAlgorithm, BaseCompatibleWorkerAlgorithm]]:
+        """Returns the algorithms in the protocol."""
+        if isinstance(self.algorithm, Sequence):
+            return list(self.algorithm)
+        return [self.algorithm]
 
-class _BaseCompatibleModeller(Protocol):
-    """Protocol defining base modeller-side compatibility."""
+
+class BaseCompatibleModellerAlgorithm(Protocol):
+    """Protocol defining base modeller-side algorithm compatibility."""
 
     pass
 
 
-class _BaseModellerProtocol(_BaseProtocol[_ModellerMailbox], ABC):
+class BaseModellerProtocol(_BaseProtocol[_ModellerMailbox], ABC):
     """Modeller side of the protocol.
 
     Calls the modeller side of the algorithm.
@@ -99,7 +112,9 @@ class _BaseModellerProtocol(_BaseProtocol[_ModellerMailbox], ABC):
     def __init__(
         self,
         *,
-        algorithm: _BaseCompatibleModeller,
+        algorithm: Union[
+            BaseCompatibleModellerAlgorithm, Sequence[BaseCompatibleModellerAlgorithm]
+        ],
         mailbox: _ModellerMailbox,
         **kwargs: Any,
     ):
@@ -114,13 +129,13 @@ class _BaseModellerProtocol(_BaseProtocol[_ModellerMailbox], ABC):
         pass
 
 
-class _BaseCompatibleWorker(Protocol):
-    """Protocol defining base worker-side compatibility."""
+class BaseCompatibleWorkerAlgorithm(Protocol):
+    """Protocol defining base worker-side algorithm compatibility."""
 
     pass
 
 
-class _BaseWorkerProtocol(_BaseProtocol[_WorkerMailbox], ABC):
+class BaseWorkerProtocol(_BaseProtocol[_WorkerMailbox], ABC):
     """Worker side of the protocol.
 
     Calls the worker side of the algorithm.
@@ -129,7 +144,9 @@ class _BaseWorkerProtocol(_BaseProtocol[_WorkerMailbox], ABC):
     def __init__(
         self,
         *,
-        algorithm: _BaseCompatibleWorker,
+        algorithm: Union[
+            BaseCompatibleWorkerAlgorithm, Sequence[BaseCompatibleWorkerAlgorithm]
+        ],
         mailbox: _WorkerMailbox,
         **kwargs: Any,
     ):
@@ -143,18 +160,18 @@ class _BaseWorkerProtocol(_BaseProtocol[_WorkerMailbox], ABC):
         pod_vitals: Optional[_PodVitals] = None,
         pod_identifier: Optional[str] = None,
         **kwargs: Any,
-    ) -> None:
+    ) -> Any:
         """Runs the worker-side of the algorithm."""
         pass
 
 
 # The mutable underlying dict that holds the registry information
-_registry: Dict[str, Type[_BaseProtocolFactory]] = {}
+_registry: Dict[str, Type[BaseProtocolFactory]] = {}
 # The read-only version of the registry that is allowed to be imported
-registry: Mapping[str, Type[_BaseProtocolFactory]] = MappingProxyType(_registry)
+registry: Mapping[str, Type[BaseProtocolFactory]] = MappingProxyType(_registry)
 
 
-class _BaseCompatibleAlgoFactory(Protocol):
+class BaseCompatibleAlgoFactory(Protocol):
     """Protocol defining base algorithm factory compatibility."""
 
     class_name: str
@@ -162,7 +179,7 @@ class _BaseCompatibleAlgoFactory(Protocol):
     nested_fields: ClassVar[T_NESTED_FIELDS] = {}
 
 
-class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
+class BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
     """Base Protocol from which all other protocols must inherit."""
 
     fields_dict: ClassVar[T_FIELDS_DICT] = {}
@@ -171,13 +188,21 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
     def __init__(
         self,
         *,
-        algorithm: _BaseCompatibleAlgoFactory,
+        algorithm: Union[
+            BaseCompatibleAlgoFactory, Sequence[BaseCompatibleAlgoFactory]
+        ],
         **kwargs: Any,
     ) -> None:
-        self.class_name = ProtocolType[type(self).__name__].value
+        try:
+            self.class_name = ProtocolType[type(self).__name__].value
+        except KeyError:
+            # Check if the protocol is a plug-in
+            self.class_name = type(self).__name__
+
         super().__init__(**kwargs)
-        self._validate_algorithm(algorithm)
         self.algorithm = algorithm
+        for algo in self.algorithms:
+            self._validate_algorithm(algo)
 
     @classmethod
     def __init_subclass__(cls, **kwargs: Any):
@@ -185,9 +210,16 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
             logger.debug(f"Adding {cls.__name__}: {cls} to Protocol registry")
             _registry[cls.__name__] = cls
 
+    @property
+    def algorithms(self) -> List[BaseCompatibleAlgoFactory]:
+        """Returns the algorithms in the protocol."""
+        if isinstance(self.algorithm, Sequence):
+            return list(self.algorithm)
+        return [self.algorithm]
+
     @classmethod
     @abstractmethod
-    def _validate_algorithm(cls, algorithm: _BaseCompatibleAlgoFactory) -> None:
+    def _validate_algorithm(cls, algorithm: BaseCompatibleAlgoFactory) -> None:
         """Checks that `algorithm` is compatible with the protocol.
 
         Raises TypeError if `algorithm` is not compatible with the protocol.
@@ -197,21 +229,20 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
     @abstractmethod
     def modeller(
         self, mailbox: _ModellerMailbox, **kwargs: Any
-    ) -> _BaseModellerProtocol:
+    ) -> BaseModellerProtocol:
         """Creates an instance of the modeller-side for this protocol."""
         raise NotImplementedError
 
     @abstractmethod
     def worker(
         self, mailbox: _WorkerMailbox, hub: BitfountHub, **kwargs: Any
-    ) -> _BaseWorkerProtocol:
+    ) -> BaseWorkerProtocol:
         """Creates an instance of the worker-side for this protocol."""
         raise NotImplementedError
 
-    @abstractmethod
     def dump(self) -> SerializedProtocol:
-        """Serialize object, returns as JSON dict."""
-        raise NotImplementedError
+        """Returns the JSON-serializable representation of the protocol."""
+        return cast(SerializedProtocol, bf_dump(self))
 
     def run(
         self,
@@ -226,6 +257,7 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
         private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
         idp_url: Optional[str] = None,
         require_all_pods: bool = False,
+        run_on_new_data_only: bool = False,
     ) -> Optional[Any]:
         """Sets up a local Modeller instance and runs the protocol.
 
@@ -244,6 +276,8 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
             idp_url: The IDP URL.
             require_all_pods: If true raise PodResponseError if at least one pod
                 identifier specified rejects or fails to respond to a task request.
+            run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
         Returns:
             Results of the protocol.
@@ -256,29 +290,30 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
         """
         hub = _default_bitfounthub(hub=hub, username=username)
 
-        if isinstance(self.algorithm, _BaseModelAlgorithmFactory):
-            if (
-                len(pod_identifiers) > 1
-                and hasattr(self.algorithm.model.datastructure, "table")
-                and isinstance(self.algorithm.model.datastructure.table, str)
-            ):
-                raise ValueError(
-                    "You are attempting to train on multiple pods, and the"
-                    "provided the DataStructure table name is a string. "
-                    "Please make sure that the `table` argument to the "
-                    "`DataStructure` is a mapping of Pod names to table names. "
-                )
-            pod_identifiers = _check_and_update_pod_ids(pod_identifiers, hub)
-            datastructure_pod_identifiers = (
-                self.algorithm.model.datastructure.get_pod_identifiers()
-            )
-            if datastructure_pod_identifiers:
-                datastructure_pod_identifiers = _check_and_update_pod_ids(
-                    datastructure_pod_identifiers, hub
-                )
-                self.algorithm.model.datastructure._update_datastructure_with_hub_identifiers(
-                    datastructure_pod_identifiers
+        for algo in self.algorithms:
+            if isinstance(algo, _BaseModelAlgorithmFactory):
+                if (
+                    len(pod_identifiers) > 1
+                    and hasattr(algo.model.datastructure, "table")
+                    and isinstance(algo.model.datastructure.table, str)
+                ):
+                    raise ValueError(
+                        "You are attempting to train on multiple pods, and the"
+                        "provided the DataStructure table name is a string. "
+                        "Please make sure that the `table` argument to the "
+                        "`DataStructure` is a mapping of Pod names to table names. "
+                    )
+                pod_identifiers = _check_and_update_pod_ids(pod_identifiers, hub)
+                datastructure_pod_identifiers = (
+                    algo.model.datastructure.get_pod_identifiers()
                 )
+                if datastructure_pod_identifiers:
+                    datastructure_pod_identifiers = _check_and_update_pod_ids(
+                        datastructure_pod_identifiers, hub
+                    )
+                    algo.model.datastructure._update_datastructure_with_hub_identifiers(
+                        datastructure_pod_identifiers
+                    )
         if not session:
             session = hub.session
         if not idp_url:
@@ -300,8 +335,12 @@ class _BaseProtocolFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
         )
         name = type(self).__name__
 
-        logger.info(f"Starting {name} Job...")
+        logger.info(f"Starting {name} Task...")
 
-        result = modeller.run(pod_identifiers, require_all_pods=require_all_pods)
-        logger.info(f"Completed {name} Job...")
+        result = modeller.run(
+            pod_identifiers,
+            require_all_pods=require_all_pods,
+            run_on_new_data_only=run_on_new_data_only,
+        )
+        logger.info(f"Completed {name} Task.")
         return result
