@@ -28,8 +28,8 @@ from bitfount.federated.aggregators.base import (
     _BaseAggregatorFactory,
     _BaseModellerAggregator,
     _BaseWorkerAggregator,
+    registry as aggregators_registry,
 )
-from bitfount.federated.aggregators.base import registry as aggregators_registry
 from bitfount.federated.aggregators.secure import _InterPodAggregatorWorkerFactory
 from bitfount.federated.algorithms.model_algorithms.base import (
     registry as algorithms_registry,
@@ -41,10 +41,10 @@ from bitfount.federated.model_reference import BitfountModelReference
 from bitfount.federated.pod_vitals import _PodVitals
 from bitfount.federated.privacy.differential import DPPodConfig
 from bitfount.federated.protocols.base import (
-    _BaseCompatibleAlgoFactory,
-    _BaseModellerProtocol,
-    _BaseProtocolFactory,
-    _BaseWorkerProtocol,
+    BaseCompatibleAlgoFactory,
+    BaseModellerProtocol,
+    BaseProtocolFactory,
+    BaseWorkerProtocol,
 )
 from bitfount.federated.transport.modeller_transport import (
     _get_parameter_updates_from_workers,
@@ -59,9 +59,7 @@ from bitfount.federated.transport.worker_transport import (
     _send_training_metrics,
     _WorkerMailbox,
 )
-from bitfount.federated.types import SerializedProtocol
 from bitfount.hub.api import BitfountHub
-from bitfount.schemas.utils import bf_dump
 from bitfount.types import (
     T_DTYPE,
     T_FIELDS_DICT,
@@ -148,7 +146,7 @@ class _FederatedAveragingCompatibleModeller(
 
 
 @runtime_checkable
-class _FederatedAveragingCompatibleAlgoFactory(_BaseCompatibleAlgoFactory, Protocol):
+class _FederatedAveragingCompatibleAlgoFactory(BaseCompatibleAlgoFactory, Protocol):
     """Defines algorithm factory compatibility."""
 
     model: _DistributedModelTypeOrReference
@@ -229,7 +227,7 @@ class _BaseFederatedAveragingMixIn:
         return num_iterations // num_iterations_between_updates
 
 
-class _ModellerSide(_BaseModellerProtocol, _BaseFederatedAveragingMixIn):
+class _ModellerSide(BaseModellerProtocol, _BaseFederatedAveragingMixIn):
     """Modeller side of the FederatedAveraging protocol."""
 
     aggregator: _BaseModellerAggregator
@@ -288,8 +286,7 @@ class _ModellerSide(_BaseModellerProtocol, _BaseFederatedAveragingMixIn):
         self.algorithm.initialise(task_id=self.mailbox._task_id)
         self.perform_iterations_checks()
         initial_parameters = self.algorithm.run(update=None)
-        logger.federated_info("Sending model parameters to Pods")
-        logger.info("Sending model parameters to Pods")
+        logger.federated_info("Sending model parameters to pods...")
         await self._send_parameters(initial_parameters)
         logger.info("Sent model parameters to Pods")
         num_federated_iterations = self.get_num_federated_iterations()
@@ -359,7 +356,7 @@ class _ModellerSide(_BaseModellerProtocol, _BaseFederatedAveragingMixIn):
         return modeller_results
 
 
-class _WorkerSide(_BaseWorkerProtocol, _BaseFederatedAveragingMixIn):
+class _WorkerSide(BaseWorkerProtocol, _BaseFederatedAveragingMixIn):
     """Worker side of the FederatedAveraging protocol."""
 
     aggregator: _BaseWorkerAggregator
@@ -415,7 +412,7 @@ class _WorkerSide(_BaseWorkerProtocol, _BaseFederatedAveragingMixIn):
         pod_vitals: Optional[_PodVitals] = None,
         pod_identifier: Optional[str] = None,
         **kwargs: Any,
-    ) -> None:
+    ) -> Any:
         """Receives parameters and sends updates in a loop."""
         self.algorithm.initialise(datasource=datasource, pod_dp=pod_dp)
         self.perform_iterations_checks()
@@ -466,7 +463,7 @@ class _WorkerSide(_BaseWorkerProtocol, _BaseFederatedAveragingMixIn):
 
 
 @delegates()
-class FederatedAveraging(_BaseProtocolFactory):
+class FederatedAveraging(BaseProtocolFactory):
     """Original Federated Averaging algorithm by McMahan et al. (2017).
 
     This protocol performs a predetermined number of epochs or steps of training on
@@ -570,7 +567,7 @@ class FederatedAveraging(_BaseProtocolFactory):
                     self.steps_between_parameter_updates = 1
 
     @classmethod
-    def _validate_algorithm(cls, algorithm: _BaseCompatibleAlgoFactory) -> None:
+    def _validate_algorithm(cls, algorithm: BaseCompatibleAlgoFactory) -> None:
         """Checks that `algorithm` is compatible with the protocol.
 
         Raises:
@@ -635,7 +632,3 @@ class FederatedAveraging(_BaseProtocolFactory):
             mailbox=mailbox,
             **kwargs,
         )
-
-    def dump(self) -> SerializedProtocol:
-        """Returns the JSON-serializable representation of the protocol."""
-        return cast(SerializedProtocol, bf_dump(self))
