@@ -23,8 +23,8 @@ from bitfount.federated.aggregators.base import (
     _BaseAggregatorFactory,
     _BaseModellerAggregator,
     _BaseWorkerAggregator,
+    registry as aggregators_registry,
 )
-from bitfount.federated.aggregators.base import registry as aggregators_registry
 from bitfount.federated.aggregators.secure import _InterPodAggregatorWorkerFactory
 from bitfount.federated.algorithms.base import registry as algorithms_registry
 from bitfount.federated.algorithms.model_algorithms.base import (
