@@ -3,8 +3,7 @@ import asyncio
 from asyncio import FIRST_COMPLETED, AbstractEventLoop, Task
 from asyncio.futures import Future as AsyncFuture
 from collections import defaultdict
-from concurrent.futures import Future as ConcurrentFuture
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import Future as ConcurrentFuture, ThreadPoolExecutor
 from contextlib import contextmanager
 import inspect
 import threading
