@@ -5,8 +5,7 @@ import asyncio
 from asyncio.futures import Future as AsyncFuture
 import atexit
 from collections.abc import Iterable
-from concurrent.futures import Future as ConcurrentFuture
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import Future as ConcurrentFuture, ThreadPoolExecutor
 from dataclasses import dataclass
 from functools import wraps
 import inspect
