@@ -39,8 +39,8 @@ import msgpack
 import numpy as np
 import pandas as pd
 import psutil
-import pyarrow as pa
 from pyarrow import feather
+import pyarrow as pa
 from requests.exceptions import RequestException
 
 from bitfount.config import get_gpu_metadata
@@ -49,18 +49,15 @@ from bitfount.federated.exceptions import PodConnectFailedError
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.transport.exceptions import BitfountMessageServiceError
 from bitfount.federated.transport.protos.messages_pb2 import (
+    Acknowledgement,
     BitfountMessage as GrpcBitfountMessage,
-)
-from bitfount.federated.transport.protos.messages_pb2 import (
+    BitfountTask,
+    BitfountTasks,
     CommunicationDetails as GrpcCommunicationDetails,
-)
-from bitfount.federated.transport.protos.messages_pb2 import (
     LargeStorageRequest,
     PodData,
     SuccessResponse,
 )
-from bitfount.federated.transport.protos.messages_pb2 import Acknowledgement
-from bitfount.federated.transport.protos.messages_pb2 import BitfountTask, BitfountTasks
 from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub
 from bitfount.federated.transport.types import CommunicationDetails
 from bitfount.federated.transport.utils import _auto_retry_grpc
