@@ -1,17 +1,17 @@
 """Workers for handling task running on pods."""
 from __future__ import annotations
 
+import copy
 import hashlib
 import json
 import sqlite3
 from sqlite3 import Connection
 from typing import Any, List, Optional, Sequence, cast
 
-import numpy as np
 import pandas as pd
 import sqlvalidator
 
-from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.base_source import BaseSource, FileSystemIterableSource
 from bitfount.data.datasources.database_source import DatabaseSource
 from bitfount.data.datastructure import DataStructure
 from bitfount.data.exceptions import DataStructureError
@@ -22,6 +22,10 @@ from bitfount.federated.algorithms.model_algorithms.base import (
 from bitfount.federated.authorisation_checkers import _AuthorisationChecker
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.monitoring.monitor import task_config_update
+from bitfount.federated.pod_db_utils import (
+    _map_task_to_hash_add_to_db,
+    _save_results_to_db,
+)
 from bitfount.federated.pod_vitals import _PodVitals
 from bitfount.federated.privacy.differential import DPPodConfig
 from bitfount.federated.protocols.base import (
@@ -59,6 +63,12 @@ class _Worker:
             from the Modeller.
         pod_vitals: PodVitals object.
         pod_dp: DPPodConfig object.
+        pod_db: Whether the pod has a databse associated with it. Defaults to False.
+        show_datapoints_in_results_db: Whether the datasource records are shown
+            in the results database. Defaults to True.
+        project_id: The project id. Defaults to None.
+        run_on_new_data_only: Whether to run on the whole dataset or only on
+            new data. Defaults to False.
     """
 
     def __init__(
@@ -73,6 +83,8 @@ class _Worker:
         pod_dp: Optional[DPPodConfig] = None,
         pod_db: bool = False,
         show_datapoints_in_results_db: bool = True,
+        project_id: Optional[str] = None,
+        run_on_new_data_only: bool = False,
         **_kwargs: Any,
     ):
         self.datasource = datasource
@@ -83,8 +95,17 @@ class _Worker:
         self.serialized_protocol = serialized_protocol
         self.pod_vitals = pod_vitals
         self._pod_dp = pod_dp
-        self._pod_db = pod_db
-        self._show_datapoints_in_results_db = show_datapoints_in_results_db
+        self.project_id = project_id
+        self._pod_db = (
+            pod_db
+            if project_id is not None
+            and not isinstance(self.datasource, DatabaseSource)
+            else False
+        )
+        self._show_datapoints_in_results_db = (
+            show_datapoints_in_results_db if self._pod_db else False
+        )
+        self.run_on_new_data_only = run_on_new_data_only if self._pod_db else False
         # Compute task hash on ordered json dictionary
         self._task_hash = (
             hashlib.sha256(
@@ -94,12 +115,24 @@ class _Worker:
             else None
         )
 
+    def _update_task_config(self) -> None:
+        """Send task config update to monitor service."""
+        # remove schema from task_config to limit request body size
+        task_config = copy.deepcopy(self.serialized_protocol)
+        for algorithm in task_config["algorithm"]:
+            if not isinstance(algorithm, str):
+                if "model" in algorithm.keys():
+                    model = algorithm["model"]
+                    if "schema" in model:
+                        del model["schema"]
+        task_config_update(dict(task_config))
+
     async def run(self) -> None:
         """Calls relevant training procedure and sends back weights/results."""
         # Send task to Monitor service. This is done regardless of whether or not
         # the task is accepted. This method is being run in a task monitor context
         # manager so no need to set the task monitor prior to sending.
-        task_config_update(dict(self.serialized_protocol))
+        self._update_task_config()
 
         # Check authorisation with access manager
         authorisation_errors = await self.authorisation.check_authorisation()
@@ -147,23 +180,27 @@ class _Worker:
             algorithm_ = [algorithm_]
 
         algorithm_ = cast(List[BaseCompatibleAlgoFactory], algorithm_)
+        if self._pod_db:
+            con = sqlite3.connect(f"{self.project_id}.sqlite")
+            cur = con.cursor()
+            cur.execute(
+                f"""CREATE TABLE IF NOT EXISTS "{self._task_hash}" (rowID INTEGER PRIMARY KEY, 'datapoint_hash' VARCHAR, 'results' VARCHAR)"""  # noqa: B950
+            )
         for algo_ in algorithm_:
             if isinstance(algo_, _BaseModelAlgorithmFactory):
                 datastructure = algo_.model.datastructure
 
             if not isinstance(algo_, _DataLessAlgorithm):
+                # TODO: [BIT-2709] This should not be run once per algorithm, but once
+                # per protocol
                 if self._pod_db:
-                    # For now, use pod name.
-                    # TODO: [BIT-2486] Update this to reflect the project id.
-                    con = sqlite3.connect(
-                        f'{self.mailbox.pod_identifier.split("/")[-1]}.db'
+                    table = self._load_data_for_worker(
+                        datastructure=datastructure, con=con
                     )
-                    cur = con.cursor()
-                    cur.execute(
-                        f"""CREATE TABLE IF NOT EXISTS "{self._task_hash}" (rowID INTEGER PRIMARY KEY, 'datapoint_hash' VARCHAR, 'results' VARCHAR)"""  # noqa: B950
+                    # task_hash is set if pod_db is true, so it's safe to cast
+                    _map_task_to_hash_add_to_db(
+                        self.serialized_protocol, cast(str, self._task_hash), con
                     )
-                    self._load_data_for_worker(datastructure=datastructure, con=con)
-                    self._map_task_to_hash_add_to_db(con)
                 else:
                     # We execute the query directly on the db connection,
                     # or load the data at runtime for a csv.
@@ -183,8 +220,7 @@ class _Worker:
         for worker_algo in worker_protocol.algorithms:
             if isinstance(worker_algo, _BaseModelAlgorithm):
                 worker_algo.model.set_pod_identifier(self.pod_identifier)
-        # TODO: [BIT-2486] check if projectid is given and only pass
-        #  connection in that case
+
         results = await worker_protocol.run(
             datasource=self.datasource,
             pod_dp=self._pod_dp,
@@ -192,17 +228,32 @@ class _Worker:
             pod_identifier=self.mailbox.pod_identifier,
         )
         if self._pod_db:
+            # pod_db is always false for DatabaseSource,
+            # which is the only datasource that accepts sqlquery
+            # instead of table name, so we can cast
+            # if pod_db is true, task_hash is a str,
+            # so it's safe to cast
             if isinstance(results, list):
-                self._save_results_to_db(results, con)
-                con.close()
+                _save_results_to_db(
+                    results=results,
+                    pod_identifier=self.pod_identifier,
+                    task_hash=cast(str, self._task_hash),
+                    table=cast(str, table),
+                    datasource=self.datasource,
+                    show_datapoints_in_results_db=self._show_datapoints_in_results_db,
+                    run_on_new_data_only=self.run_on_new_data_only,
+                    con=con,
+                )
             else:
-                logger.info(
-                    "Results cannot be saved to pod database."
+                logger.warning(
+                    "Results cannot be saved to pod database. "
                     "Results can be only saved to database if "
                     "they are returned from the algorithm as a list, "
                     f"whereas the chosen protocol returns {type(results)}."
                 )
-
+            con.close()
+        if isinstance(self.datasource, FileSystemIterableSource):
+            self.datasource.selected_file_names = []
         logger.info("Task complete.")
         self.mailbox.delete_all_handlers(_BitfountMessageType.LOG_MESSAGE)
 
@@ -210,12 +261,11 @@ class _Worker:
         self,
         datastructure: Optional[DataStructure] = None,
         con: Optional[Connection] = None,
-    ) -> None:
-        """Load the data for the worker."""
+    ) -> Optional[str]:
+        """Load the data for the worker and returns table_name."""
         sql_query: Optional[str] = None
         table: Optional[str] = None
         kwargs = {}
-
         if datastructure:
             if datastructure.table:
                 if isinstance(datastructure.table, dict):
@@ -227,6 +277,13 @@ class _Worker:
                         )
                     kwargs["table_name"] = table
                 elif isinstance(datastructure.table, str):
+                    table = datastructure.table
+                    if not table == self.pod_identifier.split("/")[1]:
+                        raise DataStructureError(
+                            f"Table definition not found for {self.pod_identifier}. "
+                            f"Table definitions provided in this DataStructure: "
+                            f"{str(datastructure.table)}"
+                        )
                     kwargs["table_name"] = datastructure.table
             elif datastructure.query:
                 if isinstance(datastructure.query, dict):
@@ -248,110 +305,44 @@ class _Worker:
                     kwargs["sql_query"] = sql_query
         # This call loads the data for a multi-table BaseSource as specified by the
         # Modeller/DataStructure.
-        self.datasource.load_data(**kwargs)
-        if self._pod_db:
-            self.load_new_records_only_for_task(cast(Connection, con))
 
-    def load_new_records_only_for_task(self, con: Connection) -> None:
+        self.datasource.load_data(**kwargs)
+        if self._pod_db and self.run_on_new_data_only:
+            # pod database is incompatible with DatabaseSource,
+            # which is the only datasource that supports
+            # datastructure queries, so it's safe to cast
+            # for the table name
+            self.load_new_records_only_for_task(
+                cast(Connection, con), table=cast(str, table)
+            )
+        return table
+
+    def load_new_records_only_for_task(self, con: Connection, table: str) -> None:
         # Ignoring the security warning because the sql query is trusted and
         # the task_hash is calculated at __init__.
+        """Loads only records that the task has not seen before."""
+        logger.debug("Loading new records only for task.")
         task_data = pd.read_sql(
             f'SELECT "datapoint_hash" FROM "{self._task_hash}"', con  # nosec
         )
         # check hash in from static datasource table -
-        # - needs new sql connection(TODO: [BIT-2486])
-        data = pd.read_sql('SELECT * FROM "datasource"', con)
-        # set datasource_data for specific task to only run on new records.
-        new_records = data[~data["datapoint_hash"].isin(task_data["datapoint_hash"])]
-        self.datasource._ignore_cols.append("datapoint_hash")
-        self.datasource._data = new_records
-
-    def _map_task_to_hash_add_to_db(self, con: Connection) -> None:
-        algorithm_ = self.serialized_protocol["algorithm"]
-        if not isinstance(algorithm_, Sequence):
-            algorithm_ = [algorithm_]
-        for algorithm in algorithm_:
-            if "model" in algorithm:
-                algorithm["model"].pop("schema")
-        cur = con.cursor()
-        cur.execute(
-            """CREATE TABLE IF NOT EXISTS "task_definitions" ('index' INTEGER  PRIMARY KEY AUTOINCREMENT  NOT NULL, 'taskhash' TEXT,'protocol' TEXT,'algorithm' TEXT)"""  # noqa: B950
-        )
-        data = pd.read_sql("SELECT * FROM 'task_definitions' ", con)
-        if self._task_hash not in data["taskhash"].unique():
-            cur.execute(
-                """INSERT INTO "task_definitions" ('taskhash',  'protocol', 'algorithm' ) VALUES (?,?,?);""",  # noqa: B950
-                (
-                    self._task_hash,
-                    self.serialized_protocol["class_name"],
-                    str(algorithm_),
-                ),
-            )
-        con.commit()
-
-    def _save_results_to_db(self, results: List[np.ndarray], con: Connection) -> None:
-        # TODO: [BIT-2486] change this to read from static pod
-        #  database instead of the one with the project
-        pod_data = pd.read_sql('SELECT * FROM "datasource"', con)
-        # We only care about the test data since we don't log
-        # anything in the database for validation or training data
-        run_data = self.datasource.data.iloc[self.datasource._test_idxs]
-        # convert results to string
-        results_as_str = [str(item) for item in results]
-        run_data["results"] = results_as_str
-        columns = list(pod_data.columns)
-        columns.remove("rowID")
-        columns.remove("datapoint_hash")
-        # get the datapoint hashes from the pod db
-        data_w_hash = pd.merge(
-            pod_data,
-            run_data,
-            how="outer",
-            left_on=columns,
-            right_on=columns,
-            indicator=True,
-        ).loc[lambda x: x["_merge"] == "both"]
-        # drop the indicator and index columns
-        data_w_hash.drop("_merge", inplace=True, axis=1)
-        if "rowID" in data_w_hash.columns:
-            data_w_hash.drop("rowID", inplace=True, axis=1)
-        data_w_hash.drop_duplicates(inplace=True, keep="last")
-        cur = con.cursor()
+        pod_con = sqlite3.connect(f"{self.pod_identifier.split('/')[1]}.sqlite")
         # Ignoring the security warning because the sql query is trusted and
-        # the task_hash is calculated at __init__.
-        task_data = pd.read_sql(
-            f'SELECT "datapoint_hash" FROM "{self._task_hash}"', con  # nosec
-        )
-        # If this is the first time the task is run, it will not
-        # have all the columns, so we need to make sure they are
-        # added. Otherwise, we don't need to worry about the columns
-        # as any alterations to them will be classified as a new task
-        if task_data.shape[0] == 0 and self._show_datapoints_in_results_db:
-            for col in columns:
-                cur.execute(
-                    f"ALTER TABLE '{self._task_hash}' ADD COLUMN '{col}' {data_w_hash[col].dtype}"  # noqa: B950
-                )
+        # the table is checked that it matches the datasource tables.
+        data = pd.read_sql(f'SELECT * FROM "{table}"', pod_con)  # nosec
+        pod_con.close()
 
-        # do merge and get new datapoints only
-        new_task_datapoint = pd.merge(
-            data_w_hash,
-            task_data,
-            how="left",
-            indicator=True,
-        ).loc[lambda x: x["_merge"] == "left_only"]
-        if "rowId" in new_task_datapoint.columns:
-            new_task_datapoint.drop("rowID", inplace=True, axis=1)
-        # drop the indicator and index columns
-        new_task_datapoint.drop("_merge", inplace=True, axis=1)
-        logger.info(
-            f"The task was run on {new_task_datapoint.shape[0]} "
-            f"records from the datasource."
-        )
-        if self._show_datapoints_in_results_db:
-            new_task_datapoint.to_sql(
-                f"{self._task_hash}", con=con, if_exists="append", index=False
+        # set datasource_data for specific task to only run on new records.
+        new_records = data[
+            ~data["datapoint_hash"].isin(task_data["datapoint_hash"])
+        ].drop(columns=["rowID"])
+        if (
+            isinstance(self.datasource, FileSystemIterableSource)
+            and self.datasource.iterable
+        ):
+            self.datasource.selected_file_names = list(
+                new_records["_original_filename"]
             )
         else:
-            new_task_datapoint[["datapoint_hash", "results"]].to_sql(
-                f"{self._task_hash}", con=con, if_exists="append", index=False
-            )
+            self.datasource._ignore_cols.append("datapoint_hash")
+            self.datasource._data = new_records
