@@ -1,21 +1,35 @@
 """Typed wrappers around the Bitfount REST apis."""
 from __future__ import annotations
 
+import base64
 from dataclasses import dataclass
 import hashlib
 import logging
 from pathlib import Path
-from typing import Any, Final, List, Mapping, Optional, Tuple, Type, cast
+import time
+from typing import (
+    Any,
+    Callable,
+    Final,
+    List,
+    Mapping,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 import warnings
 
 from cryptography.exceptions import InvalidSignature
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.asymmetric import padding
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
-from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
 import requests
 from requests import HTTPError, RequestException, Response
 from requests.exceptions import InvalidJSONError
+from typing_extensions import ParamSpec
 
 from bitfount.config import (
     _DEVELOPMENT_ENVIRONMENT,
@@ -39,7 +53,9 @@ from bitfount.hub.types import (
     PRODUCTION_AM_URL,
     PRODUCTION_HUB_URL,
     _AccessManagerKeyResponseJSON,
+    _ActivePublicKey,
     _AMAccessCheckResponseJSON,
+    _CreatedResourceResponseJSON,
     _ModelDetailsResponseJSON,
     _ModelUploadResponseJSON,
     _MonitorPostJSON,
@@ -48,6 +64,8 @@ from bitfount.hub.types import (
     _OIDCAccessCheckPostJSON,
     _PodDetailsResponseJSON,
     _PodRegistrationResponseJSON,
+    _PublicKeyJSON,
+    _RegisterUserPublicKeyPOSTJSON,
     _SAMLAdditionalInfoPOSTJSON,
     _SAMLChallengeResponseJSON,
     _SignatureBasedAccessCheckPostJSON,
@@ -92,6 +110,13 @@ _MAX_CUSTOM_MODEL_SIZE_BYTES: Final[int] = 3 * 1024 * 1024  # 3 Megabytes
 _MAX_CUSTOM_MODEL_SIZE_MEGABYTES: Final[int] = _get_mb_from_bytes(
     _MAX_CUSTOM_MODEL_SIZE_BYTES
 ).whole
+_MAX_WEIGHTS_SIZE_BYTES: Final[int] = 500 * 1024 * 1024  # 3 Megabytes
+_MAX_WEIGHTS_SIZE_MEGABYTES: Final[int] = _get_mb_from_bytes(
+    _MAX_CUSTOM_MODEL_SIZE_BYTES
+).whole
+
+_P = ParamSpec("_P")
+_R = TypeVar("_R")
 
 
 @dataclass
@@ -110,6 +135,10 @@ def hash_file_contents(file_path: Path) -> str:
     with open(file_path, "rb") as f:
         # Read and update hash string value in blocks of 4K
         for byte_block in iter(lambda: f.read(4096), b""):
+            # This is necessary to ensure cross-platform compatibility between windows
+            # and unix platforms which use "\r\n" and "\n" for line endings
+            # respectively.
+            byte_block = byte_block.replace(b"\r", b"")
             sha256_hash.update(byte_block)
 
     return sha256_hash.hexdigest()
@@ -238,7 +267,7 @@ class BitfountHub:
         if not self.session.authenticated:
             self.session.authenticate()
         self.url = url.rstrip("/")  # to ensure no trailing slash
-        self.user_storage_path = self.session.user_storage_path
+        self.user_storage_path = self.session.authentication_handler.user_storage_path
         self.username = self.session.username
 
     @staticmethod
@@ -384,13 +413,14 @@ class BitfountHub:
 
         return pod_key
 
-    def send_model(self, model_code_path: Path) -> None:
+    def send_model(self, model_code_path: Path, private_model: bool = False) -> None:
         """Sends the provided `model_code` to Hub, associated to session username.
 
         The name of the model is taken from the name of the file.
 
         Args:
             model_code_path: The path to the file containing the model code.
+            private_model: Whether the model is private or publically accessible.
 
         Raises:
             ModelUploadError: If the model upload fails. Specific subclasses of
@@ -426,6 +456,7 @@ class BitfountHub:
                     "modelName": model_name,
                     "modelSize": model_size,
                     "modelHash": model_hash,
+                    "privateModel": private_model,
                 },
             )
             response.raise_for_status()
@@ -474,6 +505,82 @@ class BitfountHub:
         logger.info("Model successfully uploaded to object storage.")
         return None
 
+    def send_weights(
+        self, model_name: str, model_version: int, pretrained_file_path: Path
+    ) -> None:
+        """Sends the provided `model_weights` to Hub.
+
+        Save trained model weights associated to session username,
+        model_name and model_version.
+
+        Args:
+            model_name: The name of the model to associate the weights with.
+            model_version: The version of the model to associate the weights with.
+            pretrained_file_path: The path to the pretrained model file.
+
+        Raises:
+            ModelUploadError: If the model upload fails. Specific subclasses of
+            ModelUploadError are used to indicate different failure conditions.
+        """
+        weights_text: bytes = pretrained_file_path.read_bytes()
+        weights_size = len(weights_text)
+        if weights_size > _MAX_WEIGHTS_SIZE_BYTES:
+            raise ModelTooLargeError(
+                f"Model weights are too large to upload: "
+                f"expected max {_MAX_WEIGHTS_SIZE_MEGABYTES} megabytes, "
+                f"got {_get_mb_from_bytes(weights_size).fractional} megabytes."
+            )
+
+        # Register model weight details with hub
+        try:
+            logger.info(
+                f"Uploading model weights to '{model_name}:{model_version}' to Hub"
+            )
+            response = self.session.put(
+                f"{self.url}/api/models",
+                json={
+                    "modelName": model_name,
+                    "modelVersion": model_version,
+                    "weightSize": weights_size,
+                },
+            )
+            response.raise_for_status()
+        except RequestException as err:
+            try:
+                # noinspection PyUnboundLocalVariable
+                err_response: Optional[Response] = response
+            except NameError:
+                # response variable wasn't created yet
+                err_response = None
+            self._handle_hub_request_exception(err, err_response)
+            raise ModelUploadError(
+                "Request exception occurred when uploading model weights to hub."
+            ) from err
+
+        # Extract response
+        response_json: _ModelUploadResponseJSON = _extract_response_json(response)
+        logger.info(
+            "Model weights successfully registered to Bitfount Hub, "
+            "uploading to object storage."
+        )
+
+        # Extract URL to upload model to and upload it
+        upload_url: _S3PresignedPOSTURL = response_json["uploadUrl"]
+        upload_fields: _S3PresignedPOSTFields = response_json["uploadFields"]
+        try:
+            _upload_file_to_s3(
+                upload_url,
+                upload_fields,
+                file_contents=weights_text,
+                file_name=pretrained_file_path.name,
+            )
+        except RequestException as re:
+            logger.error(f"Failed to upload model weights to S3: {re}")
+            raise ModelUploadError("Failed to upload model weights to S3") from re
+
+        logger.info("Model weights successfully uploaded to object storage.")
+        return None
+
     @staticmethod
     def _verify_bitfount_model_format(path_to_model: Path) -> Type[BitfountModel]:
         """Verifies that the model file is correctly formatted.
@@ -520,19 +627,10 @@ class BitfountHub:
 
         return model[1]
 
-    def get_model(
+    def _get_model_response(
         self, username: str, model_name: str, model_version: Optional[int] = None
-    ) -> Optional[Type[BitfountModel]]:
-        """Gets model code corresponding to `model_name` from user `username`.
-
-        Args:
-            username: The model's owner.
-            model_name: The name of the model.
-            model_version: (Optional) The version of the model. Defaults to latest.
-
-        Returns:
-            The loaded Bitfount model class or None if unable to retrieve it.
-        """
+    ) -> Optional[_ModelDetailsResponseJSON]:
+        """Gets the response for the model from the hub."""
         # Retrieve model details from API endpoint
         try:
             params = {
@@ -558,12 +656,87 @@ class BitfountHub:
 
         # Check response contains model details
         response_json: _ModelDetailsResponseJSON = _extract_response_json(response)
-        if len(response_json) == 0:  # i.e. empty JSON response
+
+        return response_json
+
+    def _response_retry(
+        self, request: Callable[_P, Optional[_R]], err_msg: str
+    ) -> Callable[_P, Optional[_R]]:
+        """Retry decorator for Hub requests."""
+
+        def retried_request(*args: _P.args, **kwargs: _P.kwargs) -> Optional[_R]:
+            attempts = 5
+            response_json = request(*args, **kwargs)
+            while response_json is None and attempts > 0:
+                time.sleep(5)  # give an extra 5 seconds in between retries
+                response_json = request(*args, **kwargs)
+                attempts -= 1
+            if not response_json:  # i.e. empty JSON response
+                logger.warning(err_msg)
+                return None
+
+            return response_json
+
+        return retried_request
+
+    def get_weights(
+        self, username: str, model_name: str, model_version: int
+    ) -> Optional[bytes]:
+        """Gets weights byte stream corresponding to `model_name` from user `username`.
+
+        Args:
+            username: The model's owner.
+            model_name: The name of the model.
+            model_version: (Optional) The version of the model. Defaults to latest.
+
+        Returns:
+            The loaded weights as a byte stream or None if one has not been uploaded.
+        """
+        retrying_get_model_response = self._response_retry(
+            self._get_model_response,
+            f"No models registered by the name of {model_name} from user {username}",
+        )
+        response_json = retrying_get_model_response(
+            username=username, model_name=model_name, model_version=model_version
+        )
+        if not response_json:
+            return None
+
+        if weights_url := response_json.get("weightsDownloadUrl"):
+            weights_bytes = _download_file_from_s3(weights_url)
+            return cast(bytes, weights_bytes)
+        else:
             logger.warning(
-                f"No models registered by the name of {model_name} from user {username}"
+                "No weight file associated with "
+                f"{username}/{model_name}:{model_version}"
             )
             return None
 
+    def get_model(
+        self, username: str, model_name: str, model_version: Optional[int] = None
+    ) -> Optional[Type[BitfountModel]]:
+        """Gets model code corresponding to `model_name` from user `username`.
+
+        Args:
+            username: The model's owner.
+            model_name: The name of the model.
+            model_version: (Optional) The version of the model. Defaults to latest.
+
+        Returns:
+            The loaded Bitfount model class or None if unable to retrieve it.
+        """
+        retrying_get_model_response = self._response_retry(
+            self._get_model_response,
+            f"No models registered by the name of {model_name} from user {username}",
+        )
+        response_json: Optional[
+            _ModelDetailsResponseJSON
+        ] = retrying_get_model_response(
+            username=username, model_name=model_name, model_version=model_version
+        )
+        if not response_json:
+            return None
+
         # Download model code from S3
         try:
             model_url: _S3PresignedURL = response_json["modelDownloadUrl"]
@@ -580,7 +753,7 @@ class BitfountHub:
         model_code: str = cast(str, _download_file_from_s3(model_url, encoding="utf-8"))
 
         model_file = self._write_model_code_to_file(
-            model_code, self.user_storage_path, model_name
+            model_code, self.user_storage_path / "models", model_name, ".py"
         )
 
         file_hash = hash_file_contents(model_file)
@@ -596,24 +769,26 @@ class BitfountHub:
             # Remove the model file as model is incompatible
             model_file.unlink()
             return None
+
         return model_class
 
     @staticmethod
     def _write_model_code_to_file(
-        model_code: str, user_storage_path: Path, model_name: str
+        contents: Union[bytes, str], storage_path: Path, model_name: str, suffix: str
     ) -> Path:
         """Saves downloaded model code to a Python file.
 
-        Creates a module from `model_code` and saves it in a directory named after
-        `username` under user_storage_path/models
+        Creates a module from `contents` and saves it in a directory named after
+        `username` under storage_path
         """
-        model_store = user_storage_path / "models"
-        model_store.mkdir(parents=True, exist_ok=True)
-        model_file = Path(model_store / model_name).with_suffix(".py")
-        model_file.touch()
-        model_file.write_text(model_code)
-
-        return model_file
+        storage_path.mkdir(parents=True, exist_ok=True)
+        file = Path(storage_path / model_name).with_suffix(suffix)
+        file.touch()
+        if isinstance(contents, bytes):
+            file.write_bytes(contents)
+        else:
+            file.write_text(contents)
+        return file
 
     def register_pod(
         self,
@@ -646,11 +821,11 @@ class BitfountHub:
         json_post = {
             "name": public_metadata.name,
             "podDisplayName": public_metadata.display_name,
-            "podPublicKey": pod_public_key.public_bytes(
-                Encoding.OpenSSH, PublicFormat.OpenSSH
+            "podPublicKey": _RSAEncryption.serialize_public_key(
+                pod_public_key, form="SSH"
             ).decode("utf-8"),
-            "accessManagerPublicKey": access_manager_key.public_bytes(
-                Encoding.OpenSSH, PublicFormat.OpenSSH
+            "accessManagerPublicKey": _RSAEncryption.serialize_public_key(
+                access_manager_key, form="SSH"
             ).decode("utf-8"),
             "description": public_metadata.description,
             "schemaSize": schema_size,
@@ -705,8 +880,8 @@ class BitfountHub:
         """
         json_patch = {
             "name": pod_name,
-            "podPublicKey": pod_public_key.public_bytes(
-                Encoding.OpenSSH, PublicFormat.OpenSSH
+            "podPublicKey": _RSAEncryption.serialize_public_key(
+                pod_public_key, form="SSH"
             ).decode("utf-8"),
         }
 
@@ -780,7 +955,7 @@ class BitfountHub:
         """
         try:
             response = self.session.post(
-                f"{self.url}/api/ingest", json=update, timeout=2
+                f"{self.url}/api/ingest", json=update, timeout=5
             )
             response.raise_for_status()
         except RequestException as err:
@@ -794,6 +969,75 @@ class BitfountHub:
             # Raise the same type of exception that was caught
             raise type(err)(err_msg) from err
 
+    def register_user_public_key(self, key: RSAPublicKey) -> Union[str, int]:
+        """Register a public key for the signed-in user.
+
+        Args:
+            key: The public key to register.
+        """
+        key_json: _RegisterUserPublicKeyPOSTJSON = {
+            "key": _RSAEncryption.serialize_public_key(key, form="SSH").decode("utf-8"),
+        }
+        try:
+            response = self.session.post(
+                f"{self.url}/api/{self.username}/keys",
+                json=key_json,
+                timeout=10,
+                params={"version": 2},
+            )
+            response.raise_for_status()
+        except RequestException as err:
+            try:
+                # noinspection PyUnboundLocalVariable
+                err_response: Optional[Response] = response
+            except NameError:
+                # response variable wasn't created yet
+                err_response = None
+            err_msg = self._handle_hub_request_exception(err, err_response)
+            # Raise the same type of exception that was caught
+            raise type(err)(err_msg) from err
+
+        response_json: _CreatedResourceResponseJSON = _extract_response_json(response)
+        return response_json["id"]
+
+    def check_public_key_registered_and_active(
+        self, key_id: str, username: Optional[str] = None
+    ) -> Optional[_ActivePublicKey]:
+        """Return key details from hub if key is registered."""
+        if username is None:
+            username = self.username
+
+        try:
+            response = self.session.get(
+                f"{self.url}/api/{username}/keys/{key_id}",
+            )
+            if response.status_code == 404:
+                return None
+            response.raise_for_status()
+        except RequestException as err:
+            try:
+                # noinspection PyUnboundLocalVariable
+                err_response: Optional[Response] = response
+            except NameError:
+                # response variable wasn't created yet
+                err_response = None
+            err_msg = self._handle_hub_request_exception(err, err_response)
+            # Raise the same type of exception that was caught
+            raise type(err)(err_msg) from err
+
+        # Load key from response
+        response_json: _PublicKeyJSON = _extract_response_json(response)
+
+        if response_json["active"]:
+            return _ActivePublicKey(
+                public_key=_RSAEncryption.load_public_key(
+                    response_json["public_key"].encode()
+                ),
+                id=response_json["id"],
+                active=response_json["active"],
+            )
+        return None
+
 
 class BitfountAM:
     """Typed API for communicating with the Bitfount Access Manager.
@@ -1011,7 +1255,7 @@ class BitfountAM:
     def check_signature_based_access_request(
         self,
         unsigned_task: bytes,
-        signed_task: bytes,
+        task_signature: bytes,
         pod_identifier: str,
         serialized_protocol: SerializedProtocol,
         modeller_name: str,
@@ -1023,7 +1267,7 @@ class BitfountAM:
 
         Args:
             unsigned_task: Unsigned task.
-            signed_task: Signed task.
+            task_signature: Task signature.
             pod_identifier: The pod identifier for the request.
             serialized_protocol: The protocol received from the modeller.
             modeller_name: The name of the modeller.
@@ -1040,10 +1284,10 @@ class BitfountAM:
         """
         post_json = _SignatureBasedAccessCheckPostJSON(
             podIdentifier=pod_identifier,
-            modellerRequest=serialized_protocol,
+            modellerProtocolRequest=serialized_protocol,
             modellerName=modeller_name,
-            unsignedTask=unsigned_task,
-            signedTask=signed_task,
+            unsignedTask=base64.b64encode(unsigned_task).decode("utf-8"),
+            taskSignature=base64.b64encode(task_signature).decode("utf-8"),
             identityProvider="SIGNATURE",
         )
 
