@@ -38,8 +38,10 @@ from bitfount.config import BITFOUNT_LOGS_DIR
 from bitfount.data.databunch import BitfountDataBunch
 from bitfount.data.dataloaders import BitfountDataLoader
 from bitfount.data.datasources.base_source import BaseSource
-from bitfount.data.datastructure import DataStructure
-from bitfount.data.datastructure import registry as datastructure_registry
+from bitfount.data.datastructure import (
+    DataStructure,
+    registry as datastructure_registry,
+)
 from bitfount.data.schema import BitfountSchema, TableSchema
 from bitfount.types import (
     T_FIELDS_DICT,
@@ -292,7 +294,7 @@ class _BaseModel(
         raise NotImplementedError
 
     @abstractmethod
-    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
+    def deserialize(self, content: Union[str, os.PathLike, bytes]) -> None:
         """Implement this method to deserialise a model."""
         raise NotImplementedError
 
@@ -383,6 +385,7 @@ class ClassifierMixIn(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
     schema: BitfountSchema
     fields_dict: ClassVar[T_FIELDS_DICT] = {
         "multilabel": fields.Bool(),
+        "n_classes": fields.Int(),
         "param_clipping": fields.Dict(
             keys=fields.String(), values=fields.Integer(), allow_none=True
         ),
