@@ -10,6 +10,7 @@ from __future__ import annotations
 
 from dataclasses import dataclass
 from enum import Enum, auto
+from io import BytesIO
 import logging
 import os
 
@@ -19,10 +20,12 @@ from typing import Any, List, Literal, Mapping, Optional, Tuple, Union, cast
 
 from marshmallow import fields
 import numpy as np
-import pandas as pd
 from pandas import DataFrame
-from sklearn.linear_model import LinearRegression as sklearnLinearRegression
-from sklearn.linear_model import LogisticRegression as sklearnLogisticRegression
+import pandas as pd
+from sklearn.linear_model import (
+    LinearRegression as sklearnLinearRegression,
+    LogisticRegression as sklearnLogisticRegression,
+)
 from sklearn.neighbors import KNeighborsClassifier
 import statsmodels.api as sm
 
@@ -103,7 +106,7 @@ class LogisticRegressionClassifier(ClassifierMixIn, _BaseModel):
         with open(filename, "wb") as f:
             pickle.dump(self._model, f, protocol=pickle.HIGHEST_PROTOCOL)
 
-    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
+    def deserialize(self, content: Union[str, os.PathLike, bytes]) -> None:
         """Deserialize model.
 
         :::danger
@@ -114,11 +117,14 @@ class LogisticRegressionClassifier(ClassifierMixIn, _BaseModel):
         :::
 
         Args:
-            filename: Path to file containing serialized model.
+            content: Byte strem or path to file containing serialized model.
         """
-        with open(filename, "rb") as f:
-            # TODO: [BIT-987] Review use of pickle.load()
-            self._model = pickle.load(f)  # nosec
+        # TODO: [BIT-987] Review use of pickle.load()
+        if isinstance(content, bytes):
+            self.model = pickle.load(BytesIO(content))  # nosec
+        else:
+            with open(content, "rb") as f:
+                self._model = pickle.load(f)  # nosec
 
     def evaluate(
         self, test_dl: Optional[BitfountDataLoader] = None, *args: Any, **kwargs: Any
@@ -645,7 +651,7 @@ class RegBoostRegressor(RegressorMixIn, _BaseModel):
         with open(filename, "wb") as f:
             pickle.dump(self._model, f, protocol=pickle.HIGHEST_PROTOCOL)
 
-    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
+    def deserialize(self, content: Union[str, os.PathLike, bytes]) -> None:
         """Deserialize model.
 
         :::danger
@@ -656,9 +662,9 @@ class RegBoostRegressor(RegressorMixIn, _BaseModel):
         :::
 
         Args:
-            filename: Path to file containing serialized model.
+            content: Byte stream or path to file containing serialized model.
         """
         self._set_dataloaders()
-        with open(filename, "rb") as f:
+        with open(content, "rb") as f:
             # TODO: [BIT-987] Review use of pickle.load()
             self._model = pickle.load(f)  # nosec
