@@ -51,7 +51,7 @@ logger = logging.getLogger(__name__)
 
 def setup_modeller_from_config_file(
     path_to_config_yaml: Union[str, PathLike],
-) -> Tuple[_Modeller, List[str], Optional[str]]:
+) -> Tuple[_Modeller, List[str], Optional[str], bool]:
     """Creates a modeller from a YAML config file.
 
     Args:
@@ -74,7 +74,7 @@ def setup_modeller_from_config_file(
 
 def setup_modeller_from_config(
     config: ModellerConfig,
-) -> Tuple[_Modeller, List[str], Optional[str]]:
+) -> Tuple[_Modeller, List[str], Optional[str], bool]:
     """Creates a modeller from a loaded config mapping.
 
     Args:
@@ -90,7 +90,7 @@ def setup_modeller_from_config(
         raise FileNotFoundError("Transformation file specified but doesn't exist")
 
     bitfount_hub = _create_bitfounthub(
-        username=config.modeller.username, url=config.hub.url
+        username=config.modeller.username, url=config.hub.url, secrets=config.secrets
     )
 
     # We assume that if the user has not included a username in
@@ -109,7 +109,7 @@ def setup_modeller_from_config(
         idp_url=config.modeller._identity_provider_url,
     )
 
-    return modeller, pod_identifiers, config.project_id
+    return modeller, pod_identifiers, config.project_id, config.run_on_new_data_only
 
 
 def setup_modeller(
@@ -127,7 +127,6 @@ def setup_modeller(
 
     Args:
         pod_identifiers: The pod identifiers of the pods to be used in the task.
-        data_config: The details of the data schema and options to be used in the task.
         task_details: The task details as a TaskConfig instance.
         bitfount_hub: The BitfountHub instance.
         ms_config: The message service settings as a MessageServiceConfig instance.
@@ -367,6 +366,7 @@ async def run_modeller_async(
     require_all_pods: bool = False,
     model_out: Optional[Path] = DEFAULT_MODEL_OUT,
     project_id: Optional[str] = None,
+    run_on_new_data_only: bool = False,
 ) -> Optional[Any]:
     """Runs the modeller.
 
@@ -379,14 +379,19 @@ async def run_modeller_async(
             request to complete task execution.
         model_out: The path to save the model out to. Defaults to "./output-model.pt".
         project_id: Project Id the task belongs to.
+        run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
     Raises:
         PodResponseError: If require_all_pods is true and at least one pod
-            identifier specifed rejects or fails to respond to a task request.
+            identifier specified rejects or fails to respond to a task request.
     """
     # Start task running
     result = await modeller.run_async(
-        pod_identifiers, require_all_pods=require_all_pods, project_id=project_id
+        pod_identifiers,
+        require_all_pods=require_all_pods,
+        project_id=project_id,
+        run_on_new_data_only=run_on_new_data_only,
     )
 
     if model_out:
@@ -405,6 +410,7 @@ def run_modeller(
     require_all_pods: bool = False,
     model_out: Optional[Path] = DEFAULT_MODEL_OUT,
     project_id: Optional[str] = None,
+    run_on_new_data_only: bool = False,
 ) -> Optional[Any]:
     """Runs the modeller.
 
@@ -416,15 +422,22 @@ def run_modeller(
         require_all_pods: Require all pod identifiers specified to accept the task
             request to complete task execution.
         model_out: The path to save the model out to. Defaults to "./output-model.pt".
-        project_id: Project Id the task belongs to.
+        project_id: Project Id the task belongs to. Defaults to None.
+        run_on_new_data_only: Whether to run the task on new datapoints only.
+                Defaults to False.
 
     Raises:
         PodResponseError: If require_all_pods is true and at least one pod
-            identifier specifed rejects or fails to respond to a task request.
+            identifier specified rejects or fails to respond to a task request.
     """
     pod_identifiers = _check_and_update_pod_ids(pod_identifiers, modeller._hub)
     return asyncio.run(
         run_modeller_async(
-            modeller, pod_identifiers, require_all_pods, model_out, project_id
+            modeller,
+            pod_identifiers,
+            require_all_pods,
+            model_out,
+            project_id,
+            run_on_new_data_only,
         )
     )
