@@ -4,7 +4,6 @@ from __future__ import annotations
 from collections import defaultdict
 from contextlib import contextmanager
 from dataclasses import dataclass
-from functools import reduce
 import importlib
 import importlib.util
 import inspect
@@ -28,17 +27,23 @@ from typing import (
     NoReturn,
     Optional,
     Tuple,
+    Type,
     TypeVar,
     Union,
     cast,
 )
 
 import docstring_parser
-import matplotlib.pyplot as plt
+from docstring_parser import DocstringMeta
 import numpy as np
 from sklearn.preprocessing import OneHotEncoder
 
-__all__: List[str] = ["one_hot_encode_list", "seed_all", "ExampleSegmentationData"]
+__all__: List[str] = [
+    "is_notebook",
+    "one_hot_encode_list",
+    "seed_all",
+    "ExampleSegmentationData",
+]
 
 from bitfount import config
 
@@ -48,7 +53,7 @@ DEFAULT_SEED: int = 42
 DOCSTRING_STYLE = docstring_parser.DocstringStyle.GOOGLE
 
 
-def _is_notebook() -> bool:
+def is_notebook() -> bool:
     """Checks if code is being executed in a notebook or not."""
     try:
         # get_ipython() is always available in a notebook, no need to import
@@ -98,11 +103,9 @@ def _get_ipython_extract_symbols() -> Callable[[Any, Any], Tuple[List, List]]:
 
     This should only be imported from an environment which has IPython installed.
     """
-    from IPython.core.magics.code import (  # type: ignore[import] # Reason: No stubs for this # noqa: B950
-        extract_symbols,
-    )
+    from IPython.core.magics.code import extract_symbols
 
-    return extract_symbols  # type: ignore[no-any-return] # Reason: No stubs for this # noqa: B950
+    return extract_symbols
 
 
 def _get_object_source_code(obj: Any) -> str:
@@ -112,7 +115,7 @@ def _get_object_source_code(obj: Any) -> str:
     `inspect.getsource`. If the object is defined in a notebook, a different approach
     must be taken to extract the source code.
     """
-    if _is_notebook():
+    if is_notebook():
         extract_symbols = _get_ipython_extract_symbols()
 
         # Extract the source code from the cell the object is defined in as a string.
@@ -224,7 +227,15 @@ def one_hot_encode_list(
 
 
 def _get_module_from_file(path_to_module: Path) -> ModuleType:
-    """Imports a module from a file path and returns the module.
+    """Creates a module from a file and returns the module.
+
+    :::caution
+
+    This explicitly does NOT add it to sys.modules and hence it is not available
+    for import elsewhere. If you want to have the module added to sys.modules, use
+    _import_module_from_file instead.
+
+    :::
 
     Args:
         path_to_module: The path to the module file.
@@ -238,13 +249,14 @@ def _get_module_from_file(path_to_module: Path) -> ModuleType:
     path = path_to_module.expanduser().absolute()
     module_name = path.stem
 
-    # Load the module from the file without saving it to sys.modules
+    # Load the module from the file WITHOUT SAVING IT TO sys.modules
     # See: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly  # noqa: B950
     spec = importlib.util.spec_from_file_location(module_name, path)
     try:
         # `spec` has type `ModuleSpec | None`. If it is None, an error is raised which
         # caught by the `try...except` block
         module = importlib.util.module_from_spec(spec)  # type: ignore[arg-type] # Reason: See comment # noqa: B950
+        sys.modules[module_name] = module
         spec.loader.exec_module(module)  # type: ignore[union-attr] # Reason: See comment # noqa: B950
     except Exception as ex:
         raise ImportError(f"Unable to load code from {path_to_module}") from ex
@@ -252,6 +264,75 @@ def _get_module_from_file(path_to_module: Path) -> ModuleType:
     return module
 
 
+def _import_module_from_file(
+    path_to_module: Path, parent_module: Optional[str] = None
+) -> Tuple[ModuleType, str]:
+    """Imports a module from a file path and returns the module.
+
+    Ensures module is added to sys.modules.
+
+    Args:
+        path_to_module: The path to the module file.
+        parent_module: The parent module this module should be imported under.
+
+    Returns:
+        The module and the base module name (i.e. the module name without any parent
+        package elements such as `path` for `os.path`).
+
+    Raises:
+        ImportError: If the module cannot be imported.
+    """
+    path = path_to_module.expanduser().absolute()
+    module_name = path.stem
+
+    # Modify module name to reference parent if included
+    if parent_module:
+        fqual_module_name = f"{parent_module}.{module_name}"
+    else:
+        fqual_module_name = module_name
+
+    if (existing_module := sys.modules.get(fqual_module_name)) is None:
+        _logger.debug(f"Dynamically importing {path_to_module} as {fqual_module_name}")
+
+        # Load the module from the file and import it
+        # See: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly  # noqa: B950
+        spec = importlib.util.spec_from_file_location(fqual_module_name, path)
+        try:
+            # `spec` has type `ModuleSpec | None`. If it is None, an error is raised
+            # which is caught by the `try...except` block
+            module = importlib.util.module_from_spec(spec)  # type: ignore[arg-type] # Reason: See comment # noqa: B950
+            sys.modules[fqual_module_name] = module
+            spec.loader.exec_module(module)  # type: ignore[union-attr] # Reason: See comment # noqa: B950
+        except Exception as ex:
+            # If we added the module to sys.modules, remove it as the import failed
+            sys.modules.pop(fqual_module_name, None)  # "None" default prevents KeyError
+
+            raise ImportError(
+                f"Unable to load code from {path_to_module} ({str(ex)})"
+            ) from ex
+
+        return module, module_name
+    else:
+        if (
+            hasattr(existing_module, "__file__")
+            and existing_module.__file__ is not None
+            and Path(existing_module.__file__) != path
+        ):
+            _logger.error(
+                "A module with the same name but a different source already exists,"
+                " cannot register"
+            )
+            raise ImportError(
+                f"A module with this name already exists: {existing_module}"
+            )
+        else:
+            _logger.debug(
+                f"{str(path)} is already loaded as {existing_module.__name__},"
+                f" will not reload"
+            )
+            return existing_module, module_name
+
+
 def _get_non_abstract_classes_from_module(path_to_module: Path) -> Dict[str, type]:
     """Returns non-abstract classes from a module path.
 
@@ -516,7 +597,7 @@ def _handle_fatal_error(
 class ExampleSegmentationData:
     """A synthetic segmentation dataset example.
 
-    This class is used mainly in Tutorial 11.
+    This class is used mainly in Tutorial 10 and testing.
     """
 
     def masks_to_colorimg(self, masks: np.ndarray) -> np.ndarray:
@@ -629,37 +710,14 @@ class ExampleSegmentationData:
         )
         return new_arr
 
-    def _plot_img_array(self, img_array: np.ndarray, ncol: int = 3) -> None:
-        """Helper function for tutorial 11."""
-        nrow = len(img_array) // ncol
 
-        f, plots = plt.subplots(
-            nrow, ncol, sharex="all", sharey="all", figsize=(ncol * 4, nrow * 4)  # type: ignore[arg-type] # Reason: see below # noqa: B950
-        )
-        # From matplotlib documentation:
-        # sharex, sharey: bool or {'none', 'all', 'row', 'col'}, default: False
-        for i in range(len(img_array)):
-            plots[i // ncol, i % ncol]  # type: ignore[index] # reason: see below # noqa: B950
-            plots[i // ncol, i % ncol].imshow(img_array[i])  # type: ignore[index] # reason: see below # noqa: B950
-        # matplotlib allows indexing of `Axes` objects,
-        # not sure why mypy raises this error
-
-    def plot_side_by_side(self, img_arrays: np.ndarray) -> None:
-        """Helper function for tutorial 11.
-
-        Plots side by side images and masks.
-
-        Args:
-            img_arrays: The array of images and masks pairs to plot.
-        """
-        flatten_list = reduce(lambda x, y: x + y, zip(*img_arrays))
-
-        self._plot_img_array(np.array(flatten_list), ncol=len(img_arrays))
+# Type var for class
+_C = TypeVar("_C", bound=Type)
 
 
 def delegates(
     keep: bool = False, exclude_doc_meta: Optional[List[Any]] = None
-) -> Callable[[Any], None]:
+) -> Callable[[_C], _C]:
     """Decorator to replace `**kwargs` in signature.
 
     The `@delegates()` decorator should be added to all classes that inherit
@@ -669,69 +727,81 @@ def delegates(
     and errors are reflected in the class it decorates.
     """
 
-    def _f(f: Any) -> Any:
+    def _decorate_class(klass: _C) -> _C:
         # Mypy complains that DocstringMeta is not iterable, when it is,
         # hence the "type: ignore"s.
+
         # get the signature of the decorated class
-        from_f = f.__init__
-        sig = inspect.signature(from_f)
-        sigd = dict(sig.parameters)
+        from_klass = klass.__init__
+        init_sig = inspect.signature(from_klass)
+        init_params = dict(init_sig.parameters)
+
         # get the docstring of the decorated class
-        fdoc = docstring_parser.parse(f.__doc__, style=DOCSTRING_STYLE)
+        klass_doc = docstring_parser.parse(
+            klass.__doc__ if klass.__doc__ else "", style=DOCSTRING_STYLE
+        )
+
         # aux variables to keep track of everything
         # "combined" keeps track of all args, attributes and errors across all classes.
-        combined: Dict[Any, Any] = {}
+        combined: DefaultDict[Type[DocstringMeta], List[DocstringMeta]] = defaultdict(
+            list
+        )
         # "metas" keeps track of all args, attributes and errors for one class only.
-        metas: Dict[Any, Any] = {}
+        metas: DefaultDict[Type[DocstringMeta], List[DocstringMeta]] = defaultdict(list)
         # "arg_names" keeps track of all the arg (also attrs & errors)
         # names to ensure there are no duplicates.
         arg_names: List[str] = []
-        # get all the metadata from the decorated class' docstring
-        for meta in fdoc.meta:
+
+        # get all the metadata from the decorated class' docstring and...
+        for meta in klass_doc.meta:
             meta_type = type(meta)
             if exclude_doc_meta is not None:
                 if meta_type in exclude_doc_meta:
                     continue
-            metas.setdefault(meta_type, []).append(meta)
-        for (meta_type, meta) in metas.items():
-            combined[meta_type] = meta
-            arg_names += [arg.args[1] for arg in meta]  # type: ignore[attr-defined] # Reason: see top of function # noqa: B950
-        if "kwargs" in sigd:
-            k = sigd.pop("kwargs")
+            metas[meta_type].append(meta)
+        # ...feed it into our aux variables
+        for meta_type, meta_list in metas.items():
+            combined[meta_type].extend(meta_list)
+            arg_names += [arg.args[1] for arg in meta_list]
+
+        if "kwargs" in init_params:
+            init_kwargs = init_params.pop("kwargs")
         else:
-            k = None
+            init_kwargs = None
+
         # We only need the ancestors from our own repo.
         ancestors = [
-            item
-            for item in f.__mro__
-            if item.__module__.startswith("bitfount") and item != f
+            ancestor
+            for ancestor in klass.__mro__
+            if ancestor.__module__.startswith("bitfount") and ancestor != klass
         ]
         # ancestors.sort(key=lambda x: len(x.__module__.split(".")))
 
         # now we loop through all the base classes of the
         # decorated class and extract the relevant items
-        for item in ancestors:
+        for ancestor in ancestors:
             # Load public methods from parent classes
 
             # pdoc checks whether the methods are in the class's dictionary,
             # so we need to add them to the relevant class to ensure they
             # show up in the docs.
-            to_f_dict = item.__dict__
+            to_klass_dict = ancestor.__dict__
             # get the abstract methods
-            abstract_methods = (
-                item.__abstractmethods__ if hasattr(item, "__abstractmethods__") else []
-            )
+            if hasattr(ancestor, "__abstractmethods__"):
+                abstract_methods = ancestor.__abstractmethods__  # type: ignore[attr-defined] # Reason: hasattr check # noqa: B950
+            else:
+                abstract_methods = []
 
-            # Get the public non-abstract methods
+            # Get the public non-abstract attributes
             new_dict_items = {
-                k: v
-                for k, v in to_f_dict.items()
-                if not k.startswith("_")
-                and k not in f.__dict__
-                and k not in abstract_methods
+                attr_name: attr
+                for attr_name, attr in to_klass_dict.items()
+                if not attr_name.startswith("_")
+                and attr_name not in klass.__dict__
+                and attr_name not in abstract_methods
             }
 
-            for k, v in new_dict_items.items():
+            for attr_name, attr in new_dict_items.items():
                 # `If` checks below explained:
                 # 1. We only need to document the methods from bitfount modules.
                 # 2. Only get the public methods from parent classes.
@@ -740,81 +810,81 @@ def delegates(
                 # This will no affect in any way the error
                 # handling for abstract methods not being implemented.
                 if (
-                    item.__module__.startswith("bitfount")
-                    and not str(k).startswith("_")
-                    and k not in f.__dict__
-                    and callable(to_f_dict[k])
+                    ancestor.__module__.startswith("bitfount")
+                    and not attr_name.startswith("_")
+                    and attr_name not in klass.__dict__
+                    and callable(to_klass_dict[attr_name])
                 ):
                     # `__dict__` is a mappingproxy which cannot be
                     # updated in the same way we would update a
                     # dictionary, so need the setattr instead.
-                    setattr(f, str(k), v)
+                    setattr(klass, attr_name, attr)
 
             # get the docstring and the metadata
-            item_doc = docstring_parser.parse(item.__doc__, style=DOCSTRING_STYLE)
+            item_doc = docstring_parser.parse(
+                ancestor.__doc__ if ancestor.__doc__ else "", style=DOCSTRING_STYLE
+            )
+
             # "metas" keeps track of all args, attributes and errors for one class only.
-            metas = {}
+            metas = defaultdict(list)  # reset to empty
             for meta in item_doc.meta:
                 meta_type = type(meta)
                 if exclude_doc_meta is not None:
                     if meta_type in exclude_doc_meta:
                         continue
-                metas.setdefault(meta_type, []).append(meta)
-            for (meta_type, meta) in metas.items():
-                to_add = [arg for arg in meta if arg.args[1] not in arg_names]  # type: ignore[attr-defined] # Reason: see top of function  # noqa: B950
-                arg_names += [arg.args[1] for arg in meta]  # type: ignore[attr-defined] # Reason: see top of function  # noqa: B950
-                if meta_type in combined.keys():
-                    combined[meta_type] += to_add
-                else:
-                    combined[meta_type] = to_add
-                    # now sort alphabetically the args, attrs
-                    # and errors in the docstring.
-            for k in combined.keys():
-                combined[k] = [
-                    item
-                    for item in sorted(combined[k], key=lambda d: d.args[1])  # type: ignore[no-any-return] # Reason: see below # noqa: B950
-                ]
-                # Mypy raises error for the sorted function due to
-                # the definition of sorted having changed,
-                # see here: https://github.com/python/mypy/issues/9656
+                metas[meta_type].append(meta)
+            # keep track in aux variables
+            for meta_type, meta_list in metas.items():
+                to_add = [arg for arg in meta_list if arg.args[1] not in arg_names]
+                arg_names += [arg.args[1] for arg in meta_list]
+                combined[meta_type].extend(to_add)
+
+            # now sort alphabetically the args, attrs
+            # and errors in the docstring.
+            # NOTE: have to sort in place in order to keep the defaultdict
+            for k, v in combined.items():
+                combined[k] = sorted(v, key=lambda d: d.args[1])
 
             # Now get the signature.
-            to_f = item.__init__
+            to_klass = ancestor.__init__  # type: ignore[misc] # Reason: _can_ access __init__ directly # noqa: B950
             # extract the "self" param from the signature
-            self_param = sigd.pop("self")
+            self_param = init_params.pop("self")
             # get the values without default first
             s1 = {
-                k: v
-                for k, v in inspect.signature(to_f).parameters.items()
-                if v.default == inspect.Parameter.empty
-                and k not in sigd
-                and k not in ["args", "kwargs"]
+                param_name: param
+                for param_name, param in inspect.signature(to_klass).parameters.items()
+                if param.default == inspect.Parameter.empty
+                and param_name not in init_params
+                and param_name not in ["args", "kwargs"]
             }
             # "self" needs to go first, then unparametrised args
             # and then all the other args otherwise an error is raised.
-            sigd = {"self": self_param, **s1, **sigd}
+            init_params = {"self": self_param, **s1, **init_params}
             # Then add the params which have default values.
             s2 = {
-                k: v
-                for k, v in inspect.signature(to_f).parameters.items()
-                if v.default != inspect.Parameter.empty and k not in sigd
+                param_name: param
+                for param_name, param in inspect.signature(to_klass).parameters.items()
+                if param.default != inspect.Parameter.empty
+                and param_name not in init_params
             }
 
-            sigd.update(s2)
+            init_params.update(s2)
+
         # Check if kwargs should be kept in the signature
-        if keep and k is not None:
-            sigd["kwargs"] = k
+        if keep and init_kwargs is not None:
+            init_params["kwargs"] = init_kwargs
+
         # replace the decorated class's signature.
-        from_f.__signature__ = sig.replace(parameters=sigd.values())  # type: ignore[arg-type] # Reason: see below # noqa: B950
+        from_klass.__signature__ = init_sig.replace(parameters=init_params.values())  # type: ignore[arg-type] # Reason: see below # noqa: B950
         # Mypy complains that Argument "parameters" to "replace" of
         # "Signature" has incompatible type "dict_values[str, Parameter]";
         # expected "Union[Sequence[Parameter], Type[_void], None], even though
         # Signature.replace is untyped.
 
         # Now, replace the decorated class's docstring metadata.
-        fdoc.meta = list(chain(*combined.values()))
-        f.__doc__ = docstring_parser.compose(fdoc, style=DOCSTRING_STYLE)
+        klass_doc.meta = list(chain(*combined.values()))
+        klass.__doc__ = docstring_parser.compose(klass_doc, style=DOCSTRING_STYLE)
 
-        return f
+        return klass
 
-    return _f
+    return _decorate_class
