@@ -25,7 +25,7 @@ from bitfount.data.datastructure import DataStructure
 from bitfount.data.exceptions import DataNotLoadedError
 from bitfount.data.schema import BitfountSchema
 from bitfount.data.types import DataSplit, SemanticType
-from bitfount.data.utils import DatabaseConnection, _generate_dtypes_hash
+from bitfount.data.utils import DatabaseConnection
 from tests.utils.helper import (
     DATASET_ROW_COUNT,
     TABLE_NAME,
@@ -388,6 +388,96 @@ class TestBitfountDataset:
             datasource._test_idxs
         ) == len(datasource.data)
 
+    @integration_test
+    def test_database_query_input(
+        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
+    ) -> None:
+        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
+
+        Query database connection.
+        """
+        db_conn = DatabaseConnection(
+            db_session,
+            query="""
+            SELECT *
+            FROM dummy_data d1
+            LEFT JOIN dummy_data_2 d2
+            ON 'd1.Date' = 'd2.Date'
+            """,
+        )
+        datasource = DatabaseSource(db_conn, seed=420)
+        datasource.load_data()
+
+        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
+        mocker.patch.object(_BitfountDataset, "_reformat_data")
+
+        dataset = _BitfountDataset(
+            datasource, Mock(value="train"), Mock(), Mock(), Mock()
+        )
+        dataset._split_data()
+        assert not datasource.multi_table
+        assert datasource.data is not None
+        assert datasource._train_idxs is not None
+        assert datasource._validation_idxs is not None
+        assert datasource._test_idxs is not None
+        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
+            datasource._test_idxs
+        ) == len(datasource.data)
+
+    @integration_test
+    def test_database_single_table_input(
+        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
+    ) -> None:
+        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
+
+        Single table database connection.
+        """
+        db_conn = DatabaseConnection(db_session, table_names=["dummy_data"])
+        datasource = DatabaseSource(db_conn, seed=420)
+        datasource.load_data()
+        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
+        mocker.patch.object(_BitfountDataset, "_reformat_data")
+
+        dataset = _BitfountDataset(
+            datasource, Mock(value="train"), Mock(), Mock(), Mock()
+        )
+        dataset._split_data()
+        assert datasource.data is not None
+        assert not datasource.multi_table
+        assert datasource._train_idxs is not None
+        assert datasource._validation_idxs is not None
+        assert datasource._test_idxs is not None
+        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
+            datasource._test_idxs
+        ) == len(datasource.data)
+
+    @unit_test
+    def test_mock_database_single_table_input(
+        self, mock_engine: Mock, mock_pandas_read_sql_query: None, mocker: MockerFixture
+    ) -> None:
+        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
+
+        Mock single table database connection.
+        """
+        db_conn = DatabaseConnection(mock_engine, table_names=["dummy_data"])
+        datasource = DatabaseSource(db_conn, seed=420)
+        datasource.load_data()
+        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
+        mocker.patch.object(_BitfountDataset, "_reformat_data")
+
+        dataset = _BitfountDataset(
+            datasource, Mock(value="train"), Mock(), Mock(), Mock()
+        )
+        dataset._split_data()
+        assert datasource.data is not None
+        assert not datasource.multi_table
+        assert datasource._train_idxs is not None
+        assert datasource._validation_idxs is not None
+        assert datasource._test_idxs is not None
+        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
+            datasource._test_idxs
+        ) == len(datasource.data)
+
     @unit_test
     def test_len_tab_data(self, tabular_dataset: _BitfountDataset) -> None:
         """Tests tabular dataset __len__ method."""
@@ -412,6 +502,32 @@ class TestBitfountDataset:
         assert multiimage_dataset.datasource._train_idxs is not None
         assert len(multiimage_dataset) == len(multiimage_dataset.datasource._train_idxs)
 
+    @unit_test
+    def test_datasource_data_split_called_twice(
+        self, caplog: LogCaptureFixture, tabular_dataset: _BitfountDataset
+    ) -> None:
+        """Tests that the log is printed if split_data called twice.
+
+        The first `_split_data` call happens automatically in the `__init__` method.
+        """
+        caplog.set_level(logging.DEBUG)
+        assert tabular_dataset.datasource._data_is_split
+        tabular_dataset._split_data()
+        assert "Data is already split, keeping the current split." in caplog.text
+
+    @unit_test
+    def test_data_not_loaded_raises_error(
+        self, mock_engine: Mock, mock_pandas_read_sql_query: None, mocker: MockerFixture
+    ) -> None:
+        """Checks _BitfountDataset raises a DataNotLoadedError if data is not loaded."""
+        db_conn = DatabaseConnection(mock_engine, table_names=["dummy_data"])
+        datasource = DatabaseSource(db_conn, seed=420)
+        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
+        mocker.patch.object(_BitfountDataset, "_reformat_data")
+
+        with pytest.raises(DataNotLoadedError):
+            _BitfountDataset(datasource, Mock(value="train"), Mock(), Mock(), Mock())
+
 
 class TestIterableBitfountDataset:
     """Tests IterableBitfountDataset class."""
@@ -585,18 +701,6 @@ class TestIterableBitfountDataset:
         with pytest.raises(ValueError, match="No query or table name specified."):
             next(iter_db_dataset.yield_dataset_split(DataSplit.TRAIN))
 
-    @integration_test
-    def test_validation_set_unsplit_raises_error(
-        self, iter_db_dataset: _IterableBitfountDataset
-    ) -> None:
-        """Tests that calling the `validation_set` property raises an error.
-
-        The error should be a `DataNotLoadedError` when the data is multi-table because
-        it is never loaded.
-        """
-        with pytest.raises(DataNotLoadedError):
-            iter_db_dataset.get_dataset_split(DataSplit.VALIDATION)
-
     @unit_test
     def test_yield_dataset_split_works_correctly(
         self, dataframe: pd.DataFrame, mock_engine: Mock, mocker: MockerFixture
@@ -618,7 +722,7 @@ class TestIterableBitfountDataset:
         mock_engine.execution_options.return_value = mock_engine
         mock_db_connection.execute.return_value = mock_result
 
-        # Creates a multitable DatabaseConnection object
+        # Creates a multi-table DatabaseConnection object
         db_conn = DatabaseConnection(
             mock_engine,
             table_names=["dummy_data", "dummy_data_2"],
@@ -753,100 +857,6 @@ class TestIterableBitfountDataset:
             next(dataset.yield_dataset_split(split=DataSplit.TRAIN)), expected_output
         )
 
-    @integration_test
-    def test_database_single_table_input(
-        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
-    ) -> None:
-        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
-
-        Single table database connection.
-        """
-        db_conn = DatabaseConnection(db_session, table_names=["dummy_data"])
-        datasource = DatabaseSource(db_conn, seed=420)
-        datasource.load_data()
-        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
-        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
-        dataset._split_data()
-        assert datasource.data is not None
-        assert not datasource.multi_table
-        assert datasource._train_idxs is not None
-        assert datasource._validation_idxs is not None
-        assert datasource._test_idxs is not None
-        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
-            datasource._test_idxs
-        ) == len(datasource.data)
-
-    @unit_test
-    def test_mock_database_single_table_input(
-        self, mock_engine: Mock, mock_pandas_read_sql_query: None, mocker: MockerFixture
-    ) -> None:
-        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
-
-        Mock single table database connection.
-        """
-        db_conn = DatabaseConnection(mock_engine, table_names=["dummy_data"])
-        datasource = DatabaseSource(db_conn, seed=420)
-        datasource.load_data()
-        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
-        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
-        dataset._split_data()
-        assert datasource.data is not None
-        assert not datasource.multi_table
-        assert datasource._train_idxs is not None
-        assert datasource._validation_idxs is not None
-        assert datasource._test_idxs is not None
-        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
-            datasource._test_idxs
-        ) == len(datasource.data)
-
-    @integration_test
-    def test_database_query_input(
-        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
-    ) -> None:
-        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.
-
-        Query database connection.
-        """
-        db_conn = DatabaseConnection(
-            db_session,
-            query="""
-            SELECT *
-            FROM dummy_data d1
-            LEFT JOIN dummy_data_2 d2
-            ON 'd1.Date' = 'd2.Date'
-            """,
-        )
-        datasource = DatabaseSource(db_conn, seed=420)
-        datasource.load_data()
-
-        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
-        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
-        dataset._split_data()
-        assert not datasource.multi_table
-        assert datasource.data is not None
-        assert datasource._train_idxs is not None
-        assert datasource._validation_idxs is not None
-        assert datasource._test_idxs is not None
-        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
-            datasource._test_idxs
-        ) == len(datasource.data)
-
-    @unit_test
-    def test_datasource_data_split_called_twice(
-        self, caplog: LogCaptureFixture, iter_tabular_dataset: _IterableBitfountDataset
-    ) -> None:
-        """Tests that the log is printed if split_data called twice."""
-        caplog.set_level(logging.DEBUG)
-
-        assert iter_tabular_dataset.datasource._data_is_split is False
-
-        iter_tabular_dataset._split_data()
-
-        assert iter_tabular_dataset.datasource._data_is_split is True
-        # split data in DataBunch into training, validation, test sets
-        iter_tabular_dataset._split_data()
-        assert "Data is already split, keeping the current split." in caplog.text
-
     @unit_test
     @pytest.mark.parametrize(
         "datasource_splitter, datastructure_splitter",
@@ -904,68 +914,6 @@ class TestIterableBitfountDataset:
             assert resolved_splitter is not datastructure_splitter
             assert resolved_splitter is not datasource_splitter
 
-    @unit_test
-    def test_training_set_unloaded_raises_error(
-        self, mock_engine: Mock, mocker: MockerFixture
-    ) -> None:
-        """Tests that calling the `training_set` property raises an error.
-
-        The error should be a `DataNotLoadedError` when the data is multi-table because
-        it is never loaded.
-        """
-        target = "TARGET"
-        table_name = "dummy_data"
-        datastructure = DataStructure(
-            target=target, ignore_cols=["image"], table=table_name
-        )
-
-        db_conn = DatabaseConnection(
-            mock_engine, table_names=["dummy_data", "dummy_data_2"]
-        )
-        datasource = DatabaseSource(db_conn, seed=420)
-        mock_dtypes = {"col": np.dtype(int)}
-        mocker.patch.object(datasource, "get_dtypes", return_value=mock_dtypes)
-        datasource._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
-        datasource.load_data()
-        schema = BitfountSchema()
-        schema.add_datasource_tables(datasource, table_name=table_name)
-        # Using IterableBitfountDataset but get_dataset_split is a
-        # concrete method of _BaseBitfountDataset
-        db_dataset = _IterableBitfountDataset(
-            datasource=datasource,
-            target=target,
-            selected_cols=datastructure.selected_cols,
-            data_split=DataSplit.TRAIN,
-            schema=schema.get_table_schema(table_name),
-            selected_cols_semantic_types=datastructure.selected_cols_w_types,
-        )
-        with pytest.raises(DataNotLoadedError):
-            db_dataset.get_dataset_split(DataSplit.TRAIN)
-
-    @unit_test
-    def test_datasource_data_split_flag_updated(self, dataframe: pd.DataFrame) -> None:
-        """Tests that the data_is_split flag is updated."""
-        target = "TARGET"
-        ds = DataFrameSource(dataframe, selected_cols=["M", "F"])
-        ds.load_data()
-        assert ds._data_is_split is False
-
-        schema = BitfountSchema()
-        schema.add_datasource_tables(ds, table_name=TABLE_NAME)
-        datastructure = DataStructure(
-            target=target, ignore_cols=["image"], table=TABLE_NAME
-        )
-        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
-        dataset = _IterableBitfountDataset(
-            datasource=ds,
-            selected_cols_semantic_types=datastructure.selected_cols_w_types,
-            selected_cols=datastructure.selected_cols,
-            data_split=DataSplit.TRAIN,
-            schema=schema.get_table_schema(TABLE_NAME),
-        )
-        dataset._split_data()
-        assert ds._data_is_split is True
-
     @unit_test
     def test_get_dataset_split_length_executes_query(
         self,
@@ -992,7 +940,7 @@ class TestIterableBitfountDataset:
         mock_connect.assert_called_once()
 
     @integration_test
-    def test_get_dataset_split_length_with_multitable_database_connection(
+    def test_get_dataset_split_length_with_multi_table_database_connection(
         self,
         dataframe: pd.DataFrame,
         db_session: sqlalchemy.engine.base.Engine,
