@@ -326,7 +326,9 @@ class TestPrivateSetIntersection:
             message_service=Mock(),
         )
 
-        mock_modeller.run.assert_called_once_with(["fake/fake"], require_all_pods=False)
+        mock_modeller.run.assert_called_once_with(
+            ["fake/fake"], require_all_pods=False, run_on_new_data_only=False
+        )
 
 
 @unit_test
