@@ -419,7 +419,7 @@ class TestPod:
         )
         assert pod.show_datapoints_with_results_in_db is False
 
-    def test_multi_table_datasource_pod_db_false(
+    def test__update_pod_db_multi_table(
         self,
         approved_pods: List[str],
         mock_access_manager: Mock,
@@ -428,11 +428,14 @@ class TestPod:
         mock_pod_data_config: DataclassMock,
         mock_pod_details_config: DataclassMock,
         mock_pod_keys: DataclassMock,
+        mocker: MockerFixture,
         multi_table_excel_source: ExcelSource,
         pod_name: str,
         username: str,
     ) -> None:
-        """Tests that pod_db is false if datasource is multi-tab;e."""
+        """Tests that _add_data_to_db for multi-table datasources."""
+        mock_add_to_db = mocker.patch("bitfount.federated.pod._add_data_to_pod_db")
+
         pod = Pod(
             name=pod_name,
             datasource=multi_table_excel_source,
@@ -447,7 +450,52 @@ class TestPod:
             pod_db=True,
             show_datapoints_with_results_in_db=True,
         )
+        assert mock_add_to_db.call_count == len(pod.schema.tables)
+
+    def test_databasesource_pod_db_false(
+        self,
+        approved_pods: List[str],
+        caplog: LogCaptureFixture,
+        mock_access_manager: Mock,
+        mock_bitfount_hub: Mock,
+        mock_message_service_config: DataclassMock,
+        mock_pod_data_config: DataclassMock,
+        mock_pod_details_config: DataclassMock,
+        mock_pod_keys: DataclassMock,
+        mocker: MockerFixture,
+        pod_name: str,
+        username: str,
+    ) -> None:
+        """Tests that pod_db is false if datasource is DatabaseSource."""
+        db_conn = Mock(
+            spec=DatabaseConnection,
+            table_names=["table1"],
+            multi_table=False,
+            query=None,
+            con=Mock(),
+            db_schema="test",
+        )
+        mocker.patch.object(Pod, "_setup_schema")
+        datasource = DatabaseSource(db_conn=db_conn)
+        pod = Pod(
+            name=pod_name,
+            datasource=datasource,
+            username=username,
+            data_config=mock_pod_data_config,
+            pod_details_config=mock_pod_details_config,
+            hub=mock_bitfount_hub,
+            message_service=mock_message_service_config,
+            access_manager=mock_access_manager,
+            pod_keys=mock_pod_keys,
+            approved_pods=approved_pods,
+            pod_db=True,
+            show_datapoints_with_results_in_db=True,
+        )
         assert pod.pod_db is False
+        assert (
+            "Pod database not supported for DatabaseSource. "
+            "Starting pod without database." in caplog.text
+        )
 
     @pytest.mark.skipif(
         condition=platform.system() == "Windows",
@@ -471,8 +519,8 @@ class TestPod:
         username: str,
     ) -> None:
         """Tests database is created for pod."""
-        if os.path.exists(f"{pod_name}.db"):
-            os.remove(f"{pod_name}.db")
+        if os.path.exists(f"{pod_name}.sqlite"):
+            os.remove(f"{pod_name}.sqlite")
         ds = create_datasource(classification=True)
         Pod(
             name=pod_name,
@@ -488,15 +536,15 @@ class TestPod:
             pod_db=True,
             show_datapoints_with_results_in_db=True,
         )
-        assert os.path.exists(f"{pod_name}.db")
-        con = sqlite3.connect(f"{pod_name}.db")
-        existing_data = pd.read_sql('SELECT * FROM "datasource"', con)
+        assert os.path.exists(f"{pod_name}.sqlite")
+        con = sqlite3.connect(f"{pod_name}.sqlite")
+        existing_data = pd.read_sql(f'SELECT * FROM "{pod_name}"', con)
         assert existing_data.shape[0] == ds._data.shape[0]
         assert (
-            existing_data.shape[1] - 1 == ds._data.shape[1]
-        )  # existing_data has the extra datapoint_hash column
+            existing_data.shape[1] - 2 == ds._data.shape[1]
+        )  # existing_data has the extra columns datapoint_hash and rowID
         con.close()
-        os.remove(f"{pod_name}.db")
+        os.remove(f"{pod_name}.sqlite")
 
     def test_datapoint_hash_in_datasource_raises_error(
         self,
@@ -551,8 +599,8 @@ class TestPod:
         username: str,
     ) -> None:
         """Tests only new datapoint added to db."""
-        if os.path.exists(f"{pod_name}.db"):
-            os.remove(f"{pod_name}.db")
+        if os.path.exists(f"{pod_name}.sqlite"):
+            os.remove(f"{pod_name}.sqlite")
         dataset = create_dataset(classification=True)
         dataset.drop("Date", inplace=True, axis=1)
         ds = DataFrameSource(data=dataset[0:3000])
@@ -570,13 +618,13 @@ class TestPod:
             pod_db=True,
             show_datapoints_with_results_in_db=True,
         )
-        assert os.path.exists(f"{pod_name}.db")
-        con = sqlite3.connect(f"{pod_name}.db")
-        existing_data = pd.read_sql('SELECT * FROM "datasource"', con)
+        assert os.path.exists(f"{pod_name}.sqlite")
+        con = sqlite3.connect(f"{pod_name}.sqlite")
+        existing_data = pd.read_sql(f'SELECT * FROM "{pod_name}"', con)
         assert existing_data.shape[0] == ds._data.shape[0]
         assert (
-            existing_data.shape[1] - 1 == ds._data.shape[1]
-        )  # existing_data has the extra datapoint_hash column
+            existing_data.shape[1] - 2 == ds._data.shape[1]
+        )  # existing_data has the extra datapoint_hash and rowID columns
         ds = DataFrameSource(data=dataset)
         Pod(
             name=pod_name,
@@ -593,12 +641,12 @@ class TestPod:
             show_datapoints_with_results_in_db=True,
         )
 
-        existing_data = pd.read_sql('SELECT * FROM "datasource"', con)
+        existing_data = pd.read_sql(f'SELECT * FROM "{pod_name}"', con)
         # Check that only the new datapoints have been added to the ds
         assert existing_data.shape[0] == dataset.shape[0]
-        assert existing_data.shape[1] - 1 == dataset.shape[1]
+        assert existing_data.shape[1] - 2 == dataset.shape[1]
         con.close()
-        os.remove(f"{pod_name}.db")
+        os.remove(f"{pod_name}.sqlite")
 
     def test_schema_raises_error_basesource_flag(
         self,
@@ -821,6 +869,7 @@ class TestPod:
                     # We don't actually encrypt this due to the mocked decryption
                     encrypted_request=task_request.serialize(),
                 ).serialize(),
+                run_on_new_data_only=False,
             ).serialize()
         )
 
@@ -848,6 +897,8 @@ class TestPod:
             serialized_protocol=task_protocol_details,
             pod_db=False,
             show_datapoints_in_results_db=False,
+            run_on_new_data_only=False,
+            project_id=None,
         )
 
     async def test__create_and_run_worker_with_secure_aggregation_unapproved_pods(
@@ -947,6 +998,7 @@ class TestPod:
                     # We don't actually encrypt this due to the mocked decryption
                     encrypted_request=task_request.serialize(),
                 ).serialize(),
+                run_on_new_data_only=False,
             ).serialize()
         )
 
@@ -979,6 +1031,8 @@ class TestPod:
             serialized_protocol=task_protocol_details,
             pod_db=False,
             show_datapoints_in_results_db=False,
+            run_on_new_data_only=False,
+            project_id=None,
         )
 
         # Ensure interpod mailbox creation logged
