@@ -211,9 +211,9 @@ def mock_message_type() -> Mock:
 
 
 @fixture
-def dummy_message_awaitable() -> Callable[
-    [Optional[_BitfountMessage], Optional[bool]], object
-]:
+def dummy_message_awaitable() -> (
+    Callable[[Optional[_BitfountMessage], Optional[bool]], object]
+):
     """A dummy awaitable for testing."""
 
     class MockMessageAwaitable:
