@@ -3,7 +3,6 @@ from dataclasses import dataclass
 from typing import Optional, Union, cast
 
 from _pytest.logging import LogCaptureFixture
-import matplotlib
 import numpy as np
 import pandas as pd
 import pytest
@@ -242,27 +241,6 @@ class TestMetricCollection:
         assert m.results["Recall"] == 0.75
         assert m.optimal_threshold is None
 
-    @unit_test
-    def test_view_thresholds(self) -> None:
-        """Tests that chart displaying thresholds can be created properly.
-
-        Including the underlying data.
-        """
-        y_true = np.array([0, 1, 1, 0, 1, 1])
-        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])
-
-        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
-        m.compute(y_true, y_prob)
-
-        assert len(m.thresholds) == 201  # num of thresholds (0-100 at 0.5 increments)
-        assert isinstance(m.threshold_metrics, dict)
-        for k, v in m.threshold_metrics.items():
-            assert k in m.metrics
-            assert len(v) == len(m.thresholds)
-
-        fig = m.view_thresholds()
-        assert isinstance(fig, matplotlib.figure.Figure)
-
     @integration_test
     def test_custom_metric(self) -> None:
         """Test custom metric can be supplied."""
