@@ -1,16 +1,21 @@
 """Tests for transformation processing."""
-from typing import Tuple, Type, cast
+from typing import Mapping, Tuple, Type, Union, cast
 
 from PIL import Image
-import databricks.koalas as ks
 import numpy as np
 import pandas as pd
 from pandas._testing import assert_frame_equal
 import pytest
 from pytest import fixture
 
-from bitfount.data.datasource import DataSource
+from bitfount.data.datasources.dataframe_source import DataFrameSource
 from bitfount.data.schema import BitfountSchema, TableSchema
+from bitfount.data.types import DataSplit
+from bitfount.transformations.dataset_operations import (
+    NormalizeDataTransformation,
+    ScalarAdditionDataTransformation,
+    ScalarMultiplicationDataTransformation,
+)
 from bitfount.transformations.exceptions import (
     InvalidBatchTransformationError,
     MissingColumnReferenceError,
@@ -20,14 +25,13 @@ from bitfount.transformations.exceptions import (
 from bitfount.transformations.parser import TransformationsParser
 from bitfount.transformations.processor import TransformationProcessor
 from bitfount.transformations.unary_operations import OneHotEncodingTransformation
-from bitfount.types import _DataFrameLib, _DataFrameType
 from tests.utils import PytestRequest
 from tests.utils.helper import create_dataset, integration_test, unit_test
 
 
 def parse_and_process_data(
-    data: _DataFrameType, schema: TableSchema, transformations_yaml: str
-) -> Tuple[_DataFrameType, TableSchema]:
+    data: pd.DataFrame, schema: TableSchema, transformations_yaml: str
+) -> Tuple[pd.DataFrame, TableSchema]:
     """Parses transformations YAML and applies them to data."""
     parser = TransformationsParser()
     transformations, cols = parser.parse(transformations_yaml)
@@ -47,10 +51,10 @@ class TestTransformationProcessor:
     @fixture(scope="function")
     def data_and_schema(
         self, request: PytestRequest
-    ) -> Tuple[_DataFrameType, TableSchema]:
+    ) -> Tuple[pd.DataFrame, TableSchema]:
         """Dataset and schema to be used for tests."""
-        data = create_dataset(koalas=request.param)
-        datasource = DataSource(data)
+        data = create_dataset()
+        datasource = DataFrameSource(data)
         schema = BitfountSchema()
         schema.add_datasource_tables(datasource, table_name="test")
         datasource.load_data()
@@ -70,7 +74,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_addition_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests addition transformation works correctly."""
         data, schema = data_and_schema
@@ -95,7 +99,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_comparison_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests comparison transformation works correctly."""
         data, schema = data_and_schema
@@ -121,7 +125,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_comparison_transformation_with_constant(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests comparison transformation works correctly with a constant."""
         data, schema = data_and_schema
@@ -147,7 +151,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_division_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests division transformation works correctly."""
         data, schema = data_and_schema
@@ -174,7 +178,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_inclusion_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests inclusion transformation works correctly."""
         data, schema = data_and_schema
@@ -199,7 +203,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_multiplication_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests multiplication transformation works correctly."""
         data, schema = data_and_schema
@@ -224,7 +228,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_one_hot_encoding_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests one-hot encoding transformation works correctly."""
         data, schema = data_and_schema
@@ -252,10 +256,8 @@ class TestTransformationProcessor:
         # Check additional columns have been correctly concatenated
         assert len(data.index) == original_data_len
 
-    @fixture(params=[True, False])
-    def ohe_mini_dataset_and_results(
-        self, request: PytestRequest
-    ) -> Tuple[_DataFrameType, _DataFrameType, _DataFrameLib]:
+    @fixture
+    def ohe_mini_dataset_and_results(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
         """Provides a small dataset, results and dataframe type for tests.
 
         Generates a (4,3) test dataframe for testing one-hot encoding and a
@@ -277,40 +279,26 @@ class TestTransformationProcessor:
             "col1_2": [0, 1, 0, 0],
             "col1_UNKNOWN": [0, 0, 1, 0],
         }
-        if request.param:
-            return (
-                ks.DataFrame(data=data_in),
-                ks.DataFrame(data=results),
-                _DataFrameLib.KOALAS,
-            )
-        else:
-            return (
-                pd.DataFrame(data=data_in),
-                pd.DataFrame(data=results),
-                _DataFrameLib.PANDAS,
-            )
+
+        return (
+            pd.DataFrame(data=data_in),
+            pd.DataFrame(data=results),
+        )
 
     @unit_test
     def test__do_one_hot_encoding(
         self,
-        ohe_mini_dataset_and_results: Tuple[
-            _DataFrameType, _DataFrameType, _DataFrameLib
-        ],
+        ohe_mini_dataset_and_results: Tuple[pd.DataFrame, pd.DataFrame],
     ) -> None:
         """Tests the one hot encoding method directly.
 
         Checks that values are being encoded correctly.
         """
-        data_in, expected_results, df_type = ohe_mini_dataset_and_results
+        data_in, expected_results = ohe_mini_dataset_and_results
 
         ohe_transform = OneHotEncodingTransformation(arg="c:col1", raw_values=[1, 2])
         results = TransformationProcessor._do_one_hot_encoding(data_in, ohe_transform)
 
-        # The assertion method below only works on pandas DataFrames so need to convert
-        if df_type == _DataFrameLib.KOALAS:
-            results = results.to_pandas()
-            expected_results = expected_results.to_pandas()
-
         assert_frame_equal(results, expected_results, check_dtype=False)
 
     @pytest.mark.parametrize(
@@ -322,7 +310,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_subtraction_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests subtraction transformation works correctly."""
         data, schema = data_and_schema
@@ -347,7 +335,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_cleandata_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests clean data transformation works correctly."""
         data, schema = data_and_schema
@@ -370,7 +358,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_cleandata_transformation_no_args(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests clean data transformation with no arguments (i.e. all cols)."""
         data, schema = data_and_schema
@@ -391,7 +379,7 @@ class TestTransformationProcessor:
     )
     def test_normalize_transformation(
         self,
-        data_and_schema: Tuple[_DataFrameType, TableSchema],
+        data_and_schema: Tuple[pd.DataFrame, TableSchema],
         integer_types: IntegerTypes,
     ) -> None:
         """Tests normalize transformation works."""
@@ -428,7 +416,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_normalize_transformation_on_categorical_column(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests exception raised if normalize called on categorical column."""
         data, schema = data_and_schema
@@ -442,6 +430,20 @@ class TestTransformationProcessor:
         with pytest.raises(TransformationApplicationError):
             parse_and_process_data(data, schema, test_yaml)
 
+    @unit_test
+    def test_normalize_transformation_with_list_of_cols(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests normalization works when instantiated with a list of columns."""
+        data, schema = data_and_schema
+        transform = NormalizeDataTransformation(cols=["A", "B"])
+        processor = TransformationProcessor([transform], schema)
+        data = processor.transform(data)
+        assert round(data["A"].mean()) == 0
+        assert round(data["A"].std()) == 1
+        assert round(data["B"].mean()) == 0
+        assert round(data["B"].std()) == 1
+
     @pytest.mark.parametrize(
         "data_and_schema",
         [
@@ -452,37 +454,34 @@ class TestTransformationProcessor:
     )
     def test_normalize_transformation_no_args(
         self,
-        data_and_schema: Tuple[_DataFrameType, TableSchema],
+        data_and_schema: Tuple[pd.DataFrame, TableSchema],
         integer_types: IntegerTypes,
     ) -> None:
         """Tests NormalizeData transformation when no args supplied (all cols)."""
         data, schema = data_and_schema
+        data["E"] = 2 * data["E"]
+        data["F"] = 5 * data["F"]
+        data["G"] = 10 * data["G"]
+        data["H"] = 3 * data["H"]
+        # Check that the mean is no longer 0
+        for feature in ["E", "F", "G", "H"]:
+            assert round(data[feature].mean()) != 0
         test_yaml = """
         Transformations:
         - NormalizeData
         """
+        data, schema = parse_and_process_data(data, schema, test_yaml)
+        # Check that int columns remain unchanged
         for feature in ["A", "B", "C", "D"]:
             assert schema.features["continuous"][feature].dtype in integer_types
+        # Check that the float columns have been normalized
+        for feature in ["E", "F", "G", "H"]:
+            assert round(data[feature].mean()) == 0
+            assert round(data[feature].std()) == 1
 
-        data, schema = parse_and_process_data(data, schema, test_yaml)
-        assert round(data.A.mean()) == 0
-        assert round(data.A.std()) == 1
-        assert round(data.B.mean()) == 0
-        assert round(data.B.std()) == 1
-
-        for feature in ["A", "B", "C", "D"]:
-            assert schema.features["continuous"][feature].dtype == np.float64
-
-    @pytest.mark.parametrize(
-        "data_and_schema",
-        [
-            pytest.param(True, marks=integration_test),
-            pytest.param(False, marks=unit_test),
-        ],
-        indirect=["data_and_schema"],
-    )
+    @unit_test
     def test_errors_if_transformation_same_name_as_col(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests an error is raised when a transformation name matches a column."""
         data, schema = data_and_schema
@@ -509,7 +508,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_errors_if_multi_column_transformation_out_col_same_as_col(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests an error is raised when an MCO output column name matches a column."""
         data, schema = data_and_schema
@@ -542,7 +541,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_multi_transformation(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests the application of multiple transformations."""
         data, schema = data_and_schema
@@ -585,7 +584,7 @@ class TestTransformationProcessor:
         indirect=["data_and_schema"],
     )
     def test_multi_column_output_drop_cols(
-        self, data_and_schema: Tuple[_DataFrameType, TableSchema]
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
     ) -> None:
         """Tests that multicolumn outputs are all dropped correctly."""
         data, schema = data_and_schema
@@ -614,7 +613,7 @@ class TestTransformationProcessor:
             arg2: c:second_nonexistent_col
         """
         data = create_dataset()
-        schema = BitfountSchema(DataSource(data), table_name="test")
+        schema = BitfountSchema(DataFrameSource(data), table_name="test")
         parser = TransformationsParser()
         transformations, cols = parser.parse(test_yaml)
         processor = TransformationProcessor(transformations, schema.tables[0], cols)
@@ -637,7 +636,7 @@ class TestTransformationProcessor:
             arg2: c:I
         """
         data = create_dataset()
-        schema = BitfountSchema(DataSource(data), table_name="test")
+        schema = BitfountSchema(DataFrameSource(data), table_name="test")
         parser = TransformationsParser()
         transformations, cols = parser.parse(test_yaml)
         processor = TransformationProcessor(transformations, schema.tables[0], cols)
@@ -652,7 +651,7 @@ class TestTransformationProcessor:
         """Tests image transformation works correctly."""
         test_yaml = """
         Transformations:
-        - Image:
+        - albumentations:
             arg: col:col1
             output: True
             step: train
@@ -665,7 +664,7 @@ class TestTransformationProcessor:
         tfms, _ = parser.parse(test_yaml)
         proc = TransformationProcessor(tfms)
         image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
-        transformed_image = proc.batch_transform(image, step="train")
+        transformed_image = proc.batch_transform(image, step=DataSplit.TRAIN)
         assert isinstance(transformed_image, np.ndarray)
         assert transformed_image.shape == image.shape
 
@@ -674,7 +673,7 @@ class TestTransformationProcessor:
         """Tests that transformation is not applied if step does not match."""
         test_yaml = """
         Transformations:
-        - Image:
+        - albumentations:
             arg: col:col1
             output: True
             step: train
@@ -687,7 +686,7 @@ class TestTransformationProcessor:
         tfms, _ = parser.parse(test_yaml)
         proc = TransformationProcessor(tfms)
         image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
-        transformed_image = proc.batch_transform(image, step="validation")
+        transformed_image = proc.batch_transform(image, step=DataSplit.VALIDATION)
 
         # Assert that the 'transformed' image is exactly the same
         np.testing.assert_array_equal(transformed_image, image)
@@ -710,7 +709,137 @@ class TestTransformationProcessor:
         proc = TransformationProcessor(tfms)
         image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
         with pytest.raises(InvalidBatchTransformationError):
-            proc.batch_transform(image, step="train")
+            proc.batch_transform(image, step=DataSplit.TRAIN)
+
+    @unit_test
+    @pytest.mark.parametrize(
+        "scalar",
+        [pytest.param(2), pytest.param(0.1), pytest.param({"A": 0.2, "B": 10, "C": 4})],
+    )
+    def test_scalar_multiplication(
+        self,
+        data_and_schema: Tuple[pd.DataFrame, TableSchema],
+        scalar: Union[int, float, Mapping[str, Union[int, float]]],
+    ) -> None:
+        """Tests the scalar multiplication is applied correctly."""
+        data, schema = data_and_schema
+        orig_data = data.copy()
+        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        data = processor.transform(data)
+        if isinstance(scalar, (int, float)):
+            for feature in ["E", "F", "G", "H"]:
+                assert (orig_data[feature] * scalar).to_list() == data[
+                    feature
+                ].to_list()
+        else:
+            for feature in ["A", "B", "C"]:
+                assert (orig_data[feature] * scalar[feature]).to_list() == data[
+                    feature
+                ].to_list()
+
+    @unit_test
+    def test_scalar_multiplication_raises_error_non_existing_column(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar multiplication fails if col doesn't exist."""
+        scalar = {"X": 1.5, "Z": 0.2}
+        data, schema = data_and_schema
+        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        with pytest.raises(MissingColumnReferenceError) as tpe:
+            processor.transform(data)
+        assert (
+            tpe.value.errors.sort()
+            == [f"Reference to non-existent column: {c}" for c in scalar.keys()].sort()
+        )
+
+    @unit_test
+    def test_scalar_multiplication_raises_non_numeric_column(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar multiplication fails non-numeric column."""
+        scalar = {"Date": 1.5}
+        data, schema = data_and_schema
+        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        with pytest.raises(TransformationApplicationError):
+            processor.transform(data)
+
+    @unit_test
+    def test_scalar_multiplication_raises_error_wrong_scalar(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar multiplication if scalar is an incorrect format."""
+        scalar = [1, 2, 3]
+        with pytest.raises(TransformationApplicationError):
+            ScalarMultiplicationDataTransformation(scalar=scalar)  # type: ignore[arg-type] # Reason: incorrect assignment to check if error is raised. # noqa: B950
+
+    @unit_test
+    @pytest.mark.parametrize(
+        "scalar",
+        [pytest.param(2), pytest.param(0.1), pytest.param({"A": 0.2, "B": 10, "C": 4})],
+    )
+    def test_scalar_addition(
+        self,
+        data_and_schema: Tuple[pd.DataFrame, TableSchema],
+        scalar: Union[int, float, Mapping[str, Union[int, float]]],
+    ) -> None:
+        """Tests the scalar addition is applied correctly."""
+        data, schema = data_and_schema
+        orig_data = data.copy()
+        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        data = processor.transform(data)
+        if isinstance(scalar, (int, float)):
+            for feature in ["E", "F", "G", "H"]:
+                assert (orig_data[feature] + scalar).to_list() == data[
+                    feature
+                ].to_list()
+        else:
+            for feature in ["A", "B", "C"]:
+                assert (orig_data[feature] + scalar[feature]).to_list() == data[
+                    feature
+                ].to_list()
+
+    @unit_test
+    def test_scalar_addition_raises_error_non_existing_column(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar addition fails if col doesn't exist."""
+        scalar = {"X": 1.5, "Z": 0.2}
+        data, schema = data_and_schema
+        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        with pytest.raises(MissingColumnReferenceError) as tpe:
+            processor.transform(data)
+        assert (
+            tpe.value.errors.sort()
+            == [f"Reference to non-existent column: {c}" for c in scalar.keys()].sort()
+        )
+
+    @unit_test
+    def test_scalar_addition_raises_error_wrong_scalar(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar addition if scalar is an incorrect format."""
+        scalar = [1, 2, 3]
+        with pytest.raises(TransformationApplicationError):
+            ScalarAdditionDataTransformation(
+                scalar=scalar  # type: ignore[arg-type] # Reason: incorrect assignment to check if error is raised. # noqa: B950
+            )
+
+    @unit_test
+    def test_scalar_addition_raises_non_numeric_column(
+        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
+    ) -> None:
+        """Tests the scalar addition fails non-numeric column."""
+        scalar = {"Date": 1.5}
+        data, schema = data_and_schema
+        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
+        processor = TransformationProcessor([scalar_transform], schema)
+        with pytest.raises(TransformationApplicationError):
+            processor.transform(data)
 
 
 @unit_test
