@@ -1,10 +1,12 @@
-"""Provides methods for interacting with the Bitfount Web frontend via Selenium."""
+"""Provides methods for interacting with the Bitfount Web services."""
 from contextlib import contextmanager
+from datetime import datetime, timedelta, timezone
+import json
 import logging
 from pathlib import Path
 import threading
 import time
-from typing import Generator
+from typing import Generator, Tuple
 
 import chromedriver_autoinstaller
 from requests import HTTPError
@@ -14,20 +16,20 @@ from selenium.webdriver.remote.webdriver import WebDriver
 from selenium.webdriver.support import expected_conditions as EC
 from selenium.webdriver.support.ui import WebDriverWait
 
-from bitfount.config import _STAGING_ENVIRONMENT, _get_environment
-from bitfount.hub.api import (
-    _STAGING_AM_URL,
-    PRODUCTION_AM_URL,
-    BitfountAM,
-    BitfountSession,
+from bitfount.config import (
+    _DEVELOPMENT_ENVIRONMENT,
+    _STAGING_ENVIRONMENT,
+    _get_environment,
 )
-from bitfount.hub.authentication_flow import (
+from bitfount.hub.api import BitfountAM, BitfountSession
+from bitfount.hub.authentication_flow import _get_auth_environment
+from bitfount.hub.authentication_handlers import (
     _HUB_API_IDENTIFIER,
     _SCOPES,
-    _STAGING_AUTH_DOMAIN,
-    _STAGING_CLIENT_ID,
-    _get_auth_environment,
+    ExternallyManagedJWTHandler,
 )
+from bitfount.hub.types import _DEV_AM_URL, _STAGING_AM_URL, PRODUCTION_AM_URL
+from bitfount.utils import web_utils
 
 SCREENSHOT_DIRECTORY: Path = Path("selenium-screenshots")
 IMPLICIT_WAIT_TIME = 8  # seconds; high value due to slow GitHub runner
@@ -76,71 +78,10 @@ def _oauth_sign_in(username: str, password: str, driver: WebDriver) -> None:
         raise e
 
 
-class WebdriverBitfountSession(BitfountSession):
-    """A BitfountSession implementation using Selenium Webdriver.
-
-    An implementation of BitfountSession which uses a Selenium Webdriver to approve
-    the session access.
-    """
-
-    def __init__(
-        self,
-        password: str,
-        # Below are the inherited args
-        username: str,
-        auth_domain: str = _STAGING_AUTH_DOMAIN,
-        client_id: str = _STAGING_CLIENT_ID,
-        scopes: str = _SCOPES,
-        audience: str = _HUB_API_IDENTIFIER,
-    ):
-        logger.info(f"Webdriver is using {auth_domain} for user {username}")
-        super().__init__(
-            username=username,
-            auth_domain=auth_domain,
-            client_id=client_id,
-            scopes=scopes,
-            audience=audience,
-        )
-        self.password = password
-
-    def _do_verification(self, user_code: str, verification_uri: str) -> None:
-        """Opens web browser for verification."""
-        with webdriver_factory() as driver:
-            try:
-                # Load Auth0 location
-                driver.get(verification_uri)
-
-                # Extract code
-                secure_code_field = driver.find_element_by_css_selector(
-                    "input[aria-label='Secure code']"
-                )
-                secure_code = secure_code_field.get_attribute("value")
-                if secure_code != user_code:
-                    raise ValueError(
-                        f"Secure codes differ: {user_code} != {secure_code}"
-                    )
-
-                # If code ok, click button
-                driver.find_element_by_css_selector("button[value='confirm']").click()
-                # Perform login
-                _oauth_sign_in(self._username, self.password, driver)
-                # Wait for page to load
-                driver.find_element_by_xpath(
-                    "//p[text()='Your device is now connected.']"
-                )
-            except Exception as e:
-                logger.error(
-                    f"Exception encountered whilst attempting to perform oauth "
-                    f"verification at {verification_uri}."
-                )
-                save_screenshot("do_verification", driver)
-                raise e
-
-
 class ExtendedBitfountAM(BitfountAM):
     """Extends BitfountAM with methods relevant to testing."""
 
-    def __init__(self, session: WebdriverBitfountSession, access_manager_url: str):
+    def __init__(self, session: BitfountSession, access_manager_url: str):
         logger.info(f"Using {access_manager_url} for access manager")
         super().__init__(session, access_manager_url)
 
@@ -168,17 +109,38 @@ def get_bitfount_session(
     username: str,
     password: str,
     token_dir: Path,
-) -> WebdriverBitfountSession:
-    """Creates and returns a WebdriverBitfountSession."""
+) -> BitfountSession:
+    """Creates and returns a BitfountSession that uses Resource Owner Password Flow."""
     bf_env = _get_auth_environment()
-    session = WebdriverBitfountSession(
-        username=username,
-        password=password,
-        auth_domain=bf_env.auth_domain,
-        client_id=bf_env.client_id,
+    logger.info(f"Webdriver is using {bf_env.auth_domain} for user {username}")
+
+    def get_token() -> Tuple[str, datetime]:
+        body = {
+            "grant_type": "password",
+            "username": username,
+            "password": password,
+            "audience": _HUB_API_IDENTIFIER,
+            "scope": _SCOPES,
+            "client_id": bf_env.client_id,
+        }
+
+        response = web_utils.post(
+            f"https://{bf_env.auth_domain}/oauth/token", data=body
+        ).json()
+        logger.error(json.dumps(response))
+        return response["access_token"], datetime.now(timezone.utc) + timedelta(
+            seconds=response["expires_in"]
+        )
+
+    token, expiry = get_token()
+
+    handler = ExternallyManagedJWTHandler(token, expiry, get_token, username)
+
+    handler.user_storage_path = token_dir / username
+    session = BitfountSession(
+        authentication_handler=handler,
     )
-    session.user_storage_path = token_dir / username
-    session.token_file = session.user_storage_path / ".token"
+
     return session
 
 
@@ -239,12 +201,14 @@ def grant_proactive_access(
     modeller_username: str,
     pod_id: str,
     role: str,
-    pod_session: WebdriverBitfountSession,
+    pod_session: BitfountSession,
 ) -> None:
     """Grants proactive access to a pod for a given modeller."""
     bf_env = _get_environment()
-    if bf_env is _STAGING_ENVIRONMENT:
+    if bf_env == _STAGING_ENVIRONMENT:
         am_url = _STAGING_AM_URL
+    elif bf_env == _DEVELOPMENT_ENVIRONMENT:
+        am_url = _DEV_AM_URL
     else:
         am_url = PRODUCTION_AM_URL
 
