@@ -275,6 +275,7 @@ class TestEndToEnd:
 
         Fails tests if there are any errors.
         """
+
         # As this is used in multiprocessing, _modeller_send_task_requests() has
         # to be a def function. So we wrap the desired methods in an async def
         # and use asyncio.run() to simulate async def.
@@ -403,6 +404,12 @@ class TestEndToEnd:
                 )
                 mock_message_service_stub.return_value = mock_stub
 
+                # Mock out Key ID file handling
+                mocker.patch("bitfount.federated.modeller._store_key_id")
+                mocker.patch(
+                    "bitfount.federated.modeller._get_key_id", return_value="1"
+                )
+
                 # Ensure only testing against one pod
                 modeller_config.pods.identifiers = [pod_config.pod_id]
 
@@ -423,7 +430,7 @@ class TestEndToEnd:
                     return_value=pod.schema,
                 )
 
-                modeller, pod_identifiers, _ = setup_modeller_from_config(
+                modeller, pod_identifiers, _, _ = setup_modeller_from_config(
                     modeller_config
                 )
                 print(f"{modeller._identity_verification_method}")
@@ -572,6 +579,12 @@ class TestEndToEnd:
                 )
                 mock_message_service_stub.return_value = mock_stub
 
+                # Mock out Key ID file handling
+                mocker.patch("bitfount.federated.modeller._store_key_id")
+                mocker.patch(
+                    "bitfount.federated.modeller._get_key_id", return_value="1"
+                )
+
                 # Mock out large object handling in message service
                 apply_mock_large_object_interactions(manager)
 
@@ -583,6 +596,10 @@ class TestEndToEnd:
                     pod2_config.pod_id,
                 ]
 
+                modeller_config.task.data_structure.table_config.table = {  # type: ignore[union-attr] # reason: model algorithm has a datastructure # noqa: B950
+                    pod1_config.pod_id: "census-income-demo",
+                    pod2_config.pod_id: "census-income-demo",
+                }
                 # Generate mocked sessions for modeller and pods
                 modeller_username = modeller_config.modeller.username
                 pod1_username = pod1_config.username
@@ -602,11 +619,10 @@ class TestEndToEnd:
                 pod1 = setup_pod_from_config(pod1_config)
                 pod2 = setup_pod_from_config(pod2_config)
 
-                modeller, pod_identifiers, _ = setup_modeller_from_config(
+                modeller, pod_identifiers, _, _ = setup_modeller_from_config(
                     modeller_config
                 )
                 model_out = tmp_path / "model.out"
-
                 # Start pods and wait to spin up
                 pod_1_process = multiprocessing.Process(
                     target=pod_start, name="Pod_1_Runner", args=(pod1,)
