@@ -40,6 +40,7 @@ from bitfount.federated.types import AlgorithmType, ProtocolType
 from bitfount.federated.utils import _ALGORITHMS, _MODELS, _PROTOCOLS
 from bitfount.federated.worker import _Worker
 from bitfount.hub.api import BitfountHub
+from bitfount.hub.types import _PublicKeyJSON
 from bitfount.metrics import (
     BINARY_CLASSIFICATION_METRICS,
     MULTICLASS_CLASSIFICATION_METRICS,
@@ -54,7 +55,8 @@ from tests.utils.mocks import LocalMessageService, LocalMessageServiceSharedQueu
 # PYTEST MARKS
 # tests involving a specific backend
 backend_test = pytest.mark.backend_test
-
+# tests involving differential privacy which is an optional extra
+dp_test = pytest.mark.dp_test
 # tests that run the full end-to-end system in the form of tutorials
 tutorial_test = pytest.mark.tutorial_test
 # tests that run the full end-to-end system, but mock out web calls
@@ -571,10 +573,11 @@ def create_local_modeller_and_workers(
     data = create_dataset(classification=True)
     schema = BitfountSchema()
     pod_datasets = []
-
+    table_mapping = {"Alice/AlicePod": TABLE_NAME, "Bob/BobPod": TABLE_NAME}
     if results_only:
         # ResultsOnly only works with one pod
         pod_ids = ["Alice/AlicePod"]
+        table_mapping = {"Alice/AlicePod": TABLE_NAME}
 
     # Create datasets for each pod and combine them in the schema
     for _ in pod_ids:
@@ -591,7 +594,7 @@ def create_local_modeller_and_workers(
             ignore_cols={TABLE_NAME: ["Date"]},
         )
     data_structure = DataStructure(
-        table=TABLE_NAME, target="TARGET", ignore_cols=["Date"]
+        table=table_mapping, target="TARGET", ignore_cols=["Date"]
     )
 
     # Create local message service object
@@ -609,6 +612,7 @@ def create_local_modeller_and_workers(
     # Create model
     model_class = _MODELS[model_name]
     model = model_class(datastructure=data_structure, schema=schema, **hyperparams)
+
     # Create algorithm
     algorithm = _ALGORITHMS[AlgorithmType[algorithm_name].name](
         model=model,
@@ -688,7 +692,9 @@ def create_local_modeller_and_workers(
     mock_hub: Mock = create_autospec(BitfountHub, instance=True)
     mock_hub.username = modeller_name
     mock_modeller_private_key: Mock = create_autospec(RSAPrivateKey, instance=True)
-    mock_hub.get_user_public_key.return_value = mock_modeller_private_key.public_key()
+    mock_hub.check_public_key_registered_and_active.return_value = _PublicKeyJSON(
+        public_key=mock_modeller_private_key.public_key(), id="1", active=True
+    )
     modeller = _Modeller(
         protocol=protocol,
         message_service=LocalMessageService(
