@@ -1,20 +1,21 @@
 """File containing mock objects for the purposes of testing."""
 from __future__ import annotations
 
-from asyncio import Queue
+import asyncio
 from collections.abc import Awaitable
 import dataclasses
 import inspect
 from inspect import Parameter
 import logging
 import multiprocessing
-from multiprocessing.managers import SyncManager
+from multiprocessing.managers import DictProxy, ListProxy, SyncManager
+import queue
+import threading
 import time
 from typing import (
     Any,
     AsyncGenerator,
     AsyncIterator,
-    Collection,
     Dict,
     Generator,
     Generic,
@@ -31,6 +32,7 @@ from typing import (
     cast,
 )
 from unittest.mock import AsyncMockMixin, NonCallableMock, create_autospec
+import uuid
 
 from grpc import RpcError, StatusCode
 from typing_extensions import TypeAlias
@@ -41,18 +43,18 @@ from bitfount.federated.transport.message_service import (
     _MessageService,
 )
 from bitfount.federated.transport.protos.messages_pb2 import (
+    Acknowledgement,
     BitfountMessage as GrpcBitfountMessage,
-)
-from bitfount.federated.transport.protos.messages_pb2 import (
     BitfountTasks,
     BlobStorageData,
-    CommunicationDetails,
+    CommunicationDetails as GrpcCommunicationDetails,
     LargeStorageRequest,
     PodData,
     SuccessResponse,
 )
-from bitfount.federated.transport.protos.messages_pb2 import Acknowledgement
 from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub
+from bitfount.federated.transport.types import CommunicationDetails
+from bitfount.utils.concurrency_utils import await_threading_event
 
 logger = logging.getLogger(__name__)
 
@@ -62,19 +64,40 @@ class LocalMessageServiceSharedQueues:
 
     def __init__(self) -> None:
         """Create new LocalMessageServiceSharedQueues instance."""
-        self._queues: Dict[str, Queue[_BitfountMessage]] = {}
+        self._queues: Dict[str, queue.SimpleQueue[_BitfountMessage]] = {}
+        self._lock = threading.RLock()
 
-    async def send_message(self, mailbox_id: str, message: _BitfountMessage) -> None:
+    def send_message(self, mailbox_id: str, message: _BitfountMessage) -> None:
         """Sends a message to the queue for mailbox_id."""
-        await self._queues[mailbox_id].put(message)
+        with self._lock:
+            mailbox = self._queues[mailbox_id]
+        mailbox.put_nowait(message)
 
     async def get_message(self, mailbox_id: str) -> _BitfountMessage:
         """Awaits on a message to pull from queue for mailbox_id."""
-        return await self._queues[mailbox_id].get()
+        with self._lock:
+            mailbox = self._queues[mailbox_id]
+
+        while True:
+            try:
+                message = mailbox.get_nowait()
+                logger.debug(
+                    f"Retrieved message ({message.message_type})"
+                    f" from {message.sender}"
+                    f" from mailbox {mailbox_id}"
+                )
+                return message
+            except queue.Empty:
+                await asyncio.sleep(0)
 
     def add_mailbox(self, mailbox_id: str) -> None:
-        """Creates a new queue for mailbox_id."""
-        self._queues[mailbox_id] = Queue()
+        """Creates a new queue for mailbox_id.
+
+        If queue already exists, does nothing.
+        """
+        with self._lock:
+            if mailbox_id not in self._queues:
+                self._queues[mailbox_id] = queue.SimpleQueue()
 
 
 class LocalMessageService(_MessageService):
@@ -88,16 +111,19 @@ class LocalMessageService(_MessageService):
         self,
         username: str,
         shared_queues: LocalMessageServiceSharedQueues,
-        modeller_mailbox_id: Optional[str] = None,
-        pod_ids: Optional[Collection[str]] = None,
+        modeller_mailbox_id: str,
+        pod_ids: List[str],
+        task_id: Optional[str] = None,
     ):
         """Create new LocalMessageService.
 
         Args:
             username: The username of the account "running" this message service.
             shared_queues: The shared queues to act as the mailboxes.
-            modeller_mailbox_id: Optional. The modeller's mailbox ID to simulate with.
-            pod_ids: Optional. The group of pod_ids to simulate with.
+            modeller_mailbox_id: The modeller's mailbox ID to simulate with.
+            pod_ids: The group of pod_ids to simulate with.
+            task_id: Optional. The task ID to simulate with. If not provided, one
+                will be generated.
         """
         # We don't have a BitfountSession to keep track of username with, so we
         # need it manually.
@@ -106,6 +132,29 @@ class LocalMessageService(_MessageService):
         self._modeller_mailbox_id = modeller_mailbox_id
         self._pod_ids = pod_ids
 
+        self._task_id: str
+        if task_id is None:
+            self._task_id = uuid.uuid4().hex
+        else:
+            self._task_id = task_id
+
+        # Pre-create queues
+        # NOTE: In the actual message service these won't be created until
+        #       connect_pod() or setup_communication_with_pods() are called. For
+        #       the mock use case, because we often have the pods/workers/modellers
+        #       running in parallel, we need to ensure the queues are already created
+        #       before they might be accessed.
+        # Create modeller queue
+        self._shared_queues.add_mailbox(self.modeller_mailbox_id)
+
+        # Create pod mailboxes
+        for pod_id in self._pod_ids:
+            self._shared_queues.add_mailbox(self._get_pod_mailbox_id(pod_id))
+
+        # Create worker mailboxes
+        for worker_mailbox_id in self.worker_mailbox_ids.values():
+            self._shared_queues.add_mailbox(worker_mailbox_id)
+
     @property
     def username(self) -> str:
         """Username running this message service."""
@@ -135,6 +184,84 @@ class LocalMessageService(_MessageService):
 
     async def connect_pod(self, pod_name: str) -> str:
         """See parent for more information."""
+        # NOTE: In the actual message service the pod queue will be created at this
+        #       point (by the message service itself). For the mock we pre-create
+        #       it as the way the mock is often used will require them to already
+        #       exist.
+        return self._get_pod_mailbox_id(pod_name)
+
+    async def setup_communication_with_pods(
+        self, tasks_per_pod: Mapping[str, bytes]
+    ) -> CommunicationDetails:
+        """See parent for more information."""
+        # NOTE: In the actual message service the modeller and message queues will
+        #       be created at this point (by the message service itself). For the
+        #       mock we pre-create them as the mock is often used as though the
+        #       Pod had already approved the tasks and the workers already spun
+        #       up which means the worker queues need to already exist.
+        return CommunicationDetails(
+            self.modeller_mailbox_id, self.worker_mailbox_ids, self._task_id
+        )
+
+    async def poll_for_messages(
+        self,
+        mailbox_id: str,
+        stop_event: threading.Event,
+    ) -> AsyncGenerator[_BitfountMessage, None]:
+        """See parent for more information."""
+        # Create this at the beginning so it is only created once rather than per
+        # message yielded. Can set larger polling timeout as should only be a single
+        # instance of this and should be `set()` in event of failure or polling end
+        # (see `_BaseMailbox.listen()`).
+        stop_event_wait_task = asyncio.create_task(
+            await_threading_event(
+                stop_event,
+                event_name=f"LocalMessageService_poll_for_messages_stop_{mailbox_id}",
+                polling_timeout=30,
+            )
+        )
+
+        # Check stop condition at start of loop
+        while not stop_event.is_set():
+            get_message_task = asyncio.create_task(
+                self._shared_queues.get_message(mailbox_id)
+            )
+
+            # Check stop condition whilst waiting for message
+            awaitables: List[Awaitable] = [get_message_task, stop_event_wait_task]
+            done, pending = await asyncio.wait(
+                awaitables, return_when=asyncio.FIRST_COMPLETED
+            )
+
+            if get_message_task in done:
+                yield get_message_task.result()
+            else:  # stop_event is set
+                get_message_task.cancel()
+                return
+
+    async def send_message(
+        self, message: _BitfountMessage, already_packed: bool = False
+    ) -> SuccessResponse:
+        """See parent for more information."""
+        logger.debug(
+            f"Sending message {message.message_type} to {message.recipient_mailbox_id}"
+        )
+        self._shared_queues.send_message(message.recipient_mailbox_id, message)
+        return cast(SuccessResponse, create_autospec(SuccessResponse, instance=True))
+
+    @staticmethod
+    def _get_worker_mailbox_id(pod_identifier: str) -> str:
+        """Deterministically generates a worker mailbox ID from a pod identifier."""
+        return f"{pod_identifier}_worker"
+
+    @staticmethod
+    def _get_pod_mailbox_id(pod_identifier_or_name: str) -> str:
+        """Deterministically generate pod mailbox ID from a pod identifier or name."""
+        if "/" in pod_identifier_or_name:
+            _, pod_name = pod_identifier_or_name.split("/", maxsplit=1)
+        else:
+            pod_name = pod_identifier_or_name
+
         # TODO: [BIT-960] Currently this is just hardcoded to return the pod_name
         #       (which is what the mailbox ID will actually be). [BIT-960] will have
         #       the PodConnect method actually return the generated mailbox ID so
@@ -142,41 +269,8 @@ class LocalMessageService(_MessageService):
         #       the message service side. At that point this should return the
         #       generated mailbox ID instead.
         pod_mailbox_id = pod_name
-        # Create queue for pod
-        self._shared_queues.add_mailbox(pod_mailbox_id)
         return pod_mailbox_id
 
-    async def setup_communication_with_pods(
-        self, tasks_per_pod: Mapping[str, bytes]
-    ) -> Tuple[str, Dict[str, str]]:
-        """See parent for more information."""
-        # Create queues for modeller and for workers
-        self._shared_queues.add_mailbox(self.modeller_mailbox_id)
-
-        for worker_mailbox_id in self.worker_mailbox_ids.values():
-            self._shared_queues.add_mailbox(worker_mailbox_id)
-
-        return self.modeller_mailbox_id, self.worker_mailbox_ids
-
-    async def poll_for_messages(
-        self, mailbox_id: str
-    ) -> AsyncGenerator[_BitfountMessage, None]:
-        """See parent for more information."""
-        while True:
-            yield await self._shared_queues.get_message(mailbox_id)
-
-    async def send_message(
-        self, message: _BitfountMessage, already_packed: bool = False
-    ) -> SuccessResponse:
-        """See parent for more information."""
-        await self._shared_queues.send_message(message.recipient_mailbox_id, message)
-        return create_autospec(SuccessResponse, instance=True)
-
-    @staticmethod
-    def _get_worker_mailbox_id(pod_identifier: str) -> str:
-        """Deterministically generates a worker mailbox ID from a pod identifier."""
-        return f"{pod_identifier}_worker"
-
 
 class GRPCStubMock(MessageServiceStub):
     """Multiprocess friendly GRPC Stub Fake.
@@ -200,7 +294,7 @@ class GRPCStubMock(MessageServiceStub):
         self.queue_id_counter = manager.Value("i", 1)
         self.large_object_id_counter = manager.Value("i", 1)
 
-        self.queues: Dict[str, List[GrpcBitfountMessage]] = manager.dict()
+        self.queues: DictProxy[str, ListProxy[GrpcBitfountMessage]] = manager.dict()
 
         self.lock = multiprocessing.Lock()
 
@@ -210,7 +304,12 @@ class GRPCStubMock(MessageServiceStub):
         return metadata[0][1]
 
     async def PodConnect(
-        self, data: PodData, metadata: Sequence[Tuple[str, Any]]
+        self,
+        data: PodData,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
     ) -> SuccessResponse:
         """Fakes PodConnect behaviour."""
         pod_id = self.tokens_to_usernames[self._get_token(metadata)]
@@ -221,11 +320,18 @@ class GRPCStubMock(MessageServiceStub):
         return SuccessResponse()
 
     async def SetupTaskMailboxes(
-        self, data: BitfountTasks, metadata: Sequence[Tuple[str, Any]]
-    ) -> CommunicationDetails:
+        self,
+        data: BitfountTasks,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
+    ) -> GrpcCommunicationDetails:
         """Fakes SetupTaskMailboxes behaviour."""
         modeller_name = self.tokens_to_usernames[self._get_token(metadata)]
 
+        task_id: str = uuid.uuid4().hex
+
         with self.lock:
             modeller_queue_id = self.queue_id_counter.value
             self.queues[
@@ -258,16 +364,24 @@ class GRPCStubMock(MessageServiceStub):
                         recipient=task.podIdentifier,
                         recipientMailboxId=general_mailbox_name,
                         podMailboxIds=pod_mailbox_ids,
+                        taskId=task_id,
                     )
                 )
 
         print(f"POD MAILBOXES: {pod_mailbox_ids}")
-        return CommunicationDetails(
-            mailboxId=str(modeller_queue_id), podMailboxIds=pod_mailbox_ids
+        return GrpcCommunicationDetails(
+            mailboxId=str(modeller_queue_id),
+            podMailboxIds=pod_mailbox_ids,
+            taskId=task_id,
         )
 
     async def SendBitfountMessage(
-        self, data: GrpcBitfountMessage, metadata: Sequence[Tuple[str, Any]]
+        self,
+        data: GrpcBitfountMessage,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
     ) -> SuccessResponse:
         """Fakes SendBitfountMessage behaviour."""
         sender_name = self.tokens_to_usernames[self._get_token(metadata)]
@@ -286,12 +400,18 @@ class GRPCStubMock(MessageServiceStub):
                     recipientMailboxId=data.recipientMailboxId,
                     sender=sender_name,
                     senderMailboxId=data.senderMailboxId,
+                    taskId=data.taskId,
                 )
             )
         return SuccessResponse()
 
     async def GetBitfountMessage(
-        self, data: CommunicationDetails, metadata: Sequence[Tuple[str, Any]]
+        self,
+        data: GrpcCommunicationDetails,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
     ) -> GrpcBitfountMessage:
         """Fakes GetBitfountMessage behaviour."""
         user_name = self.tokens_to_usernames[self._get_token(metadata)]
@@ -324,7 +444,12 @@ class GRPCStubMock(MessageServiceStub):
         return message
 
     async def AcknowledgeMessage(
-        self, data: Acknowledgement, metadata: Sequence[Tuple[str, Any]]
+        self,
+        data: Acknowledgement,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
     ) -> SuccessResponse:
         """Fakes AcknowledgeMessage behaviour."""
         user_name = self.tokens_to_usernames[self._get_token(metadata)]
@@ -338,7 +463,12 @@ class GRPCStubMock(MessageServiceStub):
         return SuccessResponse()
 
     async def GetLargeObjectStorage(
-        self, data: LargeStorageRequest, metadata: Sequence[Tuple[str, Any]]
+        self,
+        data: LargeStorageRequest,
+        /,
+        *,
+        metadata: Sequence[Tuple[str, Any]],
+        timeout: Optional[float] = None,
     ) -> BlobStorageData:
         """Get URLs for upload/download of a large message."""
         # Locking just to ensure that we have a unique counter value
